<?php
/*
  Plugin Name: ARForms - PDF Creator Addon
  Description: Extension for ARForms to Create PDF from entries
  Version: 3.0
  Plugin URI: http://www.arformsplugin.com/
  Author: Repute InfoSystems
  Author URI: http://reputeinfosystems.com/
  Text Domain: ARForms-Pdfcreator
 */
if( version_compare(PHP_VERSION, '7.4','>=') ){
	error_reporting( ~E_DEPRECATED );
}
define( 'ARF_PDFCREATOR_DIR', WP_PLUGIN_DIR . '/arformspdfcreator' );

global $arfsiteurl;

$arfsiteurl = home_url();

if ( is_ssl() && ( ! preg_match( '/^https:\/\/.*\..*$/', $arfsiteurl ) || ! preg_match( '/^https:\/\/.*\..*$/', WP_PLUGIN_URL ) ) ) {
	$arfsiteurl = str_replace( 'http://', 'https://', $arfsiteurl );
	define( 'ARF_PDFCREATOR_URL', str_replace( 'http://', 'https://', WP_PLUGIN_URL . '/arformspdfcreator' ) );
} else {

	define( 'ARF_PDFCREATOR_URL', WP_PLUGIN_URL . '/arformspdfcreator' );
}


load_plugin_textdomain( 'ARForms-Pdfcreator', false, 'arformspdfcreator/languages/' );

global $arformspdfcreator, $arf_templates;
$arformspdfcreator = new ARF_Pdfcreator();

global $arformspdfcreator_version;
$arformspdfcreator_version = '3.0';

global $arformspdfcreatorshortname;
$arformspdfcreatorshortname = 'PDFC';

global $arf_entries_action_column_width;
$arf_entries_action_column_width += 37;

$ARFPDF_EXTRAFIELD_SHORTCODE = '{ARF_';
$ARFPDF_LABEL_SHORTCODE      = '{ARF_GetLabel';
$ARFPDF_FIELD_SHORTCODE      = ':';
$ARFPDF_SHORTCODE_OVER       = '}';

$ARFPDF_VALUE_SHORTCODE           = '{ARF_GetValue';
$ARFPDF_ENTRY_ID_SHORTCODE        = '{ARF_GetEntryID';
$ARFPDF_LOOP_FIRST                = '[ARF_LOOP';
$ARFPDF_LOOP_LAST                 = '[/ARF_LOOP]';
$ARFPDF_LOOP_VAR                  = 'var=';
$ARFPDF_LOOP_VAR_ONLY             = 'var';
$ARFPDF_LOOP_VAR_TO               = 'to';
$ARFPDF_IF_CONDITION_START        = '[ARF_IF';
$ARFPDF_IF_CONDITION_END          = '[/ARF_IF]';
$ARFPDF_IF_OR_CONDITION           = '||';
$ARFPDF_IF_AND_CONDITION          = '&&';
$ARFPDF_FIELDTYPE_SHORTCODE       = '{ARF_GetFieldType';
$ARFPDF_MODULE_SHORTCODE          = '{var mod';
$ARFPDF_SIGNATURE_SHORTCODE_START = '{ARF_Signature}';
$ARFPDF_SIGNATURE_SHORTCODE_END   = '{/ARF_Signature}';
$ARFPDF_FORMTITLE_SHORTCODE       = '{ARF_FormTitle}';

$ARFPDF_USERID_SHORTCODE          = '{ARF_UserID}';
$ARFPDF_USEREMAIL_SHORTCODE       = '{ARF_UserEmail}';
$ARFPDF_USERNAME_SHORTCODE        = '{ARF_UserName}';
$ARFPDF_USERDISPLAYNAME_SHORTCODE = '{ARF_UserDisplayName}';

class ARF_Pdfcreator {


	public $db_pdf_creator;
	public $pdf_header_img;

	function __construct() {

		global $arformspdfcreator, $is_version_compatible, $wpdb;
		$is_version_compatible = $this->is_arforms_version();

		$this->pdf_header_img = '';

		$this->db_pdf_creator = $wpdb->prefix . 'arf_pdf_creator';

		add_action( 'init', array( $this, 'arformspdfcreator_db_check' ) );

		register_activation_hook( __FILE__, array( 'ARF_Pdfcreator', 'install' ) );

		register_activation_hook( __FILE__, array( 'ARF_Pdfcreator', 'arf_pdfcreator_check_network_activation' ) );

		register_uninstall_hook( __FILE__, array( 'ARF_Pdfcreator', 'uninstall' ) );

		add_action( 'admin_menu', array( $this, 'arformspdfcreator_menu' ), 27 );

		add_action( 'admin_enqueue_scripts', array( $this, 'arf_set_js' ), 11 );

		add_action( 'wp_ajax_pdfcreatorgetformfields', array( $this, 'arf_pdfcreator_get_form_fields' ) );

		add_action( 'wp_ajax_pdf_import_font', array( $this, 'arf_pdf_import_font' ) );

		add_action( 'wp_ajax_pdf_imported_font_progress', array( $this, 'pdf_imported_font_progress' ) );

		add_action( 'wp_ajax_pdf_img_upload', array( $this, 'pdf_img_upload' ) );

		add_action( 'wp_ajax_gettemplate', array( $this, 'arf_pdfcreator_get_template' ) );

		add_action( 'wp_ajax_pdfcreator_save_data', array( $this, 'pdfcreator_save_data' ) );

		add_action( 'wp_ajax_delete_pdf_template', array( $this, 'delete_pdf_template' ) );

		add_filter( 'arf_update_form_option_outside', array( $this, 'arf_pdf_form_option_update_ids' ), 10, 4 );

		add_filter( 'arf_update_form_return_json_outside', array( $this, 'arf_pdf_form_option_update_return_json' ), 10, 2 );

		add_action( 'admin_notices', array( $this, 'arformspdfcreator_admin_notices' ) );

		add_action( 'admin_enqueue_scripts', array( $this, 'arf_set_css' ), 11 );

		add_filter( 'upgrader_pre_install', array( $this, 'arformspdfcreator_backup' ), 10, 2 );

		add_action( 'admin_init', array( $this, 'upgrade_arformspdfcreator_data' ), 9 );

		add_action( 'init', array( $this, 'load_mpdf_library' ), 10, 3 );

		add_action( 'arf_additional_action_entries', array( $this, 'arf_create_pdf' ), 10, 3 );

		add_action( 'arf_additional_autoresponder_settings', array( $this, 'arfpdfcreator_setting_v3' ), 10, 2 );

		add_filter( 'arfformoptionsbeforeupdateform', array( $this, 'arf_pdfcreator_save_opt' ), 10, 2 );

		add_filter( 'arfnotificationattachment', array( $this, 'arf_pdfcreator_sendpdf_with_mailattachment' ), 10, 3 );

		add_filter( 'arfautoresponderattachment', array( $this, 'arf_pdfcreator_sendpdf_with_mailattachment_touser' ), 10, 3 );

		add_filter( 'arf_check_date_from_outside', array( $this, 'arf_check_date_from_outside' ), 10 );

		add_action( 'wp_ajax_arf_pdf_dismiss_font_notice', array( $this, 'arf_dismiss_font_update_notice') );
	}

	function arf_check_date_from_outside( $addon_list ) {
		array_push( $addon_list, 'arformspdfcreator' );
		return $addon_list;
	}

	public static function arf_pdfcreator_check_network_activation( $network_wide ) {
		if ( ! $network_wide ) {
			return;
		}

		deactivate_plugins( plugin_basename( __FILE__ ), true, true );

		header( 'Location: ' . network_admin_url( 'plugins.php?deactivate=true' ) );
		exit;
	}

	function arf_pdfcreator_getapiurl() {
		$api_url = 'https://www.arpluginshop.com/';
		return $api_url;
	}

	function arf_pdfcreator_get_remote_post_params( $plugin_info = '' ) {
		global $wpdb, $arfversion;

		$action = '';
		$action = $plugin_info;

		if ( ! function_exists( 'get_plugins' ) ) {
			include_once ABSPATH . 'wp-admin/includes/plugin.php';
		}
		$plugin_list = get_plugins();
		$site_url    = home_url();
		$plugins     = array();

		$active_plugins = get_option( 'active_plugins' );

		foreach ( $plugin_list as $key => $plugin ) {
			$is_active = in_array( $key, $active_plugins );

			if ( strpos( strtolower( $plugin['Title'] ), 'arformspdfcreator' ) !== false ) {
				$name      = substr( $key, 0, strpos( $key, '/' ) );
				$plugins[] = array(
					'name'      => $name,
					'version'   => $plugin['Version'],
					'is_active' => $is_active,
				);
			}
		}
		$plugins = json_encode( $plugins );

		$theme            = wp_get_theme();
		$theme_name       = $theme->get( 'Name' );
		$theme_uri        = $theme->get( 'ThemeURI' );
		$theme_version    = $theme->get( 'Version' );
		$theme_author     = $theme->get( 'Author' );
		$theme_author_uri = $theme->get( 'AuthorURI' );

		$im        = is_multisite();
		$sortorder = get_option( 'arfSortOrder' );

		$post = array(
			'wp'        => get_bloginfo( 'version' ),
			'php'       => phpversion(),
			'mysql'     => $wpdb->db_version(),
			'plugins'   => $plugins,
			'tn'        => $theme_name,
			'tu'        => $theme_uri,
			'tv'        => $theme_version,
			'ta'        => $theme_author,
			'tau'       => $theme_author_uri,
			'im'        => $im,
			'sortorder' => $sortorder,
		);

		return $post;
	}

	function arformspdfcreator_db_check() {

		global $arformspdfcreator;

		$arformspdfcreator_version = get_option( 'arformspdfcreator_version' );

		if ( ! isset( $arformspdfcreator_version ) || '' == $arformspdfcreator_version ) {
			$arformspdfcreator->install();
		}
	}

	public static function install() {

		$arformspdfcreator_version = get_option( 'arformspdfcreator_version' );

		if ( ! isset( $arformspdfcreator_version ) || '' == $arformspdfcreator_version ) {

			global $wpdb, $arformspdfcreator_version, $arformspdfcreator;

			update_option( 'arformspdfcreator_version', $arformspdfcreator_version );

			$charset_collate = '';
			if ( ! empty( $wpdb->charset ) ) {
				$charset_collate = "DEFAULT CHARACTER SET $wpdb->charset";
			}
			if ( ! empty( $wpdb->collate ) ) {
				$charset_collate .= " COLLATE $wpdb->collate";
			}
			if ( file_exists( ABSPATH . 'wp-admin/includes/upgrade.php' ) ) {
				include_once ABSPATH . 'wp-admin/includes/upgrade.php';
			}
			$table_name = $arformspdfcreator->db_pdf_creator;
			$sql        = "CREATE TABLE IF NOT EXISTS {$table_name} (
					id int(11) NOT NULL auto_increment,		
					template_name varchar(100) NOT NULL,
					template_content longtext NOT NULL,
					template_header text NOT NULL,
					template_footer text NOT NULL,
					template_settings text NOT NULL,
					template_header_img text NOT NULL,
					created_at datetime NOT NULL,
					PRIMARY KEY (id)
				  ) {$charset_collate};";
			dbDelta( $sql );

			$arformspdfcreator->getwpversion();

			$templates = $arformspdfcreator->arforms_pdfcreator_templates();
			foreach ( $templates as $template => $template_arr ) {
				$tmp_name = str_replace( '_', ' ', $template );

				$content    = $template_arr['template_content'];
				$header     = maybe_serialize( $template_arr['template_header'] );
				$footer     = maybe_serialize( $template_arr['template_footer'] );
				$settings   = maybe_serialize( $template_arr['template_settings'] );
				$header_img = maybe_serialize( $template_arr['template_header_img'] );

				$wpdb->insert(
					$arformspdfcreator->db_pdf_creator,
					array(
						'template_name'       => $tmp_name,
						'template_content'    => addslashes( $content ),
						'template_header'     => $header,
						'template_footer'     => $footer,
						'template_settings'   => $settings,
						'template_header_img' => $header_img,
						'created_at'          => current_time( 'mysql' ),
					),
					array( '%s', '%s', '%s', '%s', '%s', '%s', '%s' )
				);
			}
		}
	}

	public static function uninstall() {
		global $wpdb, $arformspdfcreator;
		if ( is_multisite() ) {
			$blogs = $wpdb->get_results( "SELECT blog_id FROM {$wpdb->blogs}", ARRAY_A );
			if ( $blogs ) {
				foreach ( $blogs as $blog ) {
					switch_to_blog( $blog['blog_id'] );

					delete_option( 'old_arformspdfcreator_version' );
					$wpdb->query( 'DROP TABLE IF EXISTS ' . $arformspdfcreator->db_pdf_creator );
				}
				restore_current_blog();
			}
		} else {
			delete_option( 'arformspdfcreator_version' );
			$wpdb->query( 'DROP TABLE IF EXISTS ' . $arformspdfcreator->db_pdf_creator );
		}
	}

	function arformspdfcreator_admin_notices() {

		global $arformspdfcreator, $wp_version;

		if ( version_compare( $wp_version, '4.5.0', '<' ) ) {
			deactivate_plugins( __FILE__, false, null );
			echo "<div class='updated'><p>Please meet the minimum requirement of WordPress 4.5 to active ARForms - PDF Creator Add-on</p></div>";
		}

		if ( ! $arformspdfcreator->is_arforms_support() ) {
			echo "<div class='updated'><p>" . esc_html__( 'Pdfcreator For ARForms plugin requires ARForms Plugin installed and active.', 'ARForms-Pdfcreator' ) . '</p></div>';
		} elseif ( ! version_compare( $arformspdfcreator->get_arforms_version(), '3.3.2', '>=' ) ) {
			echo "<div class='updated'><p>" . esc_html__( 'Pdfcreator For ARForms plugin requires ARForms plugin installed with version 3.3.2 or higher.', 'ARForms-Pdfcreator' ) . '</p></div>';
		}

		if( 'yes' == get_option('arf_pdfcreator_font_notice') ){
			wp_enqueue_script('arformspdfcreatorview-js');
			$script_data = 'jQuery(document).on("click",".notice-dismiss",function(){
				if( jQuery(this).parent().hasClass("arf_pdfcreator_font_notice") ){
					jQuery.ajax({
						url:ajaxurl,
						data:"action=arf_pdf_dismiss_font_notice",
						type:"POST",
						success:function(){}
					});
				}
			});';

			wp_add_inline_script( 'arformspdfcreatorview-js', $script_data, 'after' );
			echo "<div class='notice notice-warning is-dismissible arf_pdfcreator_font_notice'><p>" . esc_html__( 'If you have imported any fonts from ARForms PDF Addon editor, then please re-import those fonts again after updating the plugin to the latest version. After re-importing fonts, it will be loaded properly in PDF.','ARForms-Pdfcreator')."</p></div>";
		}
	}

	function arf_dismiss_font_update_notice(){
		update_option('arf_pdfcreator_font_notice', 'no' );
		die;
	}

	function is_arforms_support() {

		include_once ABSPATH . 'wp-admin/includes/plugin.php';

		return is_plugin_active( 'arforms/arforms.php' );
	}

	function get_arforms_version() {

		$arf_db_version = get_option( 'arf_db_version' );

		return ( isset( $arf_db_version ) ) ? $arf_db_version : 0;
	}

	function arformspdfcreator_menu() {

		global $arformspdfcreator;
		if ( version_compare( $arformspdfcreator->get_arforms_version(), '3.3.2', '>=' ) ) {

			if ( current_user_can( 'administrator' ) ) {
				global $current_user;
				$pdfcreatorcapabilities                                     = array(
					'arformspdfcreator' => esc_html__( 'Configure ARForms Pdfcreator', 'ARForms-Pdfcreator' ),
				);
				
				$arfroles = $pdfcreatorcapabilities;

				foreach ( $arfroles as $arfrole => $arfroledescription ) {
					$current_user->add_cap( $arfrole );
				}

				unset( $arfroles );
				unset( $arfrole );
				unset( $arfroledescription );
			}

			if ( file_exists( ARF_PDFCREATOR_DIR . '/core/view/pdf_template_editor.php' ) ) {
				include_once ARF_PDFCREATOR_DIR . '/core/view/pdf_template_editor.php';
			}

			add_submenu_page(
				'ARForms',
				'ARForms | ' . addslashes( esc_html__( 'PDF Template Editor', 'ARForms-Pdfcreator' ) ),
				'<span>' . addslashes( esc_html__( 'PDF Template Editor', 'ARForms-Pdfcreator' ) ) . '</span>',
				'arformspdfcreator',
				'ARForms-PDF-editor',
				'template_editor_view'
			);
		}
	}

	function arf_set_js() {
		global $arformspdfcreator_version;
		$inline_script = "";
		wp_register_script( 'arformspdfcreatorview-js', ARF_PDFCREATOR_URL . '/js/arformspdfcreator.js', array(), $arformspdfcreator_version );

		wp_register_script( 'arformspdfcreatoreditor-js', ARF_PDFCREATOR_URL . '/js/arformspdfcreator_editor.js', array(), $arformspdfcreator_version );

		wp_register_script( 'jquery-json', ARFURL . '/js/jquery/jquery.json-2.4.js', array(), $arformspdfcreator_version );

		if ( isset( $_REQUEST['page'] ) && '' != $_REQUEST['page'] && 'ARForms-PDF-editor' == $_REQUEST['page'] ) {
			wp_enqueue_script( 'arformspdfcreatorview-js' );

			$script_data .= "var __TEMPLATE_NAME_MSG = '".esc_html__('Please select template name', 'ARForms-Pdfcreator')."';";
			$script_data .= "var __BLANK_TEMPLATE_NAME_MSG = '".esc_html__('Please fill Template Name.', 'ARForms-Pdfcreator')."';";

			wp_add_inline_script( 'arformspdfcreatorview-js', $script_data );

			wp_enqueue_script( 'jquery-json' );
		} elseif ( isset( $_REQUEST['page'] ) && '' != $_REQUEST['page'] && 'ARForms' == $_REQUEST['page'] && ( ( isset( $_REQUEST['arfaction'] ) && ( 'edit' == $_REQUEST['arfaction'] || 'new' == $_REQUEST['arfaction'] ) ) ) ) {
			wp_enqueue_script( 'arformspdfcreatoreditor-js' );
		}
	}

	function arf_pdfcreator_get_form_fields() {
		global $wpdb, $MdlDb, $arformspdfcreator;
		$output  = '';
		$form_id = ( isset( $_POST['data'] ) ) ? $_POST['data'] : 0;
		if ( $form_id == 0 ) {
			exit;
		}

		$exclude = array( 'captcha', 'break' );

		$fields = $wpdb->get_results( $wpdb->prepare( 'SELECT id, name, type, field_options FROM ' . $MdlDb->fields . ' WHERE form_id=%d', $form_id ), OBJECT );
		if ( ! empty( $fields ) ) {
			$i = 0;
			foreach ( $fields as $field ) {
				if ( ! in_array( $field->type, $exclude ) ) {
					$field_opts = json_decode( $field->field_options );
					if( isset( $field_opts->parent_field_type ) && 'arf_repeater' == $field_opts->parent_field_type ){
						continue;
					}
					$output .= "<div class='arf_pdf_form_fields' onclick='open_pdf_field_option_dropdown(this);'>
						<div class='arf_pdf_button' data-value='" . $field->id . "' data-label='" . $field->name . "' onclick='add_field_to_editor(this);' >" . $field->name . ' (ID:' . $field->id . ')</div>
					</div>
					';
				}
				$i++;
			}
		}
		echo $output;
		exit;
	}

	function arf_pdfcreator_get_template() {
		global $wpdb, $MdlDb, $arformspdfcreator;
		$res         = array();
		$template_id = ( $_POST['data'] != '' ) ? $_POST['data'] : 0;
		if ( $template_id == 0 ) {
			echo json_encode( $res );
			exit;
		}

		$result = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $arformspdfcreator->db_pdf_creator . ' WHERE id=%d', $template_id ) );

		if ( ! empty( $result ) ) {
			$settings                = maybe_unserialize( $result->template_settings );
			$res['template_content'] = stripcslashes( $result->template_content );

			$header_arr                    = maybe_unserialize( $result->template_header );
			$res['template_header_left']   = ( $header_arr['left'] != '' ) ? $header_arr['left'] : '';
			$res['template_header_center'] = ( $header_arr['center'] != '' ) ? $header_arr['center'] : '';
			$res['template_header_right']  = ( $header_arr['right'] != '' ) ? $header_arr['right'] : '';

			$footer_arr                    = maybe_unserialize( $result->template_footer );
			$res['template_footer_left']   = ( $footer_arr['left'] != '' ) ? $footer_arr['left'] : '';
			$res['template_footer_center'] = ( $footer_arr['center'] != '' ) ? $footer_arr['center'] : '';
			$res['template_footer_right']  = ( $footer_arr['center'] != '' ) ? $footer_arr['right'] : '';

			$res['template_size']        = ( $settings['arf_pdf_template_size'] != '' ) ? $settings['arf_pdf_template_size'] : '';
			$res['template_orientation'] = ( $settings['arf_pdf_template_orientation'] != '' ) ? $settings['arf_pdf_template_orientation'] : '';
			$res['template_font']        = ( $settings['arf_pdf_template_font'] != '' ) ? $settings['arf_pdf_template_font'] : '';
			$res['pdf_pageno']           = ( $settings['arf_pdf_pgno'] != '' ) ? $settings['arf_pdf_pgno'] : '0';

			$pdf_header_img              = maybe_unserialize( $result->template_header_img );
			$res['pdf_header_img_path']  = ( $pdf_header_img['pdf_header_img_name'] != '' ) ? site_url() . '/wp-content/uploads/arforms/' . $pdf_header_img['pdf_header_img_name'] : '';
			$res['pdf_header_img_name']  = $pdf_header_img['pdf_header_img_name'];
			$res['pdf_header_img_align'] = $pdf_header_img['pdf_header_img_align'];
		}

		echo json_encode( $res );
		exit;

	}

	function download_file( $url, $path ) {
		$newfilename = $path;
		$file        = fopen( $url, 'rb' );
		if ( $file ) {
			$newfile = fopen( $newfilename, 'wb' );

			if ( $newfile ) {
				while ( ! feof( $file ) ) {
					fwrite( $newfile, fread( $file, 1024 * 8 ), 1024 * 8 );
				}
			}
		}

		if ( $file ) {
			fclose( $file );
		}
		if ( $newfile ) {
			fclose( $newfile );
		}
	}

	function pdf_imported_font_progress() {
		$continue = true;
		@session_start();
		$destroy = false;
		if( $_SESSION['imported_fonts'] == $_SESSION['total_fonts'] && $_SESSION['imported_fonts'] != '' && $_SESSION['total_fonts'] != '' ){
			$continue = false;
			$destroy = true;
		}
		$arr = array(
			'imported_fonts' => $_SESSION['imported_fonts'],
			'total_fonts'    => $_SESSION['total_fonts'],
			'continue' 		 => $continue
		);
		if( $destroy ){
			unset( $_SESSION['imported_fonts'] );
			unset( $_SESSION['total_fonts'] );
		}
		echo json_encode( $arr );
		exit;
	}
	function arf_pdf_import_font() {
		global $wpdb, $MdlDb, $arformspdfcreator;
		$res_arr    = array();
		$fonts_data = json_decode( stripslashes_deep( $_REQUEST['data'] ), true );
		@session_start();
		if ( isset( $fonts_data['important_extra_fonts'] ) ) {
			$_SESSION['total_fonts']    = count( $fonts_data['arf_pdf_extra_fonts'] );
			$_SESSION['imported_fonts'] = 0;
			$font_arr                   = array();
			$font_list                  = $this->arforms_pdf_get_fonts_arr();
			$main_font_copied           = true;
			$fonts_cnt                  = 0;
			$child_font                 = true;
			$sub_font_err_cnt           = 0;

			foreach ( $font_list as $fontkey => $font_val ) {
				if ( isset( $fonts_data['arf_pdf_extra_fonts'][ $fontkey ] ) ) {

					foreach ( $font_val as $font_files ) {
						foreach ( $font_files as $font ) {
							$sub_font_cnt++;
							$destination = ARF_PDFCREATOR_DIR . '/libs/ttfonts/' . $font;
							$source      = 'https://www.arformsplugin.com/arf_pdfcreator_fonts/?action=arf_pdf_import_form&font=' . $font;

							$args = array(
								'timeout' => 50,
							);

							$response = wp_remote_get( $source, $args );

							if ( isset( $response['body'] ) && $response['body'] != '' ) {
								$newfile = fopen( $destination, 'wb' );
								if ( $newfile ) {
									   fwrite( $newfile, $response['body'] );
								} else {
									$child_font = false;
									$sub_font_err_cnt++;
								}
							} else {
								$child_font = false;
								$sub_font_err_cnt++;
							}
						}
					}
					if ( $child_font == false ) {
						if ( $sub_font_err_cnt > 0 && $sub_font_err_cnt != count( $font_files ) ) {
							   $fonts_cnt++;
							   $_SESSION['imported_fonts']++;

							@session_write_close();
							session_start();
						} else {
							unset( $fonts_data['arf_pdf_extra_fonts'][ $fontkey ] );
							$_SESSION['total_fonts'] = $_SESSION['total_fonts'] - 1;
						}
					} else {
						$fonts_cnt++;
						$_SESSION['imported_fonts']++;

						@session_write_close();
						session_start();
					}
				} else {
					$main_font_copied = false;

				}
			}

			if ( $fonts_cnt == 0 ) {
				$res_arr['state']   = 'fail';
				$res_arr['message'] = esc_html__( 'There is something error while import fonts. Please try again later.', 'ARForms-Pdfcreator' );
			} elseif ( $fonts_cnt > 0 && ( $child_font == false && $sub_font_err_cnt > 0 ) ) {
				$res_arr['state']   = 'warning';
				$res_arr['message'] = esc_html__( 'Some of the fonts are not imported. Please try again later.', 'ARForms-Pdfcreator' );
			} else {
				$res_arr['state']   = 'success';
				$res_arr['message'] = esc_html__( 'Fonts imported successfully.', 'ARForms-Pdfcreator' );
			}
		} else {
			$res_arr['state']   = 'fail';
			$res_arr['message'] = esc_html__( 'There is something error while import fonts. Please try again later.', 'ARForms-Pdfcreator' );
		}
		echo json_encode( $res_arr );
		die;
	}

	function pdf_img_upload() {

		if ( isset( $_FILES['img_file'] ) ) {
			$file_tmp  = $_FILES['img_file']['tmp_name'];
			$file_type = $_FILES['img_file']['type'];
			$file_size = $_FILES['img_file']['size'];
			$file_name = $_FILES['img_file']['name'];

			$ext = strtolower( ltrim( strrchr( $file_name, '.' ), '.' ) );

			if ( $ext == 'jpg' || $ext == 'jpeg' || $ext == 'bmp' || $ext == 'gif' || $ext == 'png' ) {

				$destination = ABSPATH . 'wp-content/uploads/arforms/' . time() . '_' . $file_name;
				if ( move_uploaded_file( $file_tmp, $destination ) === false ) {
					$res['state']    = 'fail';
					$res['response'] = esc_html__( 'File could not be moved.', 'ARForms-Pdfcreator' );
					echo json_encode( $res );
				} else {

					$res['state']    = 'success';
					$res['response'] = time() . '_' . $file_name;
					$res['img_path'] = site_url() . '/wp-content/uploads/arforms/' . time() . '_' . $file_name;
					echo json_encode( $res );
				}
			} else {
				$res['state']    = 'invalid file';
				$res['response'] = esc_html__( 'Please select only image file.', 'ARForms-Pdfcreator' );
				echo json_encode( $res );
			}
		} else {
			$res['state']    = 'error';
			$res['response'] = esc_html__( 'No file selected.', 'ARForms-Pdfcreator' );
			echo json_encode( $res );
		}
		exit;
	}

	function delete_pdf_template() {
		$response = array();
		if ( isset( $_POST['tmpl_id'] ) && $_POST['tmpl_id'] > 0 ) {
			global $wpdb, $MdlDb, $arformspdfcreator;

			$res = $wpdb->delete(
				$this->db_pdf_creator,
				array( 'id' => $_POST['tmpl_id'] ),
				array( '%d' )
			);

			if ( $res ) {
				   $response['message'] = esc_html__( 'Template deleted successfully.', 'ARForms-Pdfcreator' );
				   $response['state']   = 'success';
			} else {
				$response['message'] = esc_html__( 'Fail to delete Template.', 'ARForms-Pdfcreator' );
				$response['state']   = 'error';
			}
		} else {
			$response['message'] = esc_html__( 'No templates found.', 'ARForms-Pdfcreator' );
			$response['state']   = 'fail';
		}
		echo json_encode( $response );
		exit;
	}

	function pdfcreator_save_data() {
		global $wpdb, $MdlDb, $arformspdfcreator;

		$errors      = false;
		$errors_data = '';
		$pdfdata     = json_decode( stripslashes_deep( $_REQUEST['data'] ), true );

		if ( isset( $pdfdata['state'] ) && $pdfdata['state'] == 'update_pdfcreator_data' ) {
			$id           = ( $pdfdata['pdf_template_selector'] ) ? $pdfdata['pdf_template_selector'] : '0';
			$html_content = addslashes( $pdfdata['arf_pdfcreator_editor'] );

			$footer = array(
				'left'   => addslashes( $pdfdata['arf_pdfcreator_editor_footer_left'] ),
				'center' => addslashes( $pdfdata['arf_pdfcreator_editor_footer_center'] ),
				'right'  => addslashes( $pdfdata['arf_pdfcreator_editor_footer_right'] ),
			);

			$header = array(
				'left'   => addslashes( $pdfdata['arf_pdfcreator_editor_header_left'] ),
				'center' => addslashes( $pdfdata['arf_pdfcreator_editor_header_center'] ),
				'right'  => addslashes( $pdfdata['arf_pdfcreator_editor_header_right'] ),
			);

			$pdf_header_img = array(
				'pdf_header_img_name'  => $pdfdata['pdf_header_img_name'],
				'pdf_header_img_align' => $pdfdata['pdf_header_img_align'],
			);

			$result = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $this->db_pdf_creator . ' WHERE id=%d', $id ), ARRAY_A );

			if ( ! empty( $result ) ) {
				$res = $wpdb->update(
					$this->db_pdf_creator,
					array(
						'template_content'    => $html_content,
						'template_header'     => maybe_serialize( $header ),
						'template_footer'     => maybe_serialize( $footer ),
						'template_settings'   => maybe_serialize( $pdfdata['pdf_settings'] ),
						'template_header_img' => maybe_serialize( $pdf_header_img ),
						'created_at'          => current_time( 'mysql' ),
					),
					array( 'id' => $id ),
					array( '%s', '%s', '%s', '%s', '%s' ),
					array( '%d' )
				);

				if ( $res ) {
					   $message = esc_html__( 'Template updated successfully.', 'ARForms-Pdfcreator' );
					   $res     = array(
						   'state'                  => 'update',
						   'id'                     => $id,
						   'template_name'          => $pdf_template_name,
						   'template_content'       => stripcslashes( $html_content ),
						   'template_header_left'   => $pdfdata['arf_pdfcreator_editor_header_left'],
						   'template_header_center' => $pdfdata['arf_pdfcreator_editor_header_center'],
						   'template_header_right'  => $pdfdata['arf_pdfcreator_editor_header_right'],
						   'template_footer_left'   => $pdfdata['arf_pdfcreator_editor_footer_left'],
						   'template_footer_center' => $pdfdata['arf_pdfcreator_editor_footer_center'],
						   'template_footer_right'  => $pdfdata['arf_pdfcreator_editor_footer_right'],
						   'font'                   => $pdfdata['pdf_settings']['arf_pdf_template_font'],
						   'orientation'            => $pdfdata['pdf_settings']['arf_pdf_template_orientation'],
						   'size'                   => $pdfdata['pdf_settings']['arf_pdf_template_size'],
						   'message'                => $message,
					   );
					   echo json_encode( $res );
					   exit;
				} else {
					$message = esc_html__( 'There is an error to update template.', 'ARForms-Pdfcreator' );
					$res     = array(
						'state'                  => 'error',
						'id'                     => $id,
						'template_name'          => '',
						'template_content'       => '',
						'template_header_left'   => '',
						'template_header_center' => '',
						'template_header_right'  => '',
						'template_footer_left'   => '',
						'template_footer_center' => '',
						'template_footer_right'  => '',
						'font'                   => '',
						'orientation'            => '',
						'size'                   => '',
						'message'                => $message,
					);
					echo json_encode( $res );
					exit;
				}
			}
		}

		if ( isset( $pdfdata['state'] ) && $pdfdata['state'] == 'save_pdfcreator_data' ) {
			$id                = ( $pdfdata['pdf_template_selector'] ) ? $pdfdata['pdf_template_selector'] : '0';
			$pdf_template_name = trim( $pdfdata['arf_pdf_template_name'] );
			$html_content      = addslashes( $pdfdata['arf_pdfcreator_editor'] );

			$footer = array(
				'left'   => addslashes( $pdfdata['arf_pdfcreator_editor_footer_left'] ),
				'center' => addslashes( $pdfdata['arf_pdfcreator_editor_footer_center'] ),
				'right'  => addslashes( $pdfdata['arf_pdfcreator_editor_footer_right'] ),
			);

			$header = array(
				'left'   => addslashes( $pdfdata['arf_pdfcreator_editor_header_left'] ),
				'center' => addslashes( $pdfdata['arf_pdfcreator_editor_header_center'] ),
				'right'  => addslashes( $pdfdata['arf_pdfcreator_editor_header_right'] ),
			);

			$pdf_header_img = array(
				'pdf_header_img_name'  => $pdfdata['pdf_header_img_name'],
				'pdf_header_img_align' => $pdfdata['pdf_header_img_align'],
			);

			$res = $wpdb->insert(
				$this->db_pdf_creator,
				array(
					'template_name'       => stripcslashes( $pdf_template_name ),
					'template_content'    => $html_content,
					'template_header'     => maybe_serialize( $header ),
					'template_footer'     => maybe_serialize( $footer ),
					'template_settings'   => maybe_serialize( $pdfdata['pdf_settings'] ),
					'template_header_img' => maybe_serialize( $pdf_header_img ),
					'created_at'          => current_time( 'mysql' ),
				),
				array( '%s', '%s', '%s', '%s', '%s', '%s' )
			);

			$last_id = $wpdb->insert_id;

			if ( $res ) {
				$message = esc_html__( 'Template save successfully.', 'ARForms-Pdfcreator' );
				$res     = array(
					'state'                  => 'save',
					'id'                     => $last_id,
					'template_name'          => $pdf_template_name,
					'template_content'       => stripcslashes( $html_content ),
					'template_header_left'   => $pdfdata['arf_pdfcreator_editor_header_left'],
					'template_header_center' => $pdfdata['arf_pdfcreator_editor_header_center'],
					'template_header_right'  => $pdfdata['arf_pdfcreator_editor_header_right'],
					'template_footer_left'   => $pdfdata['arf_pdfcreator_editor_footer_left'],
					'template_footer_center' => $pdfdata['arf_pdfcreator_editor_footer_center'],
					'template_footer_right'  => $pdfdata['arf_pdfcreator_editor_footer_right'],
					'font'                   => $pdfdata['pdf_settings']['arf_pdf_template_font'],
					'orientation'            => $pdfdata['pdf_settings']['arf_pdf_template_orientation'],
					'size'                   => $pdfdata['pdf_settings']['arf_pdf_template_size'],
					'message'                => $message,
				);
				echo json_encode( $res );
				exit;
			} else {
				$message = esc_html__( 'There is an error to save template.', 'ARForms-Pdfcreator' );
				$res     = array(
					'state'                  => 'error',
					'id'                     => '',
					'template_name'          => '',
					'template_content'       => '',
					'template_header_left'   => '',
					'template_header_center' => '',
					'template_header_right'  => '',
					'template_footer_left'   => '',
					'template_footer_center' => '',
					'template_footer_right'  => '',
					'font'                   => '',
					'orientation'            => '',
					'size'                   => '',
					'message'                => $message,
				);
				echo json_encode( $res );
				exit;
			}
		}
	}

	function arf_pdf_form_option_update_ids( $options, $return_json_data, $values_old_id, $values_new_id ) {

		if ( ! empty( $options['arf_pdfcreator_template_name'] ) ) {
			$options['arf_pdfcreator_template_name']          = str_replace( $values_old_id, $values_new_id, $options['arf_pdfcreator_template_name'] );
			$return_json_data['arf_pdfcreator_template_name'] = $options['arf_pdfcreator_template_name'];
		}

		return $options;
	}

	function arf_pdf_form_option_update_return_json( $return_json_data, $options ) {

		if ( ! empty( $options['arf_pdfcreator_template_name'] ) ) {
			$return_json_data['arf_pdfcreator_template_name'] = $options['arf_pdfcreator_template_name'];
		}
		return $return_json_data;
	}

	function arf_set_css( $hook ) {

		global $arformspdfcreator_version;
		wp_register_style( 'arformspdfcreator-css', ARF_PDFCREATOR_URL . '/css/arformspdfcreator.css', array(), $arformspdfcreator_version );

		wp_register_style( 'arformspdfcreatorview-css', ARF_PDFCREATOR_URL . '/css/arformspdfcreator_view.css', array(), $arformspdfcreator_version );

		wp_register_style( 'fontawesome', ARFURL . '/css/font-awesome.min.css', array(), $arformspdfcreator_version );

		wp_register_style( 'arformspdfcreatoreditor-css', ARF_PDFCREATOR_URL . '/css/arformspdfcreator_editor.css', array(), $arformspdfcreator_version );

		if ( isset( $_REQUEST['page'] ) && $_REQUEST['page'] != '' && $_REQUEST['page'] == 'ARForms-entries' ) {
			wp_enqueue_style( 'arformspdfcreator-css' );
		}

		if ( isset( $_REQUEST['page'] ) && $_REQUEST['page'] != '' && $_REQUEST['page'] == 'ARForms-PDF-editor' ) {
			wp_enqueue_style( 'arformspdfcreatorview-css' );
			wp_enqueue_style( 'fontawesome' );
		}

		if ( isset( $_REQUEST['page'] ) && $_REQUEST['page'] == 'ARForms' && ( isset( $_REQUEST['arfaction'] ) && ( 'new' == $_REQUEST['arfaction'] || 'edit' == $_REQUEST['arfaction'] ) ) ) {
			wp_enqueue_style( 'arformspdfcreatoreditor-css' );
		}

		if ( isset( $hook ) && 'plugins.php' == $hook ) {
			global $wp_version;
			if ( version_compare( $wp_version, '4.5.0', '<' ) ) {
				wp_print_styles( 'arformspdfcreator-css' );
				deactivate_plugins( __FILE__, false, null );
				$redirect_url = network_admin_url( 'plugins.php?deactivate=true' );
				wp_die( '<div class="arf_pdf_wp_notice"><p class="arf_pdf_wp_notice_text" >Please meet the minimum requirement of WordPress version 4.5 to activate ARForms - PDFCreator Add-on<p class="arf_pdf_wp_notice_continue">Please <a href="javascript:void(0)" onclick="window.location.href=\'' . $redirect_url . '\'">Click Here</a> to continue.</p></div>' );
			}
		}
	}

	function getwpversion() {
		global $arformspdfcreator_version, $arformspdfcreator;
		$bloginformation = array();
		$str             = $arformspdfcreator->get_rand_alphanumeric( 10 );

		$bloginformation[] = get_bloginfo( 'name' );
		$bloginformation[] = get_bloginfo( 'description' );
		$bloginformation[] = home_url();
		$bloginformation[] = get_bloginfo( 'admin_email' );
		$bloginformation[] = get_bloginfo( 'version' );
		$bloginformation[] = get_bloginfo( 'language' );
		$bloginformation[] = $arformspdfcreator_version;
		$bloginformation[] = $_SERVER['SERVER_ADDR'];
		$bloginformation[] = $str;

		update_option( 'wp_arforms_pdfcreator_version_get_version', $str );

		$valstring  = implode( '||', $bloginformation );
		$encodedval = base64_encode( $valstring );

		$urltopost = 'http://reputeinfosystems.net/arforms/wpinfo_pdfcreator.php';
		$response  = wp_remote_post(
			$urltopost,
			array(
				'method'      => 'POST',
				'timeout'     => 45,
				'redirection' => 5,
				'httpversion' => '1.0',
				'blocking'    => true,
				'headers'     => array(),
				'body'        => array( 'wpversion' => $encodedval ),
				'cookies'     => array(),
			)
		);
	}

	function assign_rand_value( $num ) {

		switch ( $num ) {
			case '1':
				$rand_value = 'a';
				break;
			case '2':
				$rand_value = 'b';
				break;
			case '3':
				$rand_value = 'c';
				break;
			case '4':
				$rand_value = 'd';
				break;
			case '5':
				$rand_value = 'e';
				break;
			case '6':
				$rand_value = 'f';
				break;
			case '7':
				$rand_value = 'g';
				break;
			case '8':
				$rand_value = 'h';
				break;
			case '9':
				$rand_value = 'i';
				break;
			case '10':
				$rand_value = 'j';
				break;
			case '11':
				$rand_value = 'k';
				break;
			case '12':
				$rand_value = 'l';
				break;
			case '13':
				$rand_value = 'm';
				break;
			case '14':
				$rand_value = 'n';
				break;
			case '15':
				$rand_value = 'o';
				break;
			case '16':
				$rand_value = 'p';
				break;
			case '17':
				$rand_value = 'q';
				break;
			case '18':
				$rand_value = 'r';
				break;
			case '19':
				$rand_value = 's';
				break;
			case '20':
				$rand_value = 't';
				break;
			case '21':
				$rand_value = 'u';
				break;
			case '22':
				$rand_value = 'v';
				break;
			case '23':
				$rand_value = 'w';
				break;
			case '24':
				$rand_value = 'x';
				break;
			case '25':
				$rand_value = 'y';
				break;
			case '26':
				$rand_value = 'z';
				break;
			case '27':
				$rand_value = '0';
				break;
			case '28':
				$rand_value = '1';
				break;
			case '29':
				$rand_value = '2';
				break;
			case '30':
				$rand_value = '3';
				break;
			case '31':
				$rand_value = '4';
				break;
			case '32':
				$rand_value = '5';
				break;
			case '33':
				$rand_value = '6';
				break;
			case '34':
				$rand_value = '7';
				break;
			case '35':
				$rand_value = '8';
				break;
			case '36':
				$rand_value = '9';
				break;
		}
		return $rand_value;
	}

	function get_rand_alphanumeric( $length ) {
		if ( $length > 0 ) {
			global $arformspdfcreator;
			$rand_id = '';
			for ( $i = 1; $i <= $length; $i++ ) {
				mt_srand( (float) microtime() * 1000000 );
				$num      = mt_rand( 1, 36 );
				$rand_id .= $arformspdfcreator->assign_rand_value( $num );
			}
		}
		return $rand_id;
	}

	function arformspdfcreator_backup() {
		$databaseversion = get_option( 'arformspdfcreator_version' );
		update_option( 'old_arformspdfcreator_version', $databaseversion );
	}

	function upgrade_arformspdfcreator_data() {

		global $newarformspdfcreator_version;

		if ( ! isset( $newarformspdfcreator_version ) || $newarformspdfcreator_version == '' ) {
			$newarformspdfcreator_version = get_option( 'arformspdfcreator_version' );
		}

		if ( version_compare( $newarformspdfcreator_version, '3.0', '<' ) ) {
			$path = ARF_PDFCREATOR_DIR . '/core/upgrade_arformspdfcreator_latest_data.php';
			include $path;
		}
	}

	function load_mpdf_library() {

		global $arformspdfcreator, $wpdb, $MdlDb;

		set_time_limit( 0 );
		if ( function_exists( 'wp_raise_memory_limit' ) ) {
			wp_raise_memory_limit( 'admin' );
		} else {
			ini_set( 'memory_limit', '256M' );
		}

		$tamplate_id = isset( $_REQUEST['arf_pdf_template'] ) ? $_REQUEST['arf_pdf_template'] : 0;
		$paper_size  = 'A4';
		$font_family = 'arf_arial';

		if ( get_option( 'arformspdfcreator_version' ) && get_option( 'arformspdfcreator_version' ) >= '2.0' ) {
			$template = $wpdb->get_row( $wpdb->prepare( 'SELECT template_settings FROM ' . $arformspdfcreator->db_pdf_creator . ' WHERE id=%d', $tamplate_id ) );

			if ( ! empty( $template ) && $template->template_settings != '' ) {
				$arr         = maybe_unserialize( $template->template_settings );
				$paper_size  = $arr['arf_pdf_template_size'];
				$font_family = strtolower( $arr['arf_pdf_template_font'] );
				$orientation = $arr['arf_pdf_template_orientation'];
				if ( $orientation == 'landscape' ) {
					$paper_size .= '-L';
				}
			}
		}

		$lib = ARF_PDFCREATOR_DIR . '/libs/mpdf.php';
		include $lib;

		$pdf_templates = ARF_PDFCREATOR_DIR . '/core/arf_pdfcreator_templates.php';
		include $pdf_templates;

		global $arf_templates;
		$arf_templates = new ARF_pdfcreator_template( $font_family );

		global $arf_pdfcreator;
		$arf_pdfcreator = new ARF_MPDF( 'utf-8', $paper_size, 0, $font_family, 5, 5, 20, 10, 5, 5 );

		$arf_pdfcreator->setAutoTopMargin    = 'stretch';
		$arf_pdfcreator->setAutoBottomMargin = 'stretch';

		if ( is_rtl() ) {
			$arf_pdfcreator->SetDirectionality( 'rtl' );
		}
	}

	function arf_create_pdf( $item_id, $form_id, $return = false ) {
		global $arf_templates, $is_version_compatible, $PDF_button;
		$template = $arf_templates->get_arfpdf_template( $form_id );

		if ( $template == '' ) {
			$template = 'default_template';
		}
		if ( $is_version_compatible ) {
			$PDF_button = "<div class='arfformicondiv arfhelptip' title='" . esc_html__( 'PDF', 'ARForms-Pdfcreator' ) . "'><a data-id='" . $item_id . "' href='" . ARF_PDFCREATOR_URL . '/core/arf_pdfcreator.php?entry_id=' . $item_id . '&arf_pdf_template=' . $template . "' target='_blank'><svg width='30px' height='30px' viewBox='-7 -6 32 32' class='arfsvgposition'><path fill-rule='evenodd'  clip-rule='evenodd' fill='#ffffff' d='M2.759,19.001c-0.479,0-0.928-0.189-1.301-0.563c-0.966-0.967-1.061-2.158-0.252-3.187c0.823-1.049,2.72-1.984,4.96-2.665  c0.253-0.553,0.499-1.135,0.73-1.736c0.266-0.688,0.502-1.371,0.708-2.036c-1.626-2.709-2.646-5.509-2.249-7.12  c0.287-1.162,1.208-1.776,2.528-1.686c0.006,0,0.011,0.001,0.017,0.002c0.665,0.058,1.839,0.434,2.222,2.29  c0.314,1.523,0.006,3.831-0.666,6.248c0.671,1.047,1.43,2.074,2.215,2.973c2.375-0.207,4.516-0.039,5.652,0.651  c0.973,0.592,1.27,1.571,0.792,2.62c-0.019,0.042-0.041,0.081-0.066,0.119c-0.834,1.251-2.212,1.573-3.78,0.882  c-1.041-0.458-2.168-1.351-3.262-2.485c-1.195,0.144-2.444,0.373-3.614,0.674c-1.114,2.268-2.406,4.097-3.583,4.731  C3.455,18.906,3.099,19.001,2.759,19.001z M5.063,14.725c-1.15,0.455-2.061,0.989-2.521,1.576c-0.245,0.311-0.289,0.53,0.118,0.938  c0.066,0.065,0.109,0.107,0.345-0.019C3.583,16.907,4.317,15.997,5.063,14.725z M13.292,13.145c0.569,0.482,1.131,0.861,1.661,1.095  c1.075,0.475,1.458,0.057,1.643-0.209c0.094-0.228,0.051-0.28-0.156-0.406C15.852,13.266,14.699,13.12,13.292,13.145z M8.797,10.631  c-0.168,0.477-0.348,0.948-0.537,1.413c0.464-0.101,0.934-0.19,1.403-0.269C9.366,11.406,9.077,11.023,8.797,10.631z M7.62,1.697  c-0.516,0-0.571,0.223-0.616,0.403C6.788,2.979,7.307,4.614,8.227,6.416c0.322-1.561,0.414-2.893,0.232-3.772S7.92,1.717,7.76,1.703  C7.71,1.699,7.664,1.697,7.62,1.697z' /></svg></a></div>";
		} else {
			$PDF_button = "<a href='" . ARF_PDFCREATOR_URL . '/core/arf_pdfcreator.php?entry_id=' . $item_id . '&arf_pdf_template=' . $template . "' target='_blank'><img src='" . ARF_PDFCREATOR_URL . "/images/pdf-icon.png' class='arfhelptip' title='" . esc_html__( 'PDF', 'ARForms-Pdfcreator' ) . "' onmouseout='this.src=\"" . ARF_PDFCREATOR_URL . "/images/pdf-icon.png\";' onmouseover='this.src=\"" . ARF_PDFCREATOR_URL . "/images/pdf-icon_hover.png\";' /></a>";
		}

		if ( $return ) {
			return $PDF_button;
		} else {
			echo $PDF_button;
		}
	}

	function get_file_icon( $media_id ) {
		global $arfieldhelper;

		if ( ! is_numeric( $media_id ) ) {
			return;
		}

		$post_meta_data = get_post_meta( $media_id );

		$image_array_link = $post_meta_data['_wp_attached_file'][0];

		$image_name = explode( '/', $image_array_link );

		$image_name = $image_name[ count( $image_name ) - 1 ];

		$image_ext = explode( '.', $image_name );

		$image_ext = $image_ext[ count( $image_ext ) - 1 ];

		$image_ext = strtolower( $image_ext );

		$exclude_ext = array( 'png', 'jpg', 'jpeg', 'jpe', 'gif', 'bmp', 'tif', 'tiff', 'ico' );
		
		if ( in_array( $image_ext, $exclude_ext ) ) {
			$image_array_link = get_home_url() . '/' . str_replace( 'thumbs/', '', $image_array_link );
			
			$img_height       = 'height=150';
			$img_width        = 'width=150';
		} else {
			$img_height       = '';
			$img_width        = '';
			$image_array_link = $arfieldhelper->arf_get_file_icon( $image_ext );
		}

		$image = '<img class="attachment-thumnail" alt="' . $image_array_link . '" src="' . $image_array_link . '" border="0" ' . $img_height . ' ' . $img_width . '>';

		return $image;
	}

	function get_file_link( $media_id, $short = true ) {
		if ( is_numeric( $media_id ) ) {

			if ( $short ) {

				$attachment = get_post( $media_id );

				$label = basename( $attachment->guid );
			}

			global $arfieldhelper;

			$post_meta_data = get_post_meta( $media_id );

			$image_array_link = $post_meta_data['_wp_attached_file'][0];

			$image_name = explode( '/', $image_array_link );

			$image_name = $image_name[ count( $image_name ) - 1 ];

			$image_ext = explode( '.', $image_name );

			$image_ext = $image_ext[ count( $image_ext ) - 1 ];

			$image_ext = strtolower( $image_ext );

			$exclude_ext = array( 'png', 'jpg', 'jpeg', 'jpe', 'gif', 'bmp', 'tif', 'tiff', 'ico' );

			$is_image = false;
			if ( in_array( $image_ext, $exclude_ext ) ) {
				$url      = get_home_url() . '/' . str_replace( 'thumbs/', '', $image_array_link );
				$is_image = true;
			} else {
				$url = get_home_url() . '/' . $image_array_link;
			}

			if ( is_admin() ) {
				global $arfsiteurl;

				$url = '<a href="' . $url . '">';
			}

			return $url;
		}
	}

	function arfpdfcreator_setting_v3( $id, $values ) {
		global $arf_templates, $is_version_compatible;

		$templates = $arf_templates->get_arf_pdf_templates();
		?>
		<div class="arf_clear_both"></div>
			<div class="arf_separater arf_pdc_separator"></div>
			<div class="arftablerow arfpdfcreatorsetting">
				<div class="arfcolmnleft">

					<div class="arftablerow" >
						<div class="arfcolumnleft arfsettingsubtitle">
		 <?php esc_html_e( 'PDF Creator Settings', 'ARForms-Pdfcreator' ); ?>
						</div>
		<?php if ( $is_version_compatible ) { ?>
							<div class="arf_popup_checkbox_wrapper arf_pdfcreator_checkbox_wrapper">
								<div class="arf_custom_checkbox_div">
									<div class="arf_custom_checkbox_wrapper">
										<input type="checkbox" value="1" id="send_pdf_with_email" name="options[send_pdf_with_email]" <?php checked( ( isset( $values['send_pdf_with_mail'] ) ? $values['send_pdf_with_mail'] : '' ), 1 ); ?>>
										<svg width="18px" height="18px">
			<?php echo ARF_CUSTOM_UNCHECKED_ICON; ?>
			<?php echo ARF_CUSTOM_CHECKED_ICON; ?>
										</svg>
									</div>
								</div>
								<span><label class="send_pdf_with_email" for="send_pdf_with_email">
			<?php esc_html_e( ' Send pdf file of the form entry with email to admin ( If Configured )', 'ARForms-Pdfcreator' ); ?>
									</label></span>
							</div>
							<div class="arf_popup_checkbox_wrapper arf_pdfcreator_checkbox_wrapper">
								<div class="arf_custom_checkbox_div">
									<div class="arf_custom_checkbox_wrapper">
										<input type="checkbox" value="1" id="send_pdf_with_email_touser" name="options[send_pdf_with_email_touser]" <?php checked( isset( $values['send_pdf_with_email_touser'] ) ? $values['send_pdf_with_email_touser'] : '', 1 ); ?>>
										<svg width="18px" height="18px">
			<?php echo ARF_CUSTOM_UNCHECKED_ICON; ?>
			<?php echo ARF_CUSTOM_CHECKED_ICON; ?>
										</svg>
									</div>
								</div>
								<span> <label class="send_pdf_with_email" for="send_pdf_with_email_touser"><span></span>
			<?php esc_html_e( 'Send pdf file of the form entry with email to user ( If Configured )', 'ARForms-Pdfcreator' ); ?>
									</label></span>
							</div>

		<?php } else { ?>
							<div class="arfcolumnright">
								<input type="checkbox" value="1" id="send_pdf_with_email" name="options[send_pdf_with_email]" <?php checked( isset( $values['send_pdf_with_mail'] ) ? $values['send_pdf_with_mail'] : '', 1 ); ?> class="chkstanard">
								<label for="send_pdf_with_email" class="send_pdf_with_email"><span></span>
			<?php esc_html_e( 'Send pdf file of the form entry with email to admin ( If Configured )', 'ARForms-Pdfcreator' ); ?>
								</label>
							</div>
						</div> 

						<div class="arftablerow arf_pdfcreate_send_pdf_user"> 
							<div class="arfcolumnright">
								<input type="checkbox" value="1" id="send_pdf_with_email_touser" name="options[send_pdf_with_email_touser]" <?php checked( isset( $values['send_pdf_with_email_touser'] ) ? $values['send_pdf_with_email_touser'] : '', 1 ); ?> class="chkstanard">
								<label for="send_pdf_with_email_touser" class="send_pdf_with_email"><span></span>
			<?php esc_html_e( 'Send pdf file of the form entry with email to user ( If Configured )', 'ARForms-Pdfcreator' ); ?>
								</label>
							</div>
						</div>
		<?php } ?>

					<div class="arftablerow" >
						<div class="arfcolumnleft">&nbsp;</div>
						<div class="arfcolmnright">
							<div class="arfsettingsubtitle">
								<?php esc_html_e( 'PDF Template', 'ARForms-Pdfcreator' ); ?>
							</div>
							<div class="arf_clear_both"></div>
							<div class="sltstandard arf_pdfcreator_sltstandard">
								<?php

								global $arformspdfcreator, $wpdb, $MdlDb;
								$template_selected = '';
								$record            = '';
								$form_data         = array();
								$template_id       = 0;

								if ( $id != '' && $id != 0 ) {
									$record = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM `' . $MdlDb->forms . '` WHERE id = %d', $id ) );

									if ( ! empty( $record ) && isset( $record->options ) ) {
										$form_data   = maybe_unserialize( $record->options );
										$template_id = $form_data['arf_pdf_template'];
									}
								}

								$templates       = $wpdb->get_results( 'SELECT * FROM ' . $arformspdfcreator->db_pdf_creator, ARRAY_A );
								$def_val         = '';
								$def_template_id = '';
								$html            = '';

								if ( ! empty( $templates ) ) {
									$counter = 0;
									foreach ( $templates as $template ) {

										if ( $counter == 0 ) {
											$def_val         = $template['template_name'];
											$def_template_id = $template['id'];
										}
										if ( $template_id == $template['id'] ) {
											$template_selected = $template['template_name'];
											$template_id       = $template['id'];
										}

										$html .= "<li class='arf_selectbox_option' data-value='" . $template['id'] . "' data-label='" . $template['template_name'] . "'> " . $template['template_name'] . ' </li>';
										$counter++;
									}

									$current_template_id   = ( $template_id != 0 ) ? $template_id : $def_template_id;
									$current_template_name = ( $template_selected != '' ) ? $template_selected : $def_val;
								}

								?>
								<input id="arf_pdf_template" name="options[arf_pdf_template]"  value="<?php echo esc_html( $current_template_id ); ?>" type="hidden">
								<dl class="arf_selectbox arf_pdfc_selectbox" name="options[arf_pdf_template]">
									<dt class=""><span>
										<?php echo $current_template_name; ?>
									</span>

									<i class="arfa arfa-caret-down arfa-lg"></i></dt>
									<dd>
										<ul class="arf_selectbox dt" data-id="arf_pdf_template">
										<?php echo $html; ?>
										</ul>
									</dd>
								</dl>
							</div>
						</div>
						<div class="arfcolumnleft">&nbsp;</div>
		 <?php
			$template_name = '';
			if ( ! empty( $form_data ) && isset( $form_data['arf_pdfcreator_template_name'] ) ) {
				$template_name = $form_data['arf_pdfcreator_template_name'];
			}
			if ( $template_name == '' ) {
				$template_name = 'entry_{entry_id}';
			}
			?>
						<div class="arfcolmnright arf_postcreator_field_wrapper">
							<div class="arfsettingsubtitle"> <?php esc_html_e( 'Template Name', 'ARForms-Pdfcreator' ); ?> </div>
							<div class="arf_clear_both"></div>
							<input type="text" name="options[arf_pdfcreator_template_name]" id="arf_pdfcreator_template_name" value="<?php echo $template_name; ?>" class="arf_advanceemailfield arfheight34" />
							<button type="button" class="arf_add_field_button arf_pdfcreator_add_field_button" onclick="add_field_fun('add_field_template_name')" id="add_field_template_name_but"><?php echo addslashes( esc_html__( 'Add Field', 'ARForms-Pdfcreator' ) ); ?>&nbsp;&nbsp;
								<img src="<?php echo esc_url( ARFIMAGESURL . '/down-arrow.png' ); ?>" align="absmiddle" />
							</button>
							<div class="main_field_modal">
								<div class="arf_add_fieldmodal arf_pdf_field_template_name" id="add_field_template_name">
									<div class="arf_modal_header">
										<div class="arf_add_field_title">
										<?php echo addslashes( esc_html__( 'Fields', 'ARForms-Pdfcreator' ) ); ?>
											<div data-dismiss="arfmodal" onclick="close_add_field_subject('add_field_template_name')" class="arf_field_model_close">
												<svg viewBox="0 -4 32 32"><g id="email"><path fill-rule="evenodd" clip-rule="evenodd" fill="#333333" d="M10.702,10.909L6.453,6.66l-4.249,4.249L1.143,9.848l4.249-4.249L1.154,1.361l1.062-1.061l4.237,4.237l4.238-4.237l1.061,1.061L7.513,5.599l4.249,4.249L10.702,10.909z"></path></g></svg>
											</div>
										</div>
									</div>
									<div class="arfmodal-body_p">
										<div class="modal_field_val" id="arfmodalfieldval_extrafield" onclick="arfaddcodeforextrafield('arf_pdfcreator_template_name', 'date')">Date</div>
										<div class="modal_field_val" id="arfmodalfieldval_extrafield" onclick="arfaddcodeforextrafield('arf_pdfcreator_template_name', 'time')">Time</div>
										<div class="modal_field_val" id="arfmodalfieldval_extrafield" onclick="arfaddcodeforextrafield('arf_pdfcreator_template_name', 'entry_id')">Entry ID</div>
										<?php
										if ( isset( $values['id'] ) ) {
											global $wpdb, $MdlDb, $arfieldhelper;
											$field_list        = array();
											$exclude           = array( 'divider', 'section', 'arf_repeater', 'captcha', 'break' );
											$all_hidden_fields = array();
											$all_fields        = $wpdb->get_results( $wpdb->prepare( 'SELECT * FROM `' . $MdlDb->fields . '` WHERE form_id = %d ORDER BY ID ASC', $values['id'] ) );
											
											if ( ! empty( $all_fields ) ) {
												foreach ( $all_fields as $key => $field_ ) {
													if ( ! in_array( $field_->id, $exclude ) && $field_->type == 'hidden' ) {
														$all_hidden_fields[] = $field_;
														$include_fields[]    = $field_->id;
														continue;
													}
													foreach ( $field_ as $k => $field_val ) {
														if ( $k == 'type' && ! in_array( $field_val, $exclude ) ) {
															$include_fields[] = $field_->id;
														}
														if ( $k == 'options' ) {
															$arf_all_fields[ $key ][ $k ] = json_decode( $field_val, true );
															if ( json_last_error() != JSON_ERROR_NONE ) {
																$arf_all_fields[ $key ][ $k ] = maybe_unserialize( $field_val );
															}
														} elseif ( $k == 'field_options' ) {
															$field_opts = json_decode( $field_val, true );
															if ( json_last_error() != JSON_ERROR_NONE ) {
																$field_opts = maybe_unserialize( $field_val );
															}
															if ( isset( $field_opts ) && is_array( $field_opts ) ) {
																foreach ( $field_opts as $ki => $val_ ) {
																					  $arf_all_fields[ $key ][ $ki ] = $val_;
																}
															}
														} else {
																  $arf_all_fields[ $key ][ $k ] = $field_val;
														}
													}
												}
												foreach ( $all_fields as $key => $field_ ) {
													foreach ( $field_ as $k => $field_val ) {
														if ( in_array( $field_->id, $include_fields ) ) {
															if ( ! isset( $field_list[ $key ] ) ) {
																$field_list[ $key ] = new stdClass();
															}
															if ( $k == 'options' ) {
																							   $fOpt = json_decode( $field_val, true );
																if ( json_last_error() != JSON_ERROR_NONE ) {
																	$fOpt = maybe_unserialize( $field_val );
																}
																$field_list[ $key ]->$k = $fOpt;
															} elseif ( $k == 'field_options' ) {
																							  $field_opts = json_decode( $field_val, true );
																if ( json_last_error() != JSON_ERROR_NONE ) {
																	$field_opts = maybe_unserialize( $field_val );
																}
																$field_list[ $key ]->$k = $field_opts;
															} else {
																							 $field_list[ $key ]->$k = $field_val;
															}
														}
													}
												}
											}
											$arfieldhelper->get_shortcode_modal( $values['id'], 'arf_pdfcreator_template_name', 'no_email', '', false, $field_list );
										}
										?>

									</div>
								</div>
							</div>
						</div>
					</div>
					<div class="arfcolumnleft">&nbsp;</div>
					<div class="pdfcreator_documentation"><a href="<?php echo esc_url( ARF_PDFCREATOR_URL . '/documentation/index.html' ); ?>" target="_blank" class="arlinks"><?php esc_html_e( 'PDF Creator Documentation', 'ARForms-Pdfcreator' ); ?></a></div>
				</div>
			</div>
		</div>
		<?php
	}

	function arf_pdfcreator_save_opt( $options, $values ) {
		$options['send_pdf_with_mail']           = isset( $values['options']['send_pdf_with_email'] ) ? $values['options']['send_pdf_with_email'] : '';
		$options['send_pdf_with_email_touser']   = isset( $values['options']['send_pdf_with_email_touser'] ) ? $values['options']['send_pdf_with_email_touser'] : '';
		$options['arf_pdf_template']             = $values['options']['arf_pdf_template'];
		$options['arf_pdfcreator_template_name'] = isset( $values['options']['arf_pdfcreator_template_name'] ) ? $values['options']['arf_pdfcreator_template_name'] : '';

		return $options;
	}

	function arf_pdfcreator_send_pdf_with_mail( $entry_id, $form_id ) {
		global $wpdb, $db_record, $arfsettings, $armainhelper, $arfieldhelper, $arnotifymodel, $MdlDb;

		$wp_upload_dir = wp_upload_dir();

		$upload_main_dir = $wp_upload_dir['basedir'] . '/arforms/';

		$pdf = $this->arfcreatepdf( $upload_main_dir, $entry_id );

		$attachment[] = $upload_main_dir . '/' . $pdf;

		$frm_opt = $wpdb->get_results( $wpdb->prepare( 'SELECT options FROM ' . $MdlDb->forms . ' WHERE id = %d', $form_id ) );

		$options = maybe_unserialize( $frm_opt->options );

		$entry = $db_record->getOne( $entry_id, true );

		if ( $options['chk_admin_notification'] == 0 && $options['send_pdf_with_mail'] == 0 ) {
			return;
		}

		$admin_email = $options['reply_to'];

		$from_email = $options['ar_admin_from_email'];

		$subject = $options['admin_email_subject'];

		$reply_to_name = ( isset( $options['ar_user_from_name'] ) ) ? $options['ar_user_from_name'] : $arfsettings->reply_to_name;

		$reply_to_id = ( isset( $options['ar_user_from_email'] ) ) ? $options['ar_user_from_email'] : $arfsettings->reply_to;

		if ( isset( $reply_to_id ) ) {
			$reply_to = $entry->metas[ $reply_to_id ];
		}

		if ( $reply_to == '' ) {
			$reply_to = $reply_to_id;
		}

		$subject = str_replace( '[form_name]', stripslashes( $form->name ), $subject );

		$subject = str_replace( '[site_name]', $arfblogname, $subject );

		$shortcodes = $armainhelper->get_shortcodes( $options['ar_admin_from_email'], $entry->form_id );

		$reply_to = $arfieldhelper->replace_shortcodes( $options['ar_admin_from_email'], $entry, $shortcodes );

		$reply_to = $arfieldhelper->arf_replace_shortcodes( $reply_to, $entry );

		$arnotifymodel->send_notification_email_user( $admin_email, $subject, $message = '', $reply_to, $reply_to_name, true, $attachment );
	}

	function arf_pdfcreator_sendpdf_with_mailattachment( $attachments, $form, $entries = array() ) {

		if ( ! $form ) {
			return $attachments;
		}

		if ( ! $entries ) {
			return $attachments;
		}

		global $wpdb, $db_record, $arfsettings, $arformspdfcreator, $MdlDb;

		$frm_opt = $wpdb->get_row( $wpdb->prepare( 'SELECT options FROM ' . $MdlDb->forms . ' WHERE id = %d', $form->id ) );

		$options = maybe_unserialize( $frm_opt->options );

		if ( $options['send_pdf_with_mail'] == 0 ) {
			return $attachments;
		}

		$entry = $entries['entry'];

		$entry_id = $entry->id;

		if ( ! $entry_id ) {
			return $attachments;
		}

		$wp_upload_dir = wp_upload_dir();

		$upload_main_dir = $wp_upload_dir['basedir'] . '/arforms/';

		$pdf = $arformspdfcreator->arfcreatepdf( $upload_main_dir, $entry_id );

		$attachments[] = $upload_main_dir . $pdf;

		return $attachments;
	}

	function arf_pdfcreator_sendpdf_with_mailattachment_touser( $attachments, $form, $entries = array() ) {

		if ( ! $form ) {
			return $attachments;
		}

		if ( ! $entries ) {
			return $attachments;
		}

		global $wpdb, $db_record, $arfsettings, $arformspdfcreator, $MdlDb;

		$frm_opt = $wpdb->get_row( $wpdb->prepare( 'SELECT options FROM ' . $MdlDb->forms . ' WHERE id = %d', $form->id ) );

		$options = maybe_unserialize( $frm_opt->options );

		if ( $options['send_pdf_with_email_touser'] == 0 ) {
			return $attachments;
		}

		$entry = $entries['entry'];

		$entry_id = $entry->id;

		if ( ! $entry_id ) {
			return $attachments;
		}

		$str    = $options['arf_pdfcreator_template_name'];
		$string = '';

		$pattern = '/\[(.*?)\:(\d+)(|\.(\d+))\]/';

		preg_match_all( $pattern, $str, $matched );

		if ( isset( $matched[0] ) && $matched[0] != '' ) {
			foreach ( $matched[0] as $vi ) {
				$exp    = explode( ':', $vi );
				$tmp_id = str_replace( ']', '', $exp[1] );
				$fields = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $MdlDb->entry_metas . ' WHERE entry_id=%d AND field_id=%d', $id, $tmp_id ) );
				if ( is_serialized( $fields->entry_value ) ) {
					$field_value = maybe_unserialize( $fields->entry_value );
					$field_value = implode( '_', $field_value );
					$str         = str_replace( $vi, $field_value, $str );
				} else {
					$str = str_replace( $vi, $fields->entry_value, $str );
				}
			}
		}

		$pattern2 = '/\{(.*?)\}/';
		preg_match_all( $pattern2, $str, $matched );

		if ( isset( $matched[0] ) && $matched[0] != '' ) {
			foreach ( $matched[0] as $vi ) {
				$tmp_val = str_replace( '{', '', $vi );
				$tmp_val = str_replace( '}', '', $tmp_val );
				$row     = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $MdlDb->entries . ' WHERE id=%d', $id ) );
				if ( ! empty( $row ) ) {
					if ( $tmp_val == 'date' ) {
						$date = date( 'Y-m-d', strtotime( $row->created_date ) );
						$str  = str_replace( $vi, $date, $str );
					}
					if ( $tmp_val == 'time' ) {
						$time = date( 'h_i_s', strtotime( $row->created_date ) );
						$str  = str_replace( $vi, $time, $str );
					}
					if ( $tmp_val == 'entry_id' ) {
						   $str = str_replace( $vi, $row->id, $str );
					}
				}
			}
		}

		$pdf_file_name = ( $str != '' ) ? $str : 'entry_' . $entry_id;

		$wp_upload_dir = wp_upload_dir();

		$upload_main_dir = $wp_upload_dir['basedir'] . '/arforms/';

		if ( ! file_exists( $upload_main_dir . $pdf_file_name . '.pdf' ) ) {
			$pdf = $arformspdfcreator->arfcreatepdf( $upload_main_dir, $entry_id );

			$attachments[] = $upload_main_dir . $pdf;
		} else {
			$attachments[] = $upload_main_dir . $pdf_file_name . '.pdf';
		}

		return $attachments;
	}

	function arfcreatepdf( $path, $id ) {

		global $arf_pdfcreator, $wpdb, $arformspdfcreator, $arf_templates, $MdlDb;
		$style       = file_get_contents( ARF_PDFCREATOR_DIR . '/css/arformspdfcreator.css' );
		$paper_size  = 'A4';
		$orientation = 'portrait';
		$font_family = '';

		$form         = $wpdb->get_results( $wpdb->prepare( 'SELECT form_id,created_date FROM ' . $MdlDb->entries . ' WHERE id = %d', $id ) );
		$form_id      = $form[0]->form_id;
		$created_date = strtotime( $form[0]->created_date );
		$forms        = $wpdb->get_results( $wpdb->prepare( 'SELECT name,options FROM ' . $MdlDb->forms . ' WHERE id = %d', $form_id ) );
		$form_name    = stripslashes_deep( strip_tags( $forms[0]->name ) );
		$frm_options  = maybe_unserialize( $forms[0]->options );

		$tpl_id   = $frm_options['arf_pdf_template'];
        if( 'string' == gettype($tpl_id) && ! preg_match('/^[\d+]+$/', $tpl_id, $match) ){
            $tpl_id = 1;
        }
		$pdf_tbl  = $arformspdfcreator->db_pdf_creator;
		$template = $wpdb->get_row( $wpdb->prepare( 'SELECT template_settings FROM ' . $pdf_tbl . ' WHERE id=%d', $tpl_id ) );
		if ( ! empty( $template ) ) {
			$tpl_option  = maybe_unserialize( $template->template_settings );
			$paper_size  = $tpl_option['arf_pdf_template_size'];
			$orientation = $tpl_option['arf_pdf_template_orientation'];
			$font_family = strtolower( $tpl_option['arf_pdf_template_font'] );
			if ( $orientation == 'landscape' ) {
				$paper_size .= '-L';
			}
		}
		$arf_pdfcreator = new ARF_MPDF( 'utf-8', $paper_size, 0, $font_family, 5, 5, 20, 10, 5, 5 );

		$arf_pdfcreator->setAutoTopMargin    = 'stretch';
		$arf_pdfcreator->setAutoBottomMargin = 'stretch';

        $like_btn_color = '';
        $dislike_btn_color = '';

        if( isset( $id ) && '' != $id ){
            $entry_id = $id;
            $form_id = $wpdb->get_var( $wpdb->prepare( "SELECT form_id FROM `" . $MdlDb->entries . "` WHERE id = %d", $entry_id ) );
            if( '' != $form_id ){
                $has_like_field = $wpdb->get_var( $wpdb->prepare( "SELECT count(*) FROM `". $MdlDb->fields."` WHERE type = %s AND form_id = %d", 'like', $form_id) );
                if( $has_like_field > 0 ){
                    $frm_css = $wpdb->get_var( $wpdb->prepare ( "SELECT form_css FROm `". $MdlDb->forms ."` WHERE id = %d", $form_id) );
                    $form_css = maybe_unserialize( $frm_css );
                    
                    $like_btn_color = $form_css['arflikebtncolor'];
                    $dislike_btn_color = $form_css['arfdislikebtncolor'];
                }
            }
        }
        if( '' != $like_btn_color && '' != $dislike_btn_color ){            
            global $arformspdfcreator_version;
            
            $runtime_css  = '.arf_like_color_custom{margin-left:0;background:'.$like_btn_color.';display:inline-block;padding:10px;}.imgmargin{padding:6px;}';
            $runtime_css .= '.arf_dislike_color_custom{margin-left:0;background:'.$dislike_btn_color.';display:inline-block;padding:10px;}.imgmargin{padding:6px;}';
            $style .= $runtime_css;
        }

		$arf_pdfcreator->SetFColor( 'red', true );
		$arf_pdfcreator->WriteHTML( $style, 1 );
		$arf_pdfcreator->autoLangToFont = true;

		$arftemplatename = $frm_options['arf_pdf_template'];

		$str    = $frm_options['arf_pdfcreator_template_name'];
		$string = '';

		$pattern = '/\[(.*?)\:(\d+)(|\.(\d+))\]/';

		preg_match_all( $pattern, $str, $matched );

		if ( isset( $matched[0] ) && $matched[0] != '' ) {
			foreach ( $matched[0] as $vi ) {
				$exp    = explode( ':', $vi );
				$tmp_id = str_replace( ']', '', $exp[1] );
				$fields = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $MdlDb->entry_metas . ' WHERE entry_id=%d AND field_id=%d', $id, $tmp_id ) );
				if ( is_serialized( $fields->entry_value ) ) {
					$field_value = maybe_unserialize( $fields->entry_value );
					$field_value = implode( '_', $field_value );
					$str         = str_replace( $vi, $field_value, $str );
				} else {
					$str = str_replace( $vi, $fields->entry_value, $str );
				}
			}
		}

		$pattern2 = '/\{(.*?)\}/';
		preg_match_all( $pattern2, $str, $matched );

		if ( isset( $matched[0] ) && $matched[0] != '' ) {
			foreach ( $matched[0] as $vi ) {
				$tmp_val = str_replace( '{', '', $vi );
				$tmp_val = str_replace( '}', '', $tmp_val );
				$row     = $wpdb->get_row( $wpdb->prepare( 'SELECT * FROM ' . $MdlDb->entries . ' WHERE id=%d', $id ) );
				if ( ! empty( $row ) ) {
					if ( $tmp_val == 'date' ) {
						$date = date( 'Y-m-d', strtotime( $row->created_date ) );
						$str  = str_replace( $vi, $date, $str );
					}
					if ( $tmp_val == 'time' ) {
						$time = date( 'h_i_s', strtotime( $row->created_date ) );
						$str  = str_replace( $vi, $time, $str );
					}
					if ( $tmp_val == 'entry_id' ) {
						   $str = str_replace( $vi, $row->id, $str );
					}
				}
			}
		}

		$template_name = ( $str != '' ) ? $str : 'entry_' . $id;
		$arf_pdfcreator->SetTitle( $form_name );
		$arf_pdfcreator->autoLangToFont = true;
		
		$arf_templates = new ARF_pdfcreator_template( $font_family );

		$htmlcontent                    = $arf_templates->arf_get_template_html( $arftemplatename, $arf_pdfcreator, $form_id, $id );
        
		$arf_pdfcreator->WriteHTML( $htmlcontent );

		$arf_pdfcreator->Output( $path . $template_name . '.pdf', 'F' );

		return $template_name . '.pdf';
	}

	function get_label( $value ) {
		if ( ! $value ) {
			return '';
		}

		return $value->name;
	}

	function get_fieldtype( $value ) {
		if ( ! $value ) {
			return '';
		}

		return $value->type;
	}

	function get_value( $field, $entry_id, $is_conditoinal = '' ) {

		if ( ! $field ) {
			return '';
		}

		global $arformspdfcreator, $wpdb, $MdlDb, $is_version_compatible, $arfieldhelper, $arformcontroller;

		apply_filters( 'arfpdfdisplayvalue', $field );

		$formid = $wpdb->get_results( $wpdb->prepare( 'SELECT form_id FROM ' . $MdlDb->fields . ' WHERE id = %d', $field->id ) );
		$formid = isset( $formid[0]->form_id ) ? $formid[0]->form_id : '';

		if ( $formid ) {

			$forms    = $wpdb->get_results( $wpdb->prepare( 'SELECT form_css FROM ' . $MdlDb->forms . ' WHERE id = %d', $formid ) );
			$form_css = maybe_unserialize( $forms[0]->form_css );
		}

		$field->entry_value = isset( $field->entry_value ) ? $field->entry_value : '';

		$likebtn_bg_col    = isset( $form_css['arflikebtncolor'] ) ? $form_css['arflikebtncolor'] : '';
		$dislikebtn_bg_col = isset( $form_css['arfdislikebtncolor'] ) ? $form_css['arfdislikebtncolor'] : '';

		switch ( $field->type ) {
			case 'file':
				$return_file_upload_value = '';
				$post_id                  = $field->entry_value;
				if ( ! empty( $post_id ) ) {
					$post_id = ( ! empty( $post_id ) ) ? explode( '|', $post_id ) : '';

					if ( is_array( $post_id ) ) {
						foreach ( $post_id as $key => $value ) {
							$icon                      = $arformspdfcreator->get_file_icon( $value );
							$link                      = $arformspdfcreator->get_file_link( $value );
							$return_file_upload_value .= "<a target='_blank' href='" . $link . "'>" . $icon . '</a>';
						}
					}
				}
				return $return_file_upload_value;
				break;

			case 'like':
				if ( $is_version_compatible ) {
                    
					if ( $field->entry_value == '1' ) {
                        
                        return "<span style='margin-left:0;' class='arf_like_btn active field_in_list arf_like_color_custom'><img src='".ARFIMAGESURL."/like-icon.png' alt='like' class='imgmargin' /></span>";
					} elseif ( $field->entry_value == '0' ) {
						
                        return "<span style='margin-left:0' class='arf_dislike_btn active arf_dislike_color_custom field_in_list'><img src='" . ARFIMAGESURL ."/dislike-icon.png' alt='dislike' class='imgmargin' />";
					} else {
						return '-';
					}
				}
				break;
			case 'checkbox':
				$opts = maybe_unserialize( $field->entry_value );
				if ( is_array( $opts ) && count( $opts ) > 1 ) {
					$options = implode( ', ', $opts );
				} else {
					$options = $field->entry_value;
				}

				$field_opts       = $wpdb->get_row( $wpdb->prepare( 'SELECT entry_value FROM ' . $MdlDb->entry_metas . " WHERE field_id='%d' AND entry_id='%d'", '-' . $field->id, $entry_id ) );
				$field_opts_label = $wpdb->get_row( $wpdb->prepare( 'SELECT field_options, options FROM ' . $MdlDb->fields . " WHERE 	id='%d'", $field->id ) );

				if ( ! empty( $field_opts ) && ! empty( $field_opts_label ) ) {
					$json_field_options_arr = json_decode( $field_opts_label->field_options );
					$json_options_arr       = json_decode( $field_opts_label->options );

					$field_opts = maybe_unserialize( $field_opts->entry_value );

					if ( $field->type == 'checkbox' ) {

						if ( $field_opts && is_array( $field_opts ) && ! empty( $field_opts ) ) {

							$temp_value = '';
							if ( isset( $json_field_options_arr->separate_value ) && $json_field_options_arr->separate_value == 1 && ! empty( $json_options_arr ) && $is_conditoinal == '' ) {

								foreach ( $json_options_arr as $key => $new_field_opt ) {
									$search_array = $arformcontroller->arfSearchArray( $new_field_opt->value, 'value', $field_opts );

									if ( $search_array != '' || $search_array == '0' ) {
										$temp_value .= $new_field_opt->value . ' (' . $new_field_opt->label . '), ';
									}
								}
							} else {
								foreach ( $field_opts as $key => $new_field_opt ) {
									  $temp_value .= $new_field_opt . ', ';
								}
							}

							$temp_value = ( ! empty( $temp_value ) ) ? trim( $temp_value ) : '';
							$options    = rtrim( $temp_value, ',' );
						} else {
							$temp_value = '';
							if ( isset( $json_field_options_arr->separate_value ) && $json_field_options_arr->separate_value == 1 && ! empty( $json_options_arr ) && $is_conditoinal == '' ) {

								foreach ( $json_options_arr as $key => $new_field_opt ) {

									if ( $new_field_opt->value == $field_opts ) {
											$temp_value .= $new_field_opt->value . ' (' . $new_field_opt->label . ')';
									}
								}
							} else {

								  $temp_value .= $field_opts;
							}
							$options = $temp_value;
						}
					} else {
						$options = $field_opts;
					}
				}
				return $options;
				break;

			case 'radio':
				$options = $field->entry_value;

				$field_opts = $wpdb->get_row( $wpdb->prepare( 'SELECT entry_value FROM ' . $MdlDb->entry_metas . " WHERE 	field_id='%d' AND entry_id='%d'", '-' . $field->id, $entry_id ) );

				$field_opts_label = $wpdb->get_row( $wpdb->prepare( 'SELECT field_options, options FROM ' . $MdlDb->fields . " WHERE 	id='%d'", $field->id ) );

				if ( ! empty( $field_opts ) && ! empty( $field_opts_label ) ) {
					$json_field_options_arr = json_decode( $field_opts_label->field_options );
					$json_options_arr       = json_decode( $field_opts_label->options );

					if ( isset( $json_field_options_arr->separate_value ) && $json_field_options_arr->separate_value == 1 && ! empty( $json_options_arr ) && $is_conditoinal == '' ) {

						foreach ( $json_options_arr as $key => $opt_val ) {

							if ( $opt_val->value == $options ) {
								$options = $opt_val->value . ' (' . $opt_val->label . ')';
							}
						}
					} else {
						$options = $field_opts->entry_value;
					}
				}

				return $options;
				break;

			case 'select':
				$options = $field->entry_value;

				$field_opts       = $wpdb->get_row( $wpdb->prepare( 'SELECT entry_value FROM ' . $MdlDb->entry_metas . " WHERE field_id='%d' AND entry_id='%d'", '-' . $field->id, $entry_id ) );
				$field_opts_label = $wpdb->get_row( $wpdb->prepare( 'SELECT field_options, options FROM ' . $MdlDb->fields . " WHERE 	id='%d'", $field->id ) );

				$field_opts = maybe_unserialize( $field_opts->entry_value );

				if ( $field_opts && ! empty( $field_opts_label ) ) {
					$json_field_options_arr = json_decode( $field_opts_label->field_options );
					$json_options_arr       = json_decode( $field_opts_label->options );

					if ( isset( $json_field_options_arr->separate_value ) && $json_field_options_arr->separate_value == 1 && ! empty( $json_options_arr ) && $is_conditoinal == '' ) {
						foreach ( $json_options_arr as $key => $opt_val ) {

							if ( $opt_val->value == $field_opts['value'] ) {
								$options = $opt_val->value . ' (' . $opt_val->label . ')';
							}
						}
					}
				}

				return $options;
				break;

			case 'arf_autocomplete':
				$options = $field->entry_value;

				$field_opts       = $wpdb->get_row( $wpdb->prepare( 'SELECT entry_value FROM ' . $MdlDb->entry_metas . " WHERE field_id='%d' AND entry_id='%d'", '-' . $field->id, $entry_id ) );
				$field_opts_label = $wpdb->get_row( $wpdb->prepare( 'SELECT field_options, options FROM ' . $MdlDb->fields . " WHERE 	id='%d'", $field->id ) );

				$field_opts = maybe_unserialize( $field_opts->entry_value );

				if ( $field_opts && ! empty( $field_opts_label ) ) {
					$json_field_options_arr = json_decode( $field_opts_label->field_options );
					$json_options_arr       = json_decode( $field_opts_label->options );

					if ( isset( $json_field_options_arr->separate_value ) && $json_field_options_arr->separate_value == 1 && ! empty( $json_options_arr ) && $is_conditoinal == '' ) {
						foreach ( $json_options_arr as $key => $opt_val ) {
							if ( $opt_val->value == $field_opts['value'] ) {
								$options = $opt_val->value . ' (' . $opt_val->label . ')';
							}
						}
					}
				}

				return $options;
				break;

			case 'image':
				if ( $field->entry_value == '' ) {
					return '';
				} else {
					return "<img height='50px' src='" . $field->entry_value . "'>";
				}
				break;
			case 'textarea':
				return stripslashes( nl2br( $field->entry_value ) );
				break;

			case 'divider':
				return '';
				break;

			case 'section':
				return '';
				break;

			case 'date':
				if ( $is_version_compatible ) {

					$field_form_id = $wpdb->get_row( $wpdb->prepare( 'SELECT form_id,field_options FROM ' . $MdlDb->fields . " WHERE id='%d'", $field->id ) );

					if( '' == $field->field_options ){
						$field->field_options = json_decode( $field_form_id->field_options );
					}
					return $arfieldhelper->get_date_entry( $field->entry_value, $field_form_id->form_id, $field->field_options->show_time_calendar, $field->field_options->clock, $field->field_options->locale, 'arformspdfcreator' );
				} else {
					return $field->entry_value;
				}
				break;

			case 'extra_field':
				global $arrecordcontroller;
				$field_value = '';
				$field_name  = '';
				if ( $field->name == 'IPAddress' ) {
					$field_name = 'ip_address';
				}
				if ( $field->name == 'Browser' ) {
					$field_name = 'browser_info';
				}
				if ( $field->name == 'PageURL' || $field->name == 'ReferrerURL' ) {
					$field_name = 'description';
				}
				if ( $field->name == 'Country' ) {
					$field_name = 'country';
				}
				if ( $field->name == 'EntryCreationDate' ) {
					$field_name = 'created_date';
				}

				if ( $field_name != '' ) {

					$extra_field = $wpdb->get_row( $wpdb->prepare( 'SELECT ' . $field_name . ' FROM ' . $MdlDb->entries . ' WHERE id=%d', $entry_id ), ARRAY_A );
					if ( ! empty( $extra_field ) ) {
						if ( $field_name == 'description' ) {
							$tmp_val = maybe_unserialize( $extra_field[ $field_name ] );
							if ( $field->name == 'PageURL' ) {
								$field_value = $tmp_val['page_url'];
							}
							if ( $field->name == 'ReferrerURL' ) {
								$field_value = $tmp_val['http_referrer'];
							}
						} elseif ( $field_name == 'browser_info' ) {
							$browser_info = $arrecordcontroller->getBrowser( $extra_field[ $field_name ] );
							$field_value  = $browser_info['name'] . '(' . $browser_info['version'] . ')';
						} else {
							$field_value = $extra_field[ $field_name ];
						}
					}
				}
				return $field_value;
				break;
			case 'arf_repeater':
				$field_value = "";
				$field_value .= "<style>
					.arf_repeater_table, .arf_repeater_table th, .arf_repeater_table td {
					  border: 1px solid #ccc;
					  border-collapse: collapse;
					}
					.arf_repeater_table th, .arf_repeater_table td {
					  padding: 10px;
					}
				</style>";
				$get_all_inner_fields = $wpdb->get_results( $wpdb->prepare( "SELECT a.*, b.* FROM `".$MdlDb->fields."` a LEFT JOIN ".$MdlDb->entry_metas." b ON b.field_id = a.id WHERE ( field_options REGEXP '.*(\"parent_field\":\"%d\").*'  OR field_options REGEXP '.*(\"parent_field\":%d).*' ) AND entry_id=%d GROUP BY a.id", $field->id, $field->id, $entry_id ), ARRAY_A );

				if(!empty($get_all_inner_fields)) {
					if( count( $get_all_inner_fields) == 1){
						$valign = "";
	                    $inner_field_content = "";

	                    foreach ($get_all_inner_fields as $key => $inner_fields) {
	                        
	                        if(!empty($inner_fields['entry_value']) && strpos($inner_fields['entry_value'],"[ARF_JOIN]") != -1) {
	                            $valign = 'valign="top"';
	                            $entry_val = !empty($inner_fields['entry_value']) ? explode("[ARF_JOIN]", $inner_fields['entry_value']) : array(); 

	                            $row_span = count($entry_val);
	                            $cnt = 0;
	                            foreach ($entry_val as $key2 => $value) {
	                                $inner_field_content .= "<tr>";
	                                if( $inner_fields['type'] == 'date' ){
	                                	$inner_field_opts = json_decode( $inner_fields['field_options'] );
	                                	$value = $arfieldhelper->get_date_entry( $value,$inner_fields['form_id'],$inner_field_opts->show_time_calendar, $inner_field_opts->clock,$inner_field_opts->locale, 'arformspdfcreator'  );
	                                }
	                                if($cnt==0) {
	                                    $inner_field_content .= "<td rowspan='".$row_span."' valign='top'>".$inner_fields['name']."</td><td>".$value."</td>";
	                                } else {
	                                    $inner_field_content .= "<td>".$value."</td>";
	                                }                                    
	                                $cnt++;
	                                $inner_field_content .= "</tr>";
	                            }
	                            
	                        } else {
	                            if( $inner_fields['type'] == 'date' ){
	                                	$inner_field_opts = json_decode( $inner_fields['field_options'] );
	                                	$inner_fields['entry_value'] = $arfieldhelper->get_date_entry( $inner_fields['entry_value'],$inner_fields['form_id'],$inner_field_opts->show_time_calendar, $inner_field_opts->clock,$inner_field_opts->locale, 'arformspdfcreator'  );
	                                }
	                            $inner_field_content .= "<tr>";
	                                $inner_field_content .= "<td rowspan='".$row_span."'>".$inner_fields['name']."</td><td>".$inner_fields['entry_value']."</td>";
	                            $inner_field_content .= "</tr>";
	                        }                    
	                        
	                    }
	                    $field_value .= "<table class='arf_repeater_table'><tbody>";
	                    $field_value .= $inner_field_content;
	                    $field_value .= "</tbody></table>";
					} else {
						$inner_field_content  = "<table class='arf_repeater_table'>";
							$inner_field_content .= "<thead>";
								$fields_arr = [];
								$inner_field_content .= "<tr>";
									foreach( $get_all_inner_fields as $key => $inner_fields ){
										$inner_field_content .= "<th>" . $inner_fields['name'] . "</th>";
										if(!empty($inner_fields['entry_value']) && strpos($inner_fields['entry_value'],"[ARF_JOIN]") != -1) {
											
	                                        $entry_val = !empty($inner_fields['entry_value']) ? explode("[ARF_JOIN]", $inner_fields['entry_value']) : array(); 

	                                        $has_separate_value = false;
	                                        if( in_array( $inner_fields['type'], array('checkbox', 'radio', 'select', 'arf_autocomplete') ) ){
	                                            $field_opts = arf_json_decode( $inner_fields['field_options'], true );
	                                            if( isset( $field_opts['separate_value'] ) && '1' == $field_opts['separate_value'] ){
	                                                $has_separate_value = true;
	                                            }
	                                        }
	                                        $cnt = 0;

	                                        foreach ($entry_val as $key2 => $in_value) {

	                                            if( preg_match('/\!\|\!/', $in_value) ){
	                                                $value_exp = explode( '!|!', $in_value );
	                                                
	                                                $temp_val = '';
	                                                $value_arr = arf_json_decode( $value_exp[0], true );

	                                                if( $has_separate_value ){
	                                                    $sep_value = '';
	                                                    $fopts = arf_json_decode( $inner_fields['options'], true );
	                                                    foreach( $value_arr as $value_arrk => $varrv ){
	                                                        $fopt_key = $arformcontroller->arfSearchArray($varrv, 'value', $fopts);
	                                                        $sep_value .= $fopts[$fopt_key]['label'] . ' ('.$fopts[$fopt_key]['value'].'),';
	                                                    }
	                                                    $in_value = rtrim( $sep_value, ',');
	                                                } else {
	                                                	if( is_array( $value_arr ) ){
	                                                    	$temp_val = implode( ',',$value_arr );
	                                                	} else {
	                                                		$temp_val = $value_arr;
	                                                	}
	                                                    
	                                                    $in_value = $temp_val;
	                                                }
	                                            } else if( $inner_fields['type'] == 'date' ){
					                                	$inner_field_opts = json_decode( $inner_fields['field_options'] );
					                                	$in_value = $arfieldhelper->get_date_entry( $in_value,$inner_fields['form_id'],$inner_field_opts->show_time_calendar, $inner_field_opts->clock,$inner_field_opts->locale, 'arformspdfcreator'  );
	                                            }
	                                            $fields_arr[$cnt][$inner_fields['field_id']] = $in_value;
	                                            $cnt++;
	                                        }
	                                    } else{
	                                        if( preg_match('/\!\|\!/', $inner_fields['entry_value']) ){
	                                            $value_exp = explode( '!|!', $inner_fields['entry_value'] );
	                                            $temp_val = '';
	                                            $value_arr = arf_json_decode( $value_exp[0], true );
	                                            $temp_val = implode( ',',$value_arr );
	                                            $inner_fields['entry_value'] = $temp_val;
	                                        }
	                                    }
									}
								$inner_field_content .= "</tr>";
							$inner_field_content .= "</thead>";
							//die;
							$inner_field_content .= "<tbody>";
								foreach( $fields_arr as $k => $frr ){
                                    $inner_field_content .= "<tr>";
                                        foreach( $get_all_inner_fields as $inner_fields ){
                                            $inner_field_content .= "<td>" . $frr[$inner_fields['field_id']] . "</td>";
                                        }
                                    $inner_field_content .= "</tr>";
                                }
							$inner_field_content .= "</tbody>";

						$inner_field_content .= "</table>";

						$field_value .= $inner_field_content;
					}

				}

				return $field_value;
				break;
			default:
				return stripslashes( $field->entry_value );
				break;
		}

		return $field->entry_value;
	}

	function ARFPDF_Convert_Shortcode_values( $values, $content, $entry_id ) {

		$total_values = count( $values );

		global $wpdb, $MdlDb, $ARFPDF_LABEL_SHORTCODE, $ARFPDF_FIELD_SHORTCODE, $ARFPDF_SHORTCODE_OVER, $ARFPDF_VALUE_SHORTCODE, $ARFPDF_ENTRY_ID_SHORTCODE, $ARFPDF_LOOP_FIRST, $ARFPDF_LOOP_LAST, $ARFPDF_LOOP_VAR, $ARFPDF_LOOP_VAR_ONLY, $ARFPDF_LOOP_VAR_TO, $ARFPDF_IF_CONDITION_START, $ARFPDF_IF_CONDITION_END, $ARFPDF_IF_OR_CONDITION, $ARFPDF_IF_AND_CONDITION, $ARFPDF_FIELDTYPE_SHORTCODE, $ARFPDF_MODULE_SHORTCODE, $ARFPDF_SIGNATURE_SHORTCODE_START, $ARFPDF_SIGNATURE_SHORTCODE_END, $ARFPDF_EXTRAFIELD_SHORTCODE, $arformspdfcreator, $ARFPDF_FORMTITLE_SHORTCODE, $ARFPDF_USERID_SHORTCODE, $ARFPDF_USEREMAIL_SHORTCODE, $ARFPDF_USERNAME_SHORTCODE, $ARFPDF_USERDISPLAYNAME_SHORTCODE;

		$result                  = $wpdb->get_row( $wpdb->prepare( 'SELECT user_id FROM ' . $MdlDb->entries . ' WHERE id=%d', $entry_id ) );
		$arf_entry_user_id       = ( ! empty( $result ) ) ? $result->user_id : 0;
		$entry_user_id           = '';
		$entry_user_email        = '';
		$entry_user_name         = '';
		$entry_user_display_name = '';
		if ( $arf_entry_user_id != 0 ) {
			$entry_user_data         = get_userdata( $arf_entry_user_id );
			$entry_user_id           = $entry_user_data->ID;
			$entry_user_email        = $entry_user_data->user_email;
			$entry_user_name         = $entry_user_data->user_nicename;
			$entry_user_display_name = $entry_user_data->display_name;
		}

		$delimiter        = '#';
		$signaturematches = $delimiter . preg_quote( $ARFPDF_SIGNATURE_SHORTCODE_START, $delimiter )
		. '(.*?)(.*?)'
		. preg_quote( $ARFPDF_SIGNATURE_SHORTCODE_END, $delimiter )
		. $delimiter
		. 's';
		$signaturevalue   = preg_match_all( $signaturematches, $content, $signaturevalues );

		if ( $signaturevalue > 0 ) {
			foreach ( $signaturevalues[0] as $k => $v ) {
				$content = str_replace( $signaturevalues[ $k ], '', $content );
			}
		}

		$loopmatches  = $delimiter . preg_quote( $ARFPDF_LOOP_FIRST, $delimiter )
		. '(.*?)](.*?)'
		. preg_quote( $ARFPDF_LOOP_LAST, $delimiter )
		. $delimiter
		. 's';
		$chkloopvalue = preg_match_all( $loopmatches, $content, $loopvalues );

		if ( $chkloopvalue > 0 ) {

			$original_loopvalues = $loopvalues[0];
			$parameters_value    = $loopvalues[1];
			$loopvalues_value    = $loopvalues[2];

			if ( count( $loopvalues_value ) > 0 ) {
				$cntr = 0;
				foreach ( $loopvalues_value as $loopvalue ) {

					$chkstartend = strtolower( trim( $parameters_value[ $cntr ] ) );

					if ( $chkstartend != '' ) {
						$startloop_exp     = explode( $ARFPDF_LOOP_VAR, $chkstartend );
						$startloop_exp_val = explode( ' ', $startloop_exp[1] );
						$startloop         = trim( $startloop_exp_val[0] );
						if ( $startloop == '' ) {
							   $startloop = 1;
						}

						$endloop_exp     = explode( $ARFPDF_LOOP_VAR_TO, $chkstartend );
						$startloop_exp1  = explode( $ARFPDF_LOOP_VAR, $endloop_exp[1] );
						$endloop_exp_val = explode( ' ', $startloop_exp1[1] );
						$endloop         = trim( $endloop_exp_val[0] );

						if ( $endloop == '' ) {
							$endloop = 1;
						}
					} else {
						$startloop = 1;
						$endloop   = $total_values;
					}

					$newhtmlcontent = '';
					$for_counter    = 0;
					$ifnewloopvalue = '';
					if ( $startloop <= $endloop ) {
						for ( $i = $startloop; $i <= $endloop; $i++ ) {
							$newloopvalue = $loopvalue;

							$ifcondition_loopmatches = $delimiter . preg_quote( $ARFPDF_IF_CONDITION_START, $delimiter )
							. '(.*?)](.*?)'
							. preg_quote( $ARFPDF_IF_CONDITION_END, $delimiter )
							. $delimiter
							. 's';
							$chkifloopvalue          = preg_match_all( $ifcondition_loopmatches, $newloopvalue, $ifloopvalues );
							if ( $chkifloopvalue > 0 ) {

								$if_original_loopvalues = $ifloopvalues[0];
								$if_parameters_value    = $ifloopvalues[1];
								$if_loopvalues_value    = $ifloopvalues[2];

								$if_cntr                 = 0;
								$field_counter           = $i;
								$if_loopvalues_value_arr = '';
								foreach ( $if_parameters_value as $k => $v ) {
									$if_loopvalues_value_arr .= $if_loopvalues_value[ $k ];
									$pattern_for_condition    = '/' . preg_quote( $ARFPDF_IF_OR_CONDITION, '/' ) . '/';
									$is_condition_available   = preg_match_all( $pattern_for_condition, $if_parameters_value[ $k ], $condition_values_arr );
									if ( $is_condition_available > 0 ) {
										$condition_field = 'OR';
									} else {
										$condition_field = 'AND';
									}
									$check_template_conditions = $arformspdfcreator->CheckTemplateConditionsValue( $if_parameters_value[ $k ], $condition_field, $newloopvalue, $values, $field_counter, $arf_entry_user_id );
									if ( $check_template_conditions == 1 ) {
										$ifnewloopvalue = $if_loopvalues_value[ $k ];
										$newloopvalue   = str_replace( $if_original_loopvalues[ $k ], $ifnewloopvalue, $newloopvalue );
									} else {
										$ifnewloopvalue = '';
										$newloopvalue   = str_replace( $if_original_loopvalues[ $k ], $ifnewloopvalue, $newloopvalue );
									}
									$if_cntr++;
								}
							}

							$field_id          = '';
							$pattern_for_label = '/' . preg_quote( $ARFPDF_LABEL_SHORTCODE, '/' ) . '/';

							$is_label_available = preg_match_all( $pattern_for_label, $newloopvalue, $content_labels_arr );
							if ( $is_label_available > 0 ) {

								foreach ( $content_labels_arr[0] as $content_label_arr ) {
									$parts    = ( ! empty( $ARFPDF_LABEL_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_LABEL_SHORTCODE, $newloopvalue ) : '';
									$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';

									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
												$get_label_value = $arformspdfcreator->get_label( $value );
												$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
											}
										}
										$get_label_value = $arformspdfcreator->get_label( $values[ $field_id ] );
										$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
									} else {
										$get_label_value = $arformspdfcreator->get_label( $values[ $i ] );
										$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
									}
								}
							}

							$field_id          = '';
							$pattern_for_value = '/' . preg_quote( $ARFPDF_VALUE_SHORTCODE, '/' ) . '/';

							$is_value_available = preg_match_all( $pattern_for_value, $newloopvalue, $content_values_arr );
							if ( $is_value_available > 0 ) {
								foreach ( $content_values_arr[0] as $content_value_arr ) {
									$parts             = ( ! empty( $ARFPDF_VALUE_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_VALUE_SHORTCODE, $newloopvalue ) : '';
									$myidpart          = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
												$get_value_value = $arformspdfcreator->get_value( $value, $entry_id );
												$newloopvalue    = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $newloopvalue );
											}
										}
										$search_value    = $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER;
										$get_value_value = $arformspdfcreator->get_value( $values[ $field_id ], $entry_id );

										$newloopvalue = str_replace( $search_value, $get_value_value, $newloopvalue );
									} else {
										$get_value_value = $arformspdfcreator->get_value( $values[ $i ], $entry_id );

										$newloopvalue = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $newloopvalue );
									}
								}
							}

							$field_id               = '';
							$pattern_for_fieldtype  = '/' . preg_quote( $ARFPDF_FIELDTYPE_SHORTCODE, '/' ) . '/';
							$is_fieldtype_available = preg_match_all( $pattern_for_fieldtype, $newloopvalue, $content_fieldtypes_arr );
							if ( $is_fieldtype_available > 0 ) {
								foreach ( $content_fieldtypes_arr[0] as $content_fieldtype_arr ) {
									$parts = ( ! empty( $ARFPDF_FIELDTYPE_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_FIELDTYPE_SHORTCODE, $newloopvalue ) : '';

									$myidpart          = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
												$get_fieldtype_value = $arformspdfcreator->get_fieldtype( $value );
												$newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
											}
										}
											  $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $field_id ] );
											  $newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
									} else {
										   $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $i ] );
										   $newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
									}
								}
							}

							$pattern_for_user_id = '/' . preg_quote( $ARFPDF_USERID_SHORTCODE, '/' ) . '/';
							$is_user_id          = preg_match_all( $pattern_for_user_id, $content, $content_userid_arr );

							if ( $is_user_id > 0 ) {
								foreach ( $content_userid_arr[0] as $userid_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERID_SHORTCODE, $entry_user_id, $newloopvalue );
								}
							}

							$pattern_for_user_email = '/' . preg_quote( $ARFPDF_USEREMAIL_SHORTCODE, '/' ) . '/';
							$is_user_emai           = preg_match_all( $pattern_for_user_email, $content, $content_useremail_arr );

							if ( $is_user_emai > 0 ) {
								foreach ( $content_useremail_arr[0] as $useremail_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USEREMAIL_SHORTCODE, $entry_user_email, $newloopvalue );
								}
							}

							$pattern_for_user_name = '/' . preg_quote( $ARFPDF_USERNAME_SHORTCODE, '/' ) . '/';
							$is_user_name          = preg_match_all( $pattern_for_user_name, $content, $content_username_arr );

							if ( $is_user_name > 0 ) {
								foreach ( $content_username_arr[0] as $username_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERNAME_SHORTCODE, $entry_user_name, $newloopvalue );
								}
							}

							$pattern_for_user_display_name = '/' . preg_quote( $ARFPDF_USERDISPLAYNAME_SHORTCODE, '/' ) . '/';
							$is_user_display_name          = preg_match_all( $pattern_for_user_display_name, $content, $content_userdisplayname_arr );

							if ( $is_user_display_name > 0 ) {
								foreach ( $content_userdisplayname_arr[0] as $userdisplayname_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERDISPLAYNAME_SHORTCODE, $entry_user_display_name, $newloopvalue );
								}
							}

							$field_id               = '';
							$pattern_for_extrafield = '/' . preg_quote( $ARFPDF_EXTRAFIELD_SHORTCODE, '/' ) . '/';

							$is_extrafield_available = preg_match_all( $pattern_for_extrafield, $newloopvalue, $content_extrafield_arr );

							if ( $is_extrafield_available > 0 ) {

								foreach ( $content_extrafield_arr[0] as $extrafield_arr ) {

									$parts = ( ! empty( $ARFPDF_EXTRAFIELD_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_EXTRAFIELD_SHORTCODE, $newloopvalue ) : '';

									$extrafield = ( ! empty( $parts[1] ) ) ? explode( '}', $parts[1] ) : '';

									$field_arr = (object) array(
										'type' => 'extra_field',
										'name' => $extrafield[0],
									);

									if ( count( $extrafield ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {

										   $get_extrafield_value = $arformspdfcreator->get_value( $field_arr, $entry_id );

										   $newloopvalue = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $newloopvalue );
									} else {

										   $get_value_value = $arformspdfcreator->get_value( $field_arr, $entry_id );
										   $newloopvalue    = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $newloopvalue );
									}
								}
							}

							$newhtmlcontent .= $newloopvalue;
							$for_counter++;
						}
					} else {
						for ( $i = $startloop; $i >= $endloop; $i-- ) {
							$newloopvalue = $loopvalue;

							$ifcondition_loopmatches = $delimiter . preg_quote( $ARFPDF_IF_CONDITION_START, $delimiter )
							. '(.*?)](.*?)'
							. preg_quote( $ARFPDF_IF_CONDITION_END, $delimiter )
							. $delimiter
							. 's';
							$chkifloopvalue          = preg_match_all( $ifcondition_loopmatches, $newloopvalue, $ifloopvalues );
							if ( $chkifloopvalue > 0 ) {
								   $if_original_loopvalues = $ifloopvalues[0];
								   $if_parameters_value    = $ifloopvalues[1];
								   $if_loopvalues_value    = $ifloopvalues[2];

								   $if_cntr                 = 0;
								   $field_counter           = $i;
								   $if_loopvalues_value_arr = '';
								foreach ( $if_parameters_value as $k => $v ) {
									$if_loopvalues_value_arr .= $if_loopvalues_value[ $k ];
									$pattern_for_condition    = '/' . preg_quote( $ARFPDF_IF_OR_CONDITION, '/' ) . '/';
									$is_condition_available   = preg_match_all( $pattern_for_condition, $if_parameters_value[ $k ], $condition_values_arr );
									if ( $is_condition_available > 0 ) {
										   $condition_field = 'OR';
									} else {
										$condition_field = 'AND';
									}
									$check_template_conditions = $arformspdfcreator->CheckTemplateConditionsValue( $if_parameters_value[ $k ], $condition_field, $newloopvalue, $values, $field_counter, $arf_entry_user_id );
									if ( $check_template_conditions == 1 ) {
										$ifnewloopvalue = $if_loopvalues_value[ $k ];
										$newloopvalue   = str_replace( $if_original_loopvalues[ $k ], $ifnewloopvalue, $newloopvalue );
									} else {
										$ifnewloopvalue = '';
										$newloopvalue   = str_replace( $if_original_loopvalues[ $k ], $ifnewloopvalue, $newloopvalue );
									}
									$if_cntr++;
								}
							}

							$field_id           = '';
							$pattern_for_label  = '/' . preg_quote( $ARFPDF_LABEL_SHORTCODE, '/' ) . '/';
							$is_label_available = preg_match_all( $pattern_for_label, $newloopvalue, $content_labels_arr );
							if ( $is_label_available > 0 ) {
								foreach ( $content_labels_arr[0] as $content_label_arr ) {
									$parts = ( ! empty( $ARFPDF_LABEL_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_LABEL_SHORTCODE, $newloopvalue ) : '';

									$myidpart          = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
												$get_label_value = $arformspdfcreator->get_label( $value );
												$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
											}
										}
										$get_label_value = $arformspdfcreator->get_label( $values[ $field_id ] );
										$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
									} else {
										$get_label_value = $arformspdfcreator->get_label( $values[ $i ] );
										$newloopvalue    = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $newloopvalue );
									}
								}
							}

							$field_id          = '';
							$pattern_for_value = '/' . preg_quote( $ARFPDF_VALUE_SHORTCODE, '/' ) . '/';

							$is_value_available = preg_match_all( $pattern_for_value, $newloopvalue, $content_values_arr );
							if ( $is_value_available > 0 ) {
								foreach ( $content_values_arr[0] as $content_value_arr ) {
									$parts             = ( ! empty( $ARFPDF_VALUE_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_VALUE_SHORTCODE, $newloopvalue ) : '';
									$myidpart          = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
												$get_value_value = $arformspdfcreator->get_value( $value, $entry_id );
												$newloopvalue    = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $newloopvalue );
											}
										}
										$get_value_value = $arformspdfcreator->get_value( $values[ $field_id ], $entry_id );
										$newloopvalue    = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $newloopvalue );
									} else {
										   $get_value_value = $arformspdfcreator->get_value( $values[ $i ], $entry_id );
										   $newloopvalue    = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $newloopvalue );
									}
								}
							}

							$field_id               = '';
							$pattern_for_fieldtype  = '/' . preg_quote( $ARFPDF_FIELDTYPE_SHORTCODE, '/' ) . '/';
							$is_fieldtype_available = preg_match_all( $pattern_for_fieldtype, $newloopvalue, $content_fieldtypes_arr );
							if ( $is_fieldtype_available > 0 ) {
								foreach ( $content_fieldtypes_arr[0] as $content_fieldtype_arr ) {
									$parts = ( ! empty( $ARFPDF_FIELDTYPE_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_FIELDTYPE_SHORTCODE, $newloopvalue ) : '';

									$myidpart          = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
									$myid              = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
									$field_id          = trim( strtolower( $myid[0] ) );
									$original_field_id = $myid[0];
									if ( count( $myid ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {
										foreach ( $values as $value ) {
											if ( $value->id == $original_field_id ) {
													  $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $value );
													  $newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
											}
										}
										   $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $field_id ] );
										   $newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
									} else {
										   $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $i ] );
										   $newloopvalue        = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $original_field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $newloopvalue );
									}
								}
							}

							$pattern_for_user_id = '/' . preg_quote( $ARFPDF_USERID_SHORTCODE, '/' ) . '/';
							$is_user_id          = preg_match_all( $pattern_for_user_id, $content, $content_userid_arr );

							if ( $is_user_id > 0 ) {
								foreach ( $content_userid_arr[0] as $userid_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERID_SHORTCODE, $entry_user_id, $newloopvalue );
								}
							}

							$pattern_for_user_email = '/' . preg_quote( $ARFPDF_USEREMAIL_SHORTCODE, '/' ) . '/';
							$is_user_emai           = preg_match_all( $pattern_for_user_email, $content, $content_useremail_arr );

							if ( $is_user_emai > 0 ) {
								foreach ( $content_useremail_arr[0] as $useremail_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USEREMAIL_SHORTCODE, $entry_user_email, $newloopvalue );
								}
							}

							$pattern_for_user_name = '/' . preg_quote( $ARFPDF_USERNAME_SHORTCODE, '/' ) . '/';
							$is_user_name          = preg_match_all( $pattern_for_user_name, $content, $content_username_arr );

							if ( $is_user_name > 0 ) {
								foreach ( $content_username_arr[0] as $username_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERNAME_SHORTCODE, $entry_user_name, $newloopvalue );
								}
							}

							$pattern_for_user_display_name = '/' . preg_quote( $ARFPDF_USERDISPLAYNAME_SHORTCODE, '/' ) . '/';
							$is_user_display_name          = preg_match_all( $pattern_for_user_display_name, $content, $content_userdisplayname_arr );

							if ( $is_user_display_name > 0 ) {
								foreach ( $content_userdisplayname_arr[0] as $userdisplayname_arr ) {
									$newloopvalue = str_replace( $ARFPDF_USERDISPLAYNAME_SHORTCODE, $entry_user_display_name, $newloopvalue );
								}
							}

							$field_id               = '';
							$pattern_for_extrafield = '/' . preg_quote( $ARFPDF_EXTRAFIELD_SHORTCODE, '/' ) . '/';

							$is_extrafield_available = preg_match_all( $pattern_for_extrafield, $newloopvalue, $content_extrafield_arr );

							if ( $is_extrafield_available > 0 ) {

								foreach ( $content_extrafield_arr[0] as $extrafield_arr ) {

									$parts = ( ! empty( $ARFPDF_EXTRAFIELD_SHORTCODE ) && ! empty( $newloopvalue ) ) ? explode( $ARFPDF_EXTRAFIELD_SHORTCODE, $newloopvalue ) : '';

									$extrafield = ( ! empty( $parts[1] ) ) ? explode( '}', $parts[1] ) : '';

									$field_arr = (object) array(
										'type' => 'extra_field',
										'name' => $extrafield[0],
									);

									if ( count( $extrafield ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {

										   $get_extrafield_value = $arformspdfcreator->get_value( $field_arr, $entry_id );

										   $newloopvalue = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $newloopvalue );
									} else {

										   $get_value_value = $arformspdfcreator->get_value( $field_arr, $entry_id );
										   $newloopvalue    = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $newloopvalue );
									}
								}
							}

							$newhtmlcontent .= $newloopvalue;
							$for_counter++;
						}
					}

					$content = str_replace( $original_loopvalues[ $cntr ], $newhtmlcontent, $content );
					$cntr++;
				}
			}
		}

		$ifcondition_loopmatches = $delimiter . preg_quote( $ARFPDF_IF_CONDITION_START, $delimiter )
		. '(.*?)](.*?)'
		. preg_quote( $ARFPDF_IF_CONDITION_END, $delimiter )
		. $delimiter
		. 's';

		$chkifloopvalue = preg_match_all( $ifcondition_loopmatches, $content, $ifloopvalues );
		if ( $chkifloopvalue > 0 ) {

			$if_original_loopvalues = $ifloopvalues[0];
			$if_parameters_value    = $ifloopvalues[1];
			$if_loopvalues_value    = $ifloopvalues[2];

			$if_cntr = 0;
			foreach ( $if_parameters_value as $if_parameter_value ) {
				$pattern_for_condition  = '/' . preg_quote( $ARFPDF_IF_OR_CONDITION, '/' ) . '/';
				$is_condition_available = preg_match_all( $pattern_for_condition, $if_parameters_value[ $if_cntr ], $condition_values_arr );
				if ( $is_condition_available > 0 ) {
					$condition_field = 'OR';
				} else {
					$condition_field = 'AND';
				}
				$check_template_conditions = $arformspdfcreator->CheckTemplateConditionsValue( $if_parameter_value, $condition_field, $content, $values, 0, $arf_entry_user_id );

				if ( $check_template_conditions == 1 ) {
					$content = str_replace( $if_original_loopvalues[ $if_cntr ], $if_loopvalues_value[ $if_cntr ], $content );
				} else {
					$content = str_replace( $if_original_loopvalues[ $if_cntr ], '', $content );
				}
				$if_cntr++;
			}
		}

		$field_id           = '';
		$pattern_for_label  = '/' . preg_quote( $ARFPDF_LABEL_SHORTCODE, '/' ) . '/';
		$is_label_available = preg_match_all( $pattern_for_label, $content, $content_labels_arr );
		if ( $is_label_available > 0 ) {
			foreach ( $content_labels_arr[0] as $content_label_arr ) {
				$parts    = ( ! empty( $ARFPDF_LABEL_SHORTCODE ) && ! empty( $content ) ) ? explode( $ARFPDF_LABEL_SHORTCODE, $content ) : '';
				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];
					foreach ( $values as $value ) {
						if ( $value->id == $field_id ) {
							   $get_label_value = $arformspdfcreator->get_label( $value );
							   $content         = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $content );
						}
					}
					$get_label_value = $arformspdfcreator->get_label( $values[ $field_id ] );
					$content         = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_label_value, $content );
				} else {
					$content = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, '', $content );
				}
			}
		}

		$field_id          = '';
		$pattern_for_value = '/' . preg_quote( $ARFPDF_VALUE_SHORTCODE, '/' ) . '/';

		$is_value_available = preg_match_all( $pattern_for_value, $content, $content_values_arr );
		if ( $is_value_available > 0 ) {
			foreach ( $content_values_arr[0] as $content_value_arr ) {
				$parts = ( ! empty( $ARFPDF_VALUE_SHORTCODE ) && ! empty( $content ) ) ? explode( $ARFPDF_VALUE_SHORTCODE, $content ) : '';

				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];

					foreach ( $values as $value ) {
						if ( $value->id == $field_id ) {
							   $get_value_value = $arformspdfcreator->get_value( $value, $entry_id );
							   $content         = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $content );
						}
					}
					$get_value_value = $arformspdfcreator->get_value( $values[ $field_id ], $entry_id );
					$content         = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_value_value, $content );
				} else {
					$content = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, '', $content );
				}
			}
		}

		$field_id              = '';
		$pattern_for_entry_id  = '/' . preg_quote( $ARFPDF_ENTRY_ID_SHORTCODE, '/' ) . '/';
		$is_entry_id_available = preg_match_all( $pattern_for_entry_id, $content, $content_entry_ids_arr );
		if ( $is_entry_id_available > 0 ) {
			foreach ( $content_entry_ids_arr[0] as $content_entry_id_arr ) {
				$parts = ( ! empty( $ARFPDF_ENTRY_ID_SHORTCODE ) && ! empty( $content ) ) ? explode( $ARFPDF_ENTRY_ID_SHORTCODE, $content ) : '';

				$myid = ( ! empty( $parts[1] ) ) ? explode( '}', $parts[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$get_entry_id_value = $entry_id;
					$content            = str_replace( $ARFPDF_ENTRY_ID_SHORTCODE . $ARFPDF_SHORTCODE_OVER, $get_entry_id_value, $content );
				} else {
					$content = str_replace( $ARFPDF_ENTRY_ID_SHORTCODE . $ARFPDF_SHORTCODE_OVER, '', $content );
				}
			}
		}

		$field_id               = '';
		$pattern_for_fieldtype  = '/' . preg_quote( $ARFPDF_FIELDTYPE_SHORTCODE, '/' ) . '/';
		$is_fieldtype_available = preg_match_all( $pattern_for_fieldtype, $content, $content_fieldtypes_arr );

		if ( $is_fieldtype_available > 0 ) {
			foreach ( $content_fieldtypes_arr[0] as $content_fieldtype_arr ) {
				$parts = ( ! empty( $ARFPDF_FIELDTYPE_SHORTCODE ) && ! empty( $content ) ) ? explode( $ARFPDF_FIELDTYPE_SHORTCODE, $content ) : '';

				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];
					foreach ( $values as $value ) {
						if ( $value->id == $field_id ) {
							   $get_fieldtype_value = $arformspdfcreator->get_fieldtype( $value );
							   $content             = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $content );
						}
					}
					$get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $field_id ] );
					$content             = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $content );
				} else {
					$content = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $field_id . $ARFPDF_SHORTCODE_OVER, '', $content );
				}
			}
		}

		$pattern_for_form_title = '/' . preg_quote( $ARFPDF_FORMTITLE_SHORTCODE, '/' ) . '/';
		$is_form_title          = preg_match_all( $pattern_for_form_title, $content, $content_formtitle_arr );

		if ( $is_form_title > 0 ) {

			foreach ( $content_formtitle_arr[0] as $formtitle_arr ) {

				$form_id = $wpdb->get_row( $wpdb->prepare( 'SELECT f.name, e.form_id FROM `' . $MdlDb->entries . '` e LEFT JOIN `' . $MdlDb->forms . '` f ON f.id = e.form_id WHERE e.id = %d', $entry_id ) );

				$form_name = $form_id->name;

				$content = str_replace( $ARFPDF_FORMTITLE_SHORTCODE, $form_name, $content );
			}
		}

		$pattern_for_user_id = '/' . preg_quote( $ARFPDF_USERID_SHORTCODE, '/' ) . '/';
		$is_user_id          = preg_match_all( $pattern_for_user_id, $content, $content_userid_arr );

		if ( $is_user_id > 0 ) {
			foreach ( $content_userid_arr[0] as $userid_arr ) {
				$content = str_replace( $ARFPDF_USERID_SHORTCODE, $entry_user_id, $content );
			}
		}

		$pattern_for_user_email = '/' . preg_quote( $ARFPDF_USEREMAIL_SHORTCODE, '/' ) . '/';
		$is_user_emai           = preg_match_all( $pattern_for_user_email, $content, $content_useremail_arr );

		if ( $is_user_emai > 0 ) {
			foreach ( $content_useremail_arr[0] as $useremail_arr ) {
				$content = str_replace( $ARFPDF_USEREMAIL_SHORTCODE, $entry_user_email, $content );
			}
		}

		$pattern_for_user_name = '/' . preg_quote( $ARFPDF_USERNAME_SHORTCODE, '/' ) . '/';
		$is_user_name          = preg_match_all( $pattern_for_user_name, $content, $content_username_arr );

		if ( $is_user_name > 0 ) {
			foreach ( $content_username_arr[0] as $username_arr ) {
				$content = str_replace( $ARFPDF_USERNAME_SHORTCODE, $entry_user_name, $content );
			}
		}

		$pattern_for_user_display_name = '/' . preg_quote( $ARFPDF_USERDISPLAYNAME_SHORTCODE, '/' ) . '/';
		$is_user_display_name          = preg_match_all( $pattern_for_user_display_name, $content, $content_userdisplayname_arr );

		if ( $is_user_display_name > 0 ) {
			foreach ( $content_userdisplayname_arr[0] as $userdisplayname_arr ) {
				$content = str_replace( $ARFPDF_USERDISPLAYNAME_SHORTCODE, $entry_user_display_name, $content );
			}
		}

		$field_id                = '';
		$pattern_for_extrafield  = '/' . preg_quote( $ARFPDF_EXTRAFIELD_SHORTCODE, '/' ) . '/';
		$is_extrafield_available = preg_match_all( $pattern_for_extrafield, $content, $content_extrafield_arr );

		if ( $is_extrafield_available > 0 ) {

			foreach ( $content_extrafield_arr[0] as $extrafield_arr ) {
				$parts = ( ! empty( $ARFPDF_EXTRAFIELD_SHORTCODE ) && ! empty( $content ) ) ? explode( $ARFPDF_EXTRAFIELD_SHORTCODE, $content ) : '';

				$extrafield = ( ! empty( $parts[1] ) ) ? explode( '}', $parts[1] ) : '';

				$field_arr = (object) array(
					'type' => 'extra_field',
					'name' => $extrafield[0],
				);

				if ( count( $extrafield ) > 1 && $field_id != $ARFPDF_LOOP_VAR_ONLY ) {

					$get_extrafield_value = $arformspdfcreator->get_value( $field_arr, $entry_id );

					$content = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $content );
				} else {

					$get_value_value = $arformspdfcreator->get_value( $field_arr, $entry_id );
					$content         = str_replace( $ARFPDF_EXTRAFIELD_SHORTCODE . $extrafield[0] . $ARFPDF_SHORTCODE_OVER, $get_extrafield_value, $content );
				}
			}
		}

		return $content;
	}

	function CheckTemplateConditionsValue( $if_parameter_value, $condition_field, $content, $values, $field_counter = 0, $arf_entry_user_id = 0 ) {
		global $ARFPDF_IF_OR_CONDITION, $ARFPDF_IF_AND_CONDITION, $arformspdfcreator;
		if ( $condition_field == 'AND' ) {

			$exp_conditions          = explode( $ARFPDF_IF_AND_CONDITION, $if_parameter_value );
			$ParseTemplateLabelValue = array();
			foreach ( $exp_conditions as $exp_condition ) {
				$parameter_value           = trim( $exp_condition );
				$parsedvaluecheckedresult  = $arformspdfcreator->ParseTemplateLabelValue( $parameter_value, $values, 1, $field_counter, $arf_entry_user_id );
				$ParseTemplateLabelValue[] = $parsedvaluecheckedresult;
			}
			if ( in_array( '0', $ParseTemplateLabelValue ) ) {
				return 0;
			} else {
				return 1;
			}
		} else {

			$exp_conditions          = explode( $ARFPDF_IF_OR_CONDITION, $if_parameter_value );
			$ParseTemplateLabelValue = array();

			foreach ( $exp_conditions as $exp_condition ) {
				$parameter_value           = trim( $exp_condition );
				$ParseTemplateLabelValue[] = $arformspdfcreator->ParseTemplateLabelValue( $parameter_value, $values, 1, $field_counter, $arf_entry_user_id );
			}
			if ( in_array( '1', $ParseTemplateLabelValue ) ) {
				return 1;
			} else {
				return 0;
			}
		}
	}

	function ParseTemplateLabelValue( $singlecontent, $values, $flag = 0, $field_counter, $arf_entry_user_id = 0 ) {
		global $ARFPDF_LABEL_SHORTCODE, $ARFPDF_FIELD_SHORTCODE, $ARFPDF_SHORTCODE_OVER, $ARFPDF_VALUE_SHORTCODE, $ARFPDF_ENTRY_ID_SHORTCODE, $ARFPDF_LOOP_FIRST, $ARFPDF_LOOP_LAST, $ARFPDF_LOOP_VAR, $ARFPDF_LOOP_VAR_ONLY, $ARFPDF_LOOP_VAR_TO, $ARFPDF_IF_CONDITION_START, $ARFPDF_IF_CONDITION_END, $ARFPDF_IF_OR_CONDITION, $ARFPDF_IF_AND_CONDITION, $ARFPDF_FIELDTYPE_SHORTCODE, $ARFPDF_MODULE_SHORTCODE, $ARFPDF_SIGNATURE_SHORTCODE_START, $ARFPDF_SIGNATURE_SHORTCODE_END, $ARFPDF_USERID_SHORTCODE, $ARFPDF_USEREMAIL_SHORTCODE, $ARFPDF_USERNAME_SHORTCODE, $ARFPDF_USERDISPLAYNAME_SHORTCODE, $arformspdfcreator;

		$PARSED_FIELD_ID = 0;

		$entry_user_id           = '';
		$entry_user_email        = '';
		$entry_user_name         = '';
		$entry_user_display_name = '';
		if ( $arf_entry_user_id != 0 ) {
			$entry_user_data         = get_userdata( $arf_entry_user_id );
			$entry_user_id           = $entry_user_data->ID;
			$entry_user_email        = $entry_user_data->user_email;
			$entry_user_name         = $entry_user_data->user_nicename;
			$entry_user_display_name = $entry_user_data->display_name;
		}

		$field_id           = '';
		$pattern_for_label  = '/' . preg_quote( $ARFPDF_LABEL_SHORTCODE, '/' ) . '/';
		$is_label_available = preg_match_all( $pattern_for_label, $singlecontent, $content_labels_arr );
		if ( $is_label_available > 0 ) {
			foreach ( $content_labels_arr[0] as $content_label_arr ) {
				$parts = ( ! empty( $ARFPDF_LABEL_SHORTCODE ) && ! empty( $singlecontent ) ) ? explode( $ARFPDF_LABEL_SHORTCODE, $singlecontent ) : '';

				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];
					if ( strtolower( $field_id ) == $ARFPDF_LOOP_VAR_ONLY && $field_counter > 0 ) {
						$field_id        = $field_counter;
						$get_label_value = $arformspdfcreator->get_label( $values[ $field_id ] );
						$singlecontent   = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_label_value, $singlecontent );
						$PARSED_FIELD_ID = $field_id;
					} else {
						foreach ( $values as $value ) {
							if ( $value->id == $field_id ) {
								$get_label_value = $arformspdfcreator->get_label( $value );
								$singlecontent   = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_label_value, $singlecontent );
								$PARSED_FIELD_ID = $field_id;
							}
						}
					}
				} else {
					$singlecontent = str_replace( $ARFPDF_LABEL_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, '', $singlecontent );
				}
			}
			$values_field_counter = $arformspdfcreator->get_label( $values[ $field_counter ] );
		}

		$field_id           = '';
		$pattern_for_value  = '/' . preg_quote( $ARFPDF_VALUE_SHORTCODE, '/' ) . '/';
		$is_value_available = preg_match_all( $pattern_for_value, $singlecontent, $content_values_arr );
		if ( $is_value_available > 0 ) {

			foreach ( $content_values_arr[0] as $content_value_arr ) {
				$parts = ( ! empty( $ARFPDF_VALUE_SHORTCODE ) && ! empty( $singlecontent ) ) ? explode( $ARFPDF_VALUE_SHORTCODE, $singlecontent ) : '';

				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';

				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];
					if ( strtolower( $field_id ) == $ARFPDF_LOOP_VAR_ONLY && $field_counter > 0 ) {
						$field_id        = $field_counter;
						$get_value_value = $arformspdfcreator->get_value( $values[ $field_id ], $entry_id, 'condition' );
						$singlecontent   = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_value_value, $singlecontent );
						$PARSED_FIELD_ID = $field_id;
					} else {
						global $arformcontroller;
						$values_array = $arformcontroller->arfObjtoArray($values);
						$if_field_available = $arformcontroller->arfSearchArray($field_id,'id',$values_array);
						if( '' != $if_field_available ){
							foreach ( $values as $value ) {
								if ( $value->id == $field_id ) {
									$get_value_value = $arformspdfcreator->get_value( $value, $entry_id, 'condition' );
									$singlecontent   = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_value_value, $singlecontent );
									$PARSED_FIELD_ID = $field_id;
								}
							}
						} else {
							$get_value_value = '';
							$singlecontent   = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_value_value, $singlecontent );
							$PARSED_FIELD_ID = $field_id;
						}
					}
				} else {
					$singlecontent = str_replace( $ARFPDF_VALUE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, '', $singlecontent );
				}
			}
			$values_field_counter = $arformspdfcreator->get_value( $values[ $field_counter ], $entry_id, 'condition' );
		}

		$field_id               = '';
		$pattern_for_fieldtype  = '/' . preg_quote( $ARFPDF_FIELDTYPE_SHORTCODE, '/' ) . '/';
		$is_fieldtype_available = preg_match_all( $pattern_for_fieldtype, $singlecontent, $content_fieldtypes_arr );
		if ( $is_fieldtype_available > 0 ) {
			foreach ( $content_fieldtypes_arr[0] as $content_fieldtype_arr ) {
				$parts = ( ! empty( $ARFPDF_FIELDTYPE_SHORTCODE ) && ! empty( $singlecontent ) ) ? explode( $ARFPDF_FIELDTYPE_SHORTCODE, $singlecontent ) : '';

				$myidpart = ( ! empty( $ARFPDF_FIELD_SHORTCODE ) && ! empty( $parts[1] ) ) ? explode( $ARFPDF_FIELD_SHORTCODE, $parts[1] ) : '';
				$myid     = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';
				if ( count( $myid ) > 1 ) {
					$field_id = $myid[0];
					if ( strtolower( $field_id ) == $ARFPDF_LOOP_VAR_ONLY && $field_counter > 0 ) {
						$field_id = $field_counter;
					}
					$get_fieldtype_value = $arformspdfcreator->get_fieldtype( $values[ $field_id ] );
					$singlecontent       = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, $get_fieldtype_value, $singlecontent );
					$PARSED_FIELD_ID     = $field_id;
				} else {
					$singlecontent = str_replace( $ARFPDF_FIELDTYPE_SHORTCODE . $ARFPDF_FIELD_SHORTCODE . $myid[0] . $ARFPDF_SHORTCODE_OVER, '', $singlecontent );
				}
			}
			$values_field_counter = $arformspdfcreator->get_fieldtype( $values[ $field_counter ] );
		}

		$pattern_for_user_id = '/' . preg_quote( $ARFPDF_USERID_SHORTCODE, '/' ) . '/';
		$is_user_id          = preg_match_all( $pattern_for_user_id, $singlecontent, $content_userid_arr );

		if ( $is_user_id > 0 ) {
			foreach ( $content_userid_arr[0] as $userid_arr ) {
				$singlecontent = str_replace( $ARFPDF_USERID_SHORTCODE, $entry_user_id, $singlecontent );
			}
		}

		$pattern_for_user_email = '/' . preg_quote( $ARFPDF_USEREMAIL_SHORTCODE, '/' ) . '/';
		$is_user_emai           = preg_match_all( $pattern_for_user_email, $singlecontent, $content_useremail_arr );

		if ( $is_user_emai > 0 ) {
			foreach ( $content_useremail_arr[0] as $useremail_arr ) {
				$singlecontent = str_replace( $ARFPDF_USEREMAIL_SHORTCODE, $entry_user_email, $singlecontent );
			}
		}

		$pattern_for_user_name = '/' . preg_quote( $ARFPDF_USERNAME_SHORTCODE, '/' ) . '/';
		$is_user_name          = preg_match_all( $pattern_for_user_name, $singlecontent, $content_username_arr );

		if ( $is_user_name > 0 ) {
			foreach ( $content_username_arr[0] as $username_arr ) {
				$singlecontent = str_replace( $ARFPDF_USERNAME_SHORTCODE, $entry_user_name, $singlecontent );
			}
		}

		$pattern_for_user_display_name = '/' . preg_quote( $ARFPDF_USERDISPLAYNAME_SHORTCODE, '/' ) . '/';
		$is_user_display_name          = preg_match_all( $pattern_for_user_display_name, $singlecontent, $content_userdisplayname_arr );

		if ( $is_user_display_name > 0 ) {
			foreach ( $content_userdisplayname_arr[0] as $userdisplayname_arr ) {
				$singlecontent = str_replace( $ARFPDF_USERDISPLAYNAME_SHORTCODE, $entry_user_display_name, $singlecontent );
			}
		}

		if ( $field_counter > 0 ) {
			$field_id           = '';
			$pattern_for_module = '/' . preg_quote( $ARFPDF_MODULE_SHORTCODE, '/' ) . '/';

			$singlecontent_module = strtolower( $singlecontent );

			$is_module_available = preg_match_all( $pattern_for_module, $singlecontent_module, $content_modules_arr );

			if ( $is_module_available > 0 ) {
				foreach ( $content_modules_arr[0] as $content_module_arr ) {
					$parts = ( ! empty( $ARFPDF_MODULE_SHORTCODE ) && ! empty( $singlecontent_module ) ) ? explode( $ARFPDF_MODULE_SHORTCODE, $singlecontent_module ) : '';

					$myidpart = ( ! empty( $parts[1] ) ) ? explode( ' ', $parts[1] ) : '';

					$myid = ( ! empty( $myidpart[1] ) ) ? explode( '}', $myidpart[1] ) : '';

					if ( count( $myid ) > 1 ) {
						$field_id = trim( $myid[0] );

						$chkforcondition_false = explode( '!=', $field_id );
						$chkforcondition_true  = explode( '==', $field_id );

						if ( count( $chkforcondition_false ) > 1 ) {
							   $modultewith          = $chkforcondition_false[0];
							   $modultewith_val      = $chkforcondition_false[1];
							   $calculate_module_val = $field_counter % $modultewith;

							   $singlecontent = $calculate_module_val . '!=' . $modultewith_val;
						} else {
							$modultewith          = $chkforcondition_true[0];
							$modultewith_val      = $chkforcondition_true[1];
							$calculate_module_val = $field_counter % $modultewith;

							$singlecontent = $calculate_module_val . '==' . $modultewith_val;
						}
					} else {
						$singlecontent = str_replace( $pattern_for_module, '', $singlecontent );
					}
				}
			}
		}

		if ( $singlecontent != '' ) {
			$exp_singlecontent     = explode( '==', $singlecontent );
			$exp_singlecontent_neq = explode( '!=', $singlecontent );
			if ( isset( $exp_singlecontent[0] ) && $exp_singlecontent[0] == '-' ) {
				$exp_singlecontent[0] = '';
			}
			if ( isset( $exp_singlecontent_neq[0] ) && $exp_singlecontent_neq[0] == '-' ) {
				$exp_singlecontent_neq[0] = '';
			}
			if ( count( $exp_singlecontent ) > 1 ) {
				if ( $exp_singlecontent[0] == $exp_singlecontent[1] ) {

					return 1;
				}
			} elseif ( count( $exp_singlecontent_neq ) > 1 ) {
				if ( $exp_singlecontent_neq[0] != $exp_singlecontent_neq[1] ) {
					return 1;
				}
			} elseif ( count( $exp_singlecontent ) <= 1 && count( $exp_singlecontent_neq ) <= 1 && $field_counter > 0 ) {
				if ( $PARSED_FIELD_ID == $field_counter ) {
					return 1;
				}
			}
		}
		return 0;
	}

	function arf_pdf_get_template_header( $template_header, $font_family ) {

		global $arf_pdfcreator;

		$header = array(
			'L'    => array(
				'content'     => $template_header['left'],
				'font-size'   => 10,
				'font-style'  => 'B',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'C'    => array(
				'content'     => $template_header['center'],
				'font-size'   => 18,
				'font-style'  => 'B',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'R'    => array(
				'content'     => $template_header['right'],
				'font-size'   => 10,
				'font-style'  => 'B',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'line' => 1,
		);

		$arf_pdfcreator->DefHeaderByName( 'ARForms PDFCreator Header', $header );

		$arf_pdfcreator->SetHeaderByName( 'ARForms PDFCreator Header' );
	}

	function arf_pdf_get_template_footer( $form, $template_footer, $font_family ) {
		global $arf_pdfcreator, $ARFPDF_SIGNATURE_SHORTCODE_START, $ARFPDF_SIGNATURE_SHORTCODE_END;

		$delimiter        = '#';
		$signaturematches = $delimiter . preg_quote( $ARFPDF_SIGNATURE_SHORTCODE_START, $delimiter )
		. '(.*?)(.*?)'
		. preg_quote( $ARFPDF_SIGNATURE_SHORTCODE_END, $delimiter )
		. $delimiter
		. 's';
		$signaturevalue   = preg_match_all( $signaturematches, $content, $signaturevalues );

		$powered_by_content = 'Powered by ARForms';
		if ( $signaturevalue > 0 ) {
			foreach ( $signaturevalues[2] as $k => $v ) {
				$powered_by_content = $v;
			}
		}

		$powered_by_content = str_replace( array( "\n", '<br />', '<br>', '<br/>', '  ' ), ' ', $powered_by_content );

		$footer = array(
			'L'    => array(
				'content'     => $template_footer['left'],
				'font-size'   => 10,
				'font-style'  => 'B',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'C'    => array(
				'content'     => $template_footer['center'],
				'font-size'   => 18,
				'font-style'  => 'B',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'R'    => array(
				'content'     => $template_footer['right'],
				'font-size'   => 10,
				'font-style'  => 'R',
				'font-family' => $font_family,
				'color'       => '#000000',
			),
			'line' => 1,
		);

		$arf_pdfcreator->DefFooterByName( 'ARForms PDFCreator Footer', $footer );

		$arf_pdfcreator->SetFooterByName( 'ARForms PDFCreator Footer' );
	}

	function is_arforms_version() {
		if ( version_compare( $this->get_arforms_version(), '3.3.2', '>=' ) ) {
			return true;
		} else {
			return false;
		}
	}

	function arforms_pdfcreator_templates() {
		$frm_templates = array();

		$frm_templates['template_default']                        = array();
		$frm_templates['template_default']['template_header']     = array(
			'left'   => '',
			'center' => '{ARF_FormTitle}',
			'right'  => '',
		);
		$frm_templates['template_default']['template_footer']     = array(
			'left'   => '',
			'center' => '',
			'right'  => '',
		);
		$frm_templates['template_default']['template_settings']   = array(
			'arf_pdf_template_size'        => 'A4',
			'arf_pdf_template_orientation' => 'portrait',
			'arf_pdf_template_font'        => 'arf_arial',
			'arf_pdf_pgno'                 => 1,
		);
		$frm_templates['template_default']['template_header_img'] = array(
			'pdf_header_img_name'  => '',
			'pdf_header_img_align' => 'center',
		);
		$frm_templates['template_default']['template_content']    = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
				<html>
				<head>
				    <title>ARForms PDFcreator</title>
				    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
				    <link rel="stylesheet" href="' . ARF_PDFCREATOR_URL . '/css/arformspdfcreator.css" type="text/css" />
				</head>
				    <body>
				       <div class="arf_pdf_container">    	
				            <div class="arf_field_values">
				            	<table class="form-table">
				                [ARF_LOOP]
				                    <tr class="arf_pdf_viewentry_row [ARF_IF {VAR MOD 2==0}] two_row_even [/ARF_IF] [ARF_IF {VAR MOD 2==1}] two_row_odd [/ARF_IF]">
				                        <td class="arf_pdf_viewentry_left">&nbsp;{ARF_GetLabel:VAR}</td>
				                        <td class="arf_pdf_viewentry_right">&nbsp;{ARF_GetValue:VAR}</td>
				                    </tr>
				                [/ARF_LOOP]
				                </table>
				            </div>
				        </div>
				    </body>
				</html>';

		$frm_templates['template_two_row']                    = array();
		$frm_templates['template_two_row']['template_header'] = array(
			'left'   => '',
			'center' => '{ARF_FormTitle}',
			'right'  => '',
		);

		$frm_templates['template_two_row']['template_footer']     = array(
			'left'   => '',
			'center' => '',
			'right'  => '',
		);
		$frm_templates['template_two_row']['template_settings']   = array(
			'arf_pdf_template_size'        => 'A4',
			'arf_pdf_template_orientation' => 'portrait',
			'arf_pdf_template_font'        => 'arf_arial',
			'arf_pdf_pgno'                 => 1,
		);
		$frm_templates['template_two_row']['template_header_img'] = array(
			'pdf_header_img_name'  => '',
			'pdf_header_img_align' => 'center',
		);

		$frm_templates['template_two_row']['template_content'] = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
				<html>
				<head>    
				    <title>ARForms PDFcreator</title>
				    <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
				    <link rel="stylesheet" href="' . ARF_PDFCREATOR_URL . '/css/arformspdfcreator.css" type="text/css" />
				</head>
				    <body>
				        <div class="arf_pdf_container">
				            <div class="arf_field_values">
				                [ARF_LOOP]
				                    <div class="two_row_arfviewentry_row [ARF_IF {VAR MOD 2==0}] two_row_even [/ARF_IF] [ARF_IF {VAR MOD 2==1}] two_row_odd [/ARF_IF]">
				                        <div class="two_row_arfviewentry_left">&nbsp;{ARF_GetLabel:VAR}</div>
				                        <div class="two_row_arfviewentry_right">&nbsp;{ARF_GetValue:VAR}</div>
				                    </div>
				                [/ARF_LOOP]
				            </div>
				        </div>
				    </body>
				</html>';

		return $frm_templates;
	}

	function arforms_pdf_get_fonts_arr() {
		$font_arr = array(
			'aboriginalsans'        => array( 'AboriginalSansREGULAR' => array( 'AboriginalSansREGULAR.ttf' ) ),
			'abyssinicasil'         => array( 'Abyssinica_SIL' => array( 'Abyssinica_SIL.ttf' ) ),
			'aegean'                => array( 'Aegean' => array( 'Aegean.otf' ) ),
			'aegyptus'              => array( 'Aegyptus' => array( 'Aegyptus.otf' ) ),
			'akkadian'              => array( 'Akkadian' => array( 'Akkadian.otf' ) ),
			'ayar'                  => array( 'Ayar' => array( 'ayar.ttf' ) ),
			'mph2bdamase'           => array( 'Damase' => array( 'damase_v.2.ttf' ) ),
			'daibannasilbook'       => array( 'DBSILBR' => array( 'DBSILBR.ttf' ) ),
			'dejavusans'            => array( 'DejaVuSans' => array( 'DejaVuSans.ttf', 'DejaVuSans-Bold.ttf', 'DejaVuSans-Oblique.ttf', 'DejaVuSans-BoldOblique.ttf' ) ),
			'dejavusanscondensed'   => array( 'DejaVuSansCondensed' => array( 'DejaVuSansCondensed.ttf', 'DejaVuSansCondensed-Bold.ttf', 'DejaVuSansCondensed-Oblique.ttf', 'DejaVuSansCondensed-BoldOblique.ttf' ) ),
			'dejavusansmono'        => array( 'DejaVuSansMono' => array( 'DejaVuSansMono.ttf', 'DejaVuSansMono-Bold.ttf', 'DejaVuSansMono-Oblique.ttf', 'DejaVuSansMono-BoldOblique.ttf' ) ),
			'dejavuserif'           => array( 'DejaVuSerif' => array( 'DejaVuSerif.ttf', 'DejaVuSerif-Bold.ttf', 'DejaVuSerif-Italic.ttf', 'DejaVuSerif-BoldItalic.ttf' ) ),
			'dejavuserifcondensed'  => array( 'DejaVuSerifCondensed' => array( 'DejaVuSerifCondensed.ttf', 'DejaVuSerifCondensed-Bold.ttf', 'DejaVuSerifCondensed-Italic.ttf', 'DejaVuSerifCondensed-BoldItalic.ttf' ) ),
			'dhyana'                => array( 'Dhyana' => array( 'Dhyana-Regular.ttf', 'Dhyana-Bold.ttf' ) ),
			'freemono'              => array( 'FreeMono' => array( 'FreeMono.ttf', 'FreeMonoBold.ttf', 'FreeMonoOblique.ttf', 'FreeMonoBoldOblique.ttf' ) ),
			'freesans'              => array( 'FreeSans' => array( 'FreeSans.ttf', 'FreeSansBold.ttf', 'FreeSansOblique.ttf', 'FreeSansBoldOblique.ttf' ) ),
			'freeserif'             => array( 'FreeSerif' => array( 'FreeSerif.ttf', 'FreeSerifBold.ttf', 'FreeSerifItalic.ttf', 'FreeSerifBoldItalic.ttf' ) ),
			'garuda'                => array( 'Garuda' => array( 'Garuda.ttf', 'Garuda-Bold.ttf', 'Garuda-Oblique.ttf', 'Garuda-BoldOblique.ttf' ) ),
			'jomolhari'             => array( 'Jomolhari' => array( 'Jomolhari.ttf' ) ),
			'kaputaunicode'         => array( 'kaputaunicode' => array( 'kaputaunicode.ttf' ) ),
			'khmeros'               => array( 'KhmerOS' => array( 'KhmerOS.ttf' ) ),
			'lannaalif'             => array( 'lannaalif' => array( 'lannaalif-v1-03.ttf' ) ),
			'lateef'                => array( 'LateefRegOT' => array( 'LateefRegOT.ttf' ) ),
			'lohitkannada'          => array( 'Lohit-Kannada' => array( 'Lohit-Kannada.ttf' ) ),
			'ocrb'                  => array( 'ocrb' => array( 'ocrb10.ttf' ) ),
			'padaukbook'            => array( 'Padauk-book' => array( 'Padauk-book.ttf' ) ),
			'ocrb'                  => array( 'ocrb10' => array( 'ocrb10.ttf' ) ),
			'pothana2000'           => array( 'Pothana2000' => array( 'Pothana2000.ttf' ) ),
			'quivira'               => array( 'Quivira' => array( 'Quivira.otf' ) ),
			'sundaneseunicode'      => array( 'SundaneseUnicode' => array( 'SundaneseUnicode-1.0.5.ttf' ) ),
			'sun-exta'              => array( 'Sun-ExtA' => array( 'Sun-ExtA.ttf' ) ),
			'sun-extb'              => array( 'Sun-ExtB' => array( 'Sun-ExtB.ttf' ) ),
			'estrangeloedessa'      => array( 'SyrCOMEdessa' => array( 'SyrCOMEdessa.otf' ) ),
			'taameydavidclm'        => array( 'TaameyDavidCLM-Medium' => array( 'TaameyDavidCLM-Medium.ttf' ) ),
			'taiheritagepro'        => array( 'TaiHeritagePro' => array( 'TaiHeritagePro.ttf' ) ),
			'tharlon'               => array( 'Tharlon' => array( 'Tharlon-Regular.ttf' ) ),
			'unbatang'              => array( 'UnBatang_0613' => array( 'UnBatang_0613.ttf' ) ),
			'kfgqpcuthmantahanaskh' => array( 'Uthman' => array( 'Uthman.otf' ) ),
			'xbriyaz'               => array( 'XB Riyaz' => array( 'XB Riyaz.ttf', 'XB RiyazBd.ttf', 'XB RiyazIt.ttf', 'XB RiyazBdIt.ttf' ) ),
			'zawgyi-one'            => array( 'ZawgyiOne' => array( 'ZawgyiOne.ttf' ) ),
		);

		return $font_arr;
	}

	function arforms_pdf_get_default_fonts_arr() {
		$font_arr = array(
			'arf_arial'               => array( 'Arial' => array( 'arial.ttf' ) ),
			'arf_courier_new'         => array( 'Courier New' => array( 'cour_new.ttf' ) ),
			'arf_courier'             => array( 'Courier' => array( 'Courier.ttf' ) ),
			'arf_geneva'              => array( 'Geneva' => array( 'Geneva.ttf' ) ),
			'arf_helvetica'           => array( 'Helvetica' => array( 'HELVETICA.ttf' ) ),
			'arf_lucida_grande'       => array( 'Lucida Grande' => array( 'LucidaGrande.ttf', 'LucidaGrandeBold.ttf' ) ),
			'arf_lucida_sans_unicode' => array( 'Lucida Sans Unicode' => array( 'lucida-sans-unicode.ttf' ) ),
			'arf_monospace'           => array( 'Monospace' => array( 'Monospace.ttf' ) ),
			'arf_sansserif'           => array( 'Sans-Serif' => array( 'OpenSans-Regular.ttf' ) ),
			'arf_tahoma'              => array( 'Tahoma' => array( 'Tahoma.ttf' ) ),
			'arf_timesnewroman'       => array( 'Times New Roman' => array( 'TIMESR.ttf' ) ),
			'arf_verdana'             => array( 'Verdana' => array( 'Verdana.ttf' ) ),
		);
		return $font_arr;
	}

}

global $arf_pdfcreator_api_url, $arf_pdfcreator_plugin_slug;

$arf_pdfcreator_api_url     = $arformspdfcreator->arf_pdfcreator_getapiurl();
$arf_pdfcreator_plugin_slug = basename( dirname( __FILE__ ) );

add_filter( 'pre_set_site_transient_update_plugins', 'arf_pdfcreator_check_for_plugin_update' );

function arf_pdfcreator_check_for_plugin_update( $checked_data ) {
	global $arf_pdfcreator_api_url, $arf_pdfcreator_plugin_slug, $wp_version, $arformspdfcreator, $arformspdfcreator_version;

	if ( empty( $checked_data->checked ) ) {
		return $checked_data;
	}

	$args = array(
		'slug'            => $arf_pdfcreator_plugin_slug,
		'version'         => $arformspdfcreator_version,
		'other_variables' => $arformspdfcreator->arf_pdfcreator_get_remote_post_params(),
	);

	$request_string = array(
		'body'       => array(
			'action'  => 'basic_check',
			'request' => serialize( $args ),
			'api-key' => md5( home_url() ),
		),
		'user-agent' => 'ARFPDFCRTR-WordPress/' . $wp_version . '; ' . home_url(),
	);

	$raw_response = wp_remote_post( $arf_pdfcreator_api_url, $request_string );

	if ( ! is_wp_error( $raw_response ) && ( $raw_response['response']['code'] == 200 ) ) {
		$response = isset( $raw_response['body'] ) ? unserialize( $raw_response['body'] ) : '';
	}

	if ( isset( $response ) && ! empty( $response ) && isset( $response->token ) && $response->token != '' ) {
		update_option( 'arfpdfcreator_update_token', $response->token );
	}

	if ( isset( $response ) && is_object( $response ) && is_object( $checked_data ) && ! empty( $response ) ) {
		$checked_data->response[ $arf_pdfcreator_plugin_slug . '/' . $arf_pdfcreator_plugin_slug . '.php' ] = $response;
	}

	return $checked_data;
}

add_filter( 'plugins_api', 'arf_pdfcreator_plugin_api_call', 10, 3 );

function arf_pdfcreator_plugin_api_call( $def, $action, $args ) {
	global $arf_pdfcreator_plugin_slug, $arf_pdfcreator_api_url, $wp_version;

	if ( ! isset( $args->slug ) || ( $args->slug != $arf_pdfcreator_plugin_slug ) ) {
		return false;
	}

	$plugin_info     = get_site_transient( 'update_plugins' );
	$current_version = $plugin_info->checked[ $arf_pdfcreator_plugin_slug . '/' . $arf_pdfcreator_plugin_slug . '.php' ];
	$args->version   = $current_version;

	$request_string = array(
		'body'       => array(
			'action'       => $action,
			'update_token' => get_site_option( 'arfpdfcreator_update_token' ),
			'request'      => serialize( $args ),
			'api-key'      => md5( home_url() ),
		),
		'user-agent' => 'ARFPDFCRTR-WordPress/' . $wp_version . '; ' . home_url(),
	);

	$request = wp_remote_post( $arf_pdfcreator_api_url, $request_string );

	if ( is_wp_error( $request ) ) {
		$res = new WP_Error( 'plugins_api_failed', 'An Unexpected HTTP Error occurred during the API request.</p> <p><a href="?" onclick="document.location.reload(); return false;">Try again</a>', $request->get_error_message() );
	} else {
		$res = maybe_unserialize( $request['body'] );

		if ( $res === false ) {
			$res = new WP_Error( 'plugins_api_failed', 'An unknown error occurred', $request['body'] );
		}
	}

	return $res;
}
?>