<?php

class grad
{

    var $mpdf = null;

    public function __construct(ARF_MPDF $mpdf)
    {
        $this->mpdf = $mpdf;
    }

    
    function CoonsPatchMesh($x, $y, $w, $h, $patch_array = array(), $x_min = 0, $x_max = 1, $y_min = 0, $y_max = 1, $colspace = 'RGB', $return = false)
    {
        $s = ' q ';
        $s.=sprintf(' %.3F %.3F %.3F %.3F re W n ', $x * _MPDFK, ($this->mpdf->h - $y) * _MPDFK, $w * _MPDFK, -$h * _MPDFK);
        $s.=sprintf(' %.3F 0 0 %.3F %.3F %.3F cm ', $w * _MPDFK, $h * _MPDFK, $x * _MPDFK, ($this->mpdf->h - ($y + $h)) * _MPDFK);
        $n = count($this->mpdf->gradients) + 1;
        $this->mpdf->gradients[$n]['type'] = 6; 
        $this->mpdf->gradients[$n]['colorspace'] = $colspace; 
        $bpcd = 65535; 
        $trans = false;
        $this->mpdf->gradients[$n]['stream'] = '';
        for ($i = 0; $i < count($patch_array); $i++) {
            $this->mpdf->gradients[$n]['stream'].=chr($patch_array[$i]['f']); 
            for ($j = 0; $j < count($patch_array[$i]['points']); $j++) {
                
                if (($j % 2) == 1) { 
                    $patch_array[$i]['points'][$j] = (($patch_array[$i]['points'][$j] - $y_min) / ($y_max - $y_min)) * $bpcd;
                    $patch_array[$i]['points'][$j] = $bpcd - $patch_array[$i]['points'][$j];
                } else {
                    $patch_array[$i]['points'][$j] = (($patch_array[$i]['points'][$j] - $x_min) / ($x_max - $x_min)) * $bpcd;
                }
                if ($patch_array[$i]['points'][$j] < 0) {
                    $patch_array[$i]['points'][$j] = 0;
                }
                if ($patch_array[$i]['points'][$j] > $bpcd) {
                    $patch_array[$i]['points'][$j] = $bpcd;
                }
                $this->mpdf->gradients[$n]['stream'].=chr(floor($patch_array[$i]['points'][$j] / 256));
                $this->mpdf->gradients[$n]['stream'].=chr(floor($patch_array[$i]['points'][$j] % 256));
            }
            for ($j = 0; $j < count($patch_array[$i]['colors']); $j++) {
                
                if ($colspace == 'RGB') {
                    $this->mpdf->gradients[$n]['stream'].=($patch_array[$i]['colors'][$j][1]);
                    $this->mpdf->gradients[$n]['stream'].=($patch_array[$i]['colors'][$j][2]);
                    $this->mpdf->gradients[$n]['stream'].=($patch_array[$i]['colors'][$j][3]);
                    if (isset($patch_array[$i]['colors'][$j][4]) && ord($patch_array[$i]['colors'][$j][4]) < 100) {
                        $trans = true;
                    }
                } else if ($colspace == 'CMYK') {
                    $this->mpdf->gradients[$n]['stream'].=chr(ord($patch_array[$i]['colors'][$j][1]) * 2.55);
                    $this->mpdf->gradients[$n]['stream'].=chr(ord($patch_array[$i]['colors'][$j][2]) * 2.55);
                    $this->mpdf->gradients[$n]['stream'].=chr(ord($patch_array[$i]['colors'][$j][3]) * 2.55);
                    $this->mpdf->gradients[$n]['stream'].=chr(ord($patch_array[$i]['colors'][$j][4]) * 2.55);
                    if (isset($patch_array[$i]['colors'][$j][5]) && ord($patch_array[$i]['colors'][$j][5]) < 100) {
                        $trans = true;
                    }
                } else if ($colspace == 'Gray') {
                    $this->mpdf->gradients[$n]['stream'].=($patch_array[$i]['colors'][$j][1]);
                    if ($patch_array[$i]['colors'][$j][2] == 1) {
                        $trans = true;
                    } 
                }
            }
        }
        
        if ($trans) {
            $this->mpdf->gradients[$n]['stream_trans'] = '';
            for ($i = 0; $i < count($patch_array); $i++) {
                $this->mpdf->gradients[$n]['stream_trans'].=chr($patch_array[$i]['f']);
                for ($j = 0; $j < count($patch_array[$i]['points']); $j++) {
                    
                    $this->mpdf->gradients[$n]['stream_trans'].=chr(floor($patch_array[$i]['points'][$j] / 256));
                    $this->mpdf->gradients[$n]['stream_trans'].=chr(floor($patch_array[$i]['points'][$j] % 256));
                }
                for ($j = 0; $j < count($patch_array[$i]['colors']); $j++) {

                    if ($colspace == 'RGB') {
                        $this->mpdf->gradients[$n]['stream_trans'].=chr(intval(ord($patch_array[$i]['colors'][$j][4]) * 2.55));
                    } else if ($colspace == 'CMYK') {
                        $this->mpdf->gradients[$n]['stream_trans'].=chr(intval(ord($patch_array[$i]['colors'][$j][5]) * 2.55));
                    } else if ($colspace == 'Gray') {
                        $this->mpdf->gradients[$n]['stream_trans'].=chr(intval(ord($patch_array[$i]['colors'][$j][3]) * 2.55));
                    }
                }
            }
            $this->mpdf->gradients[$n]['trans'] = true;
            $s .= ' /TGS' . $n . ' gs ';
        }
        
        $s .= '/Sh' . $n . ' sh' . "\n";
        
        $s .= 'Q' . "\n";
        if ($return) {
            return $s;
        } else {
            $this->mpdf->_out($s);
        }
    }

    
    function Gradient($x, $y, $w, $h, $type, $stops = array(), $colorspace = 'RGB', $coords = '', $extend = '', $return = false, $is_mask = false)
    {
        if (strtoupper(substr($type, 0, 1)) == 'L') {
            $type = 2;
        } 
        else if (strtoupper(substr($type, 0, 1)) == 'R') {
            $type = 3;
        }
        if ($colorspace != 'CMYK' && $colorspace != 'Gray') {
            $colorspace = 'RGB';
        }
        $bboxw = $w;
        $bboxh = $h;
        $usex = $x;
        $usey = $y;
        $usew = $bboxw;
        $useh = $bboxh;

        if ($type < 1) {
            $type = 2;
        }
        if ($coords[0] !== false && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $coords[0], $m)) {
            $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
            if ($tmp) {
                $coords[0] = $tmp / $w;
            }
        }
        if ($coords[1] !== false && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $coords[1], $m)) {
            $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
            if ($tmp) {
                $coords[1] = 1 - ($tmp / $h);
            }
        }
        
        if ($type == 2) {
            $angle = (isset($coords[4]) ? $coords[4] : false);
            $repeat = (isset($coords[5]) ? $coords[5] : false);
            
            if ($coords[0] !== false && $coords[1] !== false && $coords[2] !== false && $coords[3] !== false) {
                
            }

        
            else if ($angle !== false && $coords[0] !== false && $coords[1] !== false && $coords[2] === false && $coords[3] === false) {
                if ($angle == 0 || $angle == 360) {
                    $coords[3] = $coords[1];
                    if ($coords[0] == 1) {
                        $coords[2] = 2;
                    } else {
                        $coords[2] = 1;
                    }
                }
                else if ($angle == 90) {
                    $coords[2] = $coords[0];
                    $coords[3] = 1;
                    if ($coords[1] == 1) {
                        $coords[3] = 2;
                    } else {
                        $coords[3] = 1;
                    }
                }
                else if ($angle == 180) {
                    if ($coords[4] == 0) {
                        $coords[2] = -1;
                    } else {
                        $coords[2] = 0;
                    }
                    $coords[3] = $coords[1];
                }
                else if ($angle == 270) {
                    $coords[2] = $coords[0];
                    if ($coords[1] == 0) {
                        $coords[3] = -1;
                    } else {
                        $coords[3] = 0;
                    }
                }
                else {
                    $endx = 1;
                    $endy = 1;
                    if ($angle <= 90) {
                        if ($angle <= 45) {
                               $endy = tan(deg2rad($angle));
                        } else {
                            $endx = tan(deg2rad(90 - $angle));
                        }
                        $b = atan2(($endy * $bboxh), ($endx * $bboxw));
                        $ny = 1 - $coords[1] - (tan($b) * (1 - $coords[0]));
                        $tx = sin($b) * cos($b) * $ny;
                        $ty = cos($b) * cos($b) * $ny;
                        $coords[2] = 1 + $tx;
                        $coords[3] = 1 - $ty;
                    } else if ($angle <= 180) {
                        if ($angle <= 135) {
                            $endx = tan(deg2rad($angle - 90));
                        } else {
                            $endy = tan(deg2rad(180 - $angle));
                        }
                        $b = atan2(($endy * $bboxh), ($endx * $bboxw));
                        $ny = 1 - $coords[1] - (tan($b) * ($coords[0]));
                        $tx = sin($b) * cos($b) * $ny;
                        $ty = cos($b) * cos($b) * $ny;
                        $coords[2] = -$tx;
                        $coords[3] = 1 - $ty;
                    } else if ($angle <= 270) {
                        if ($angle <= 225) {
                            $endy = tan(deg2rad($angle - 180));
                        } else {
                            $endx = tan(deg2rad(270 - $angle));
                        }
                        $b = atan2(($endy * $bboxh), ($endx * $bboxw));
                        $ny = $coords[1] - (tan($b) * ($coords[0]));
                        $tx = sin($b) * cos($b) * $ny;
                        $ty = cos($b) * cos($b) * $ny;
                        $coords[2] = -$tx;
                        $coords[3] = $ty;
                    } else {
                        if ($angle <= 315) {
                            $endx = tan(deg2rad($angle - 270));
                        } else {
                            $endy = tan(deg2rad(360 - $angle));
                        }
                        $b = atan2(($endy * $bboxh), ($endx * $bboxw));
                        $ny = $coords[1] - (tan($b) * (1 - $coords[0]));
                        $tx = sin($b) * cos($b) * $ny;
                        $ty = cos($b) * cos($b) * $ny;
                        $coords[2] = 1 + $tx;
                        $coords[3] = $ty;
                    }
                }
            }

            
            else if ($angle !== false && $coords[0] === false && $coords[1] === false) {
                if ($angle == 0 || $angle == 360) {
                    $coords[0] = 0;
                    $coords[1] = 0;
                    $coords[2] = 1;
                    $coords[3] = 0;
                } else if ($angle == 90) {
                    $coords[0] = 0;
                    $coords[1] = 0;
                    $coords[2] = 0;
                    $coords[3] = 1;
                } else if ($angle == 180) {
                    $coords[0] = 1;
                    $coords[1] = 0;
                    $coords[2] = 0;
                    $coords[3] = 0;
                } else if ($angle == 270) {
                    $coords[0] = 0;
                    $coords[1] = 1;
                    $coords[2] = 0;
                    $coords[3] = 0;
                } else {
                    if ($angle <= 90) {
                        $coords[0] = 0;
                        $coords[1] = 0;
                        if ($angle <= 45) {
                               $endx = 1;
                               $endy = tan(deg2rad($angle));
                        } else {
                            $endx = tan(deg2rad(90 - $angle));
                            $endy = 1;
                        }
                    } else if ($angle <= 180) {
                        $coords[0] = 1;
                        $coords[1] = 0;
                        if ($angle <= 135) {
                            $endx = tan(deg2rad($angle - 90));
                            $endy = 1;
                        } else {
                            $endx = 1;
                            $endy = tan(deg2rad(180 - $angle));
                        }
                    } else if ($angle <= 270) {
                        $coords[0] = 1;
                        $coords[1] = 1;
                        if ($angle <= 225) {
                            $endx = 1;
                            $endy = tan(deg2rad($angle - 180));
                        } else {
                            $endx = tan(deg2rad(270 - $angle));
                            $endy = 1;
                        }
                    } else {
                        $coords[0] = 0;
                        $coords[1] = 1;
                        if ($angle <= 315) {
                            $endx = tan(deg2rad($angle - 270));
                            $endy = 1;
                        } else {
                            $endx = 1;
                            $endy = tan(deg2rad(360 - $angle));
                        }
                    }
                    $b = atan2(($endy * $bboxh), ($endx * $bboxw));
                    $h2 = $bboxh - ($bboxh * tan($b));
                    $px = $bboxh + ($h2 * sin($b) * cos($b));
                    $py = ($bboxh * tan($b)) + ($h2 * sin($b) * sin($b));
                    $x1 = $px / $bboxh;
                    $y1 = $py / $bboxh;
                    if ($angle <= 90) {
                        $coords[2] = $x1;
                        $coords[3] = $y1;
                    } else if ($angle <= 180) {
                        $coords[2] = 1 - $x1;
                        $coords[3] = $y1;
                    } else if ($angle <= 270) {
                        $coords[2] = 1 - $x1;
                        $coords[3] = 1 - $y1;
                    } else {
                        $coords[2] = $x1;
                        $coords[3] = 1 - $y1;
                    }
                }
            }
            
            else if ((!isset($angle) || $angle === false) && $coords[0] !== false && $coords[1] !== false) {  
                $coords[2] = 1 - $coords[0];
                $coords[3] = 1 - $coords[1];
                $angle = rad2deg(atan2($coords[3] - $coords[1], $coords[2] - $coords[0]));
                if ($angle < 0) {
                    $angle += 360;
                } else if ($angle > 360) {
                    $angle -= 360;
                }
                if ($angle != 0 && $angle != 360 && $angle != 90 && $angle != 180 && $angle != 270) {
                    if ($w >= $h) {
                        $coords[1] *= $h / $w;
                        $coords[3] *= $h / $w;
                        $usew = $useh = $bboxw;
                        $usey -= ($w - $h);
                    } else {
                        $coords[0] *= $w / $h;
                        $coords[2] *= $w / $h;
                        $usew = $useh = $bboxh;
                    }
                }
            }

            
            else {
                $coords = array(0, 0, 1, 0); 
            }
            $s = ' q';
            $s .= sprintf(' %.3F %.3F %.3F %.3F re W n', $x * _MPDFK, ($this->mpdf->h - $y) * _MPDFK, $w * _MPDFK, -$h * _MPDFK) . "\n";
            $s .= sprintf(' %.3F 0 0 %.3F %.3F %.3F cm', $usew * _MPDFK, $useh * _MPDFK, $usex * _MPDFK, ($this->mpdf->h - ($usey + $useh)) * _MPDFK) . "\n";
        }

        
        else if ($type == 3) {
            $radius = (isset($coords[4]) ? $coords[4] : false);
            $angle = (isset($coords[5]) ? $coords[5] : false); 
            $shape = (isset($coords[6]) ? $coords[6] : false);
            $size = (isset($coords[7]) ? $coords[7] : false);
            $repeat = (isset($coords[8]) ? $coords[8] : false);
            
            if ($coords[0] !== false && $coords[1] !== false && $coords[2] !== false && $coords[3] !== false && $coords[4] !== false) {
                
            }
            
            else if ($shape !== false && $size !== false) {
                if ($coords[2] == false) {
                    $coords[2] = $coords[0];
                }
                if ($coords[3] == false) {
                    $coords[3] = $coords[1];
                }
                
                if ($shape == 'ellipse') {
                    $corner1 = sqrt(pow($coords[0], 2) + pow($coords[1], 2));
                    $corner2 = sqrt(pow($coords[0], 2) + pow((1 - $coords[1]), 2));
                    $corner3 = sqrt(pow((1 - $coords[0]), 2) + pow($coords[1], 2));
                    $corner4 = sqrt(pow((1 - $coords[0]), 2) + pow((1 - $coords[1]), 2));
                    if ($size == 'closest-side') {
                        $radius = min($coords[0], $coords[1], (1 - $coords[0]), (1 - $coords[1]));
                    } else if ($size == 'closest-corner') {
                        $radius = min($corner1, $corner2, $corner3, $corner4);
                    } else if ($size == 'farthest-side') {
                        $radius = max($coords[0], $coords[1], (1 - $coords[0]), (1 - $coords[1]));
                    } else {
                        $radius = max($corner1, $corner2, $corner3, $corner4);
                    } 
                }
                
                else if ($shape == 'circle') {
                    if ($w >= $h) {
                        $coords[1] = $coords[3] = ($coords[1] * $h / $w);
                        $corner1 = sqrt(pow($coords[0], 2) + pow($coords[1], 2));
                        $corner2 = sqrt(pow($coords[0], 2) + pow((($h / $w) - $coords[1]), 2));
                        $corner3 = sqrt(pow((1 - $coords[0]), 2) + pow($coords[1], 2));
                        $corner4 = sqrt(pow((1 - $coords[0]), 2) + pow((($h / $w) - $coords[1]), 2));
                        if ($size == 'closest-side') {
                               $radius = min($coords[0], $coords[1], (1 - $coords[0]), (($h / $w) - $coords[1]));
                        } else if ($size == 'closest-corner') {
                            $radius = min($corner1, $corner2, $corner3, $corner4);
                        } else if ($size == 'farthest-side') {
                            $radius = max($coords[0], $coords[1], (1 - $coords[0]), (($h / $w) - $coords[1]));
                        } else if ($size == 'farthest-corner') {
                            $radius = max($corner1, $corner2, $corner3, $corner4);
                        } 
                        $usew = $useh = $bboxw;
                        $usey -= ($w - $h);
                    } else {
                        $coords[0] = $coords[2] = ($coords[0] * $w / $h);
                        $corner1 = sqrt(pow($coords[0], 2) + pow($coords[1], 2));
                        $corner2 = sqrt(pow($coords[0], 2) + pow((1 - $coords[1]), 2));
                        $corner3 = sqrt(pow((($w / $h) - $coords[0]), 2) + pow($coords[1], 2));
                        $corner4 = sqrt(pow((($w / $h) - $coords[0]), 2) + pow((1 - $coords[1]), 2));
                        if ($size == 'closest-side') {
                            $radius = min($coords[0], $coords[1], (($w / $h) - $coords[0]), (1 - $coords[1]));
                        } else if ($size == 'closest-corner') {
                            $radius = min($corner1, $corner2, $corner3, $corner4);
                        } else if ($size == 'farthest-side') {
                            $radius = max($coords[0], $coords[1], (($w / $h) - $coords[0]), (1 - $coords[1]));
                        } else if ($size == 'farthest-corner') {
                            $radius = max($corner1, $corner2, $corner3, $corner4);
                        } 
                        $usew = $useh = $bboxh;
                    }
                }
                if ($radius == 0) {
                    $radius = 0.001;
                } 
                $coords[4] = $radius;
            }

            
            else { 
                
                $coords = array(0.5, 0.5, 0.5, 0.5);
            }
            $s = ' q';
            $s .= sprintf(' %.3F %.3F %.3F %.3F re W n', $x * _MPDFK, ($this->mpdf->h - $y) * _MPDFK, $w * _MPDFK, -$h * _MPDFK) . "\n";
            $s .= sprintf(' %.3F 0 0 %.3F %.3F %.3F cm', $usew * _MPDFK, $useh * _MPDFK, $usex * _MPDFK, ($this->mpdf->h - ($usey + $useh)) * _MPDFK) . "\n";
        }

        $n = count($this->mpdf->gradients) + 1;
        $this->mpdf->gradients[$n]['type'] = $type;
        $this->mpdf->gradients[$n]['colorspace'] = $colorspace;
        $trans = false;
        $this->mpdf->gradients[$n]['is_mask'] = $is_mask;
        if ($is_mask) {
            $trans = true;
        }
        if (count($stops) == 1) {
            $stops[1] = $stops[0];
        }
        if (!isset($stops[0]['offset'])) {
            $stops[0]['offset'] = 0;
        }
        if (!isset($stops[(count($stops) - 1)]['offset'])) {
            $stops[(count($stops) - 1)]['offset'] = 1;
        }

        
        if ($type == 2) {
            $axisx = ($coords[2] - $coords[0]) * $usew;
            $axisy = ($coords[3] - $coords[1]) * $useh;
            $axis_length = sqrt(pow($axisx, 2) + pow($axisy, 2));
        } else {
            $axis_length = $coords[4] * $usew;
        } 

        for ($i = 0; $i < count($stops); $i++) {
            if (isset($stops[$i]['offset']) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $stops[$i]['offset'], $m)) {
                $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                $stops[$i]['offset'] = $tmp / $axis_length;
            }
        }


        if (isset($stops[0]['offset']) && $stops[0]['offset'] > 0) {
            $firststop = $stops[0];
            $firststop['offset'] = 0;
            array_unshift($stops, $firststop);
        }
        if (!$repeat && isset($stops[(count($stops) - 1)]['offset']) && $stops[(count($stops) - 1)]['offset'] < 1) {
            $endstop = $stops[(count($stops) - 1)];
            $endstop['offset'] = 1;
            $stops[] = $endstop;
        }
        if ($stops[0]['offset'] > $stops[(count($stops) - 1)]['offset']) {
            $stops[0]['offset'] = 0;
            $stops[(count($stops) - 1)]['offset'] = 1;
        }

        for ($i = 0; $i < count($stops); $i++) {
            
            if ($colorspace == 'CMYK') {
                $this->mpdf->gradients[$n]['stops'][$i]['col'] = sprintf('%.3F %.3F %.3F %.3F', (ord($stops[$i]['col']{1}) / 100), (ord($stops[$i]['col']{2}) / 100), (ord($stops[$i]['col']{3}) / 100), (ord($stops[$i]['col']{4}) / 100));
            } else if ($colorspace == 'Gray') {
                $this->mpdf->gradients[$n]['stops'][$i]['col'] = sprintf('%.3F', (ord($stops[$i]['col']{1}) / 255));
            } else {
                $this->mpdf->gradients[$n]['stops'][$i]['col'] = sprintf('%.3F %.3F %.3F', (ord($stops[$i]['col']{1}) / 255), (ord($stops[$i]['col']{2}) / 255), (ord($stops[$i]['col']{3}) / 255));
            }
            if (!isset($stops[$i]['opacity'])) {
                $stops[$i]['opacity'] = 1;
            } else if ($stops[$i]['opacity'] > 1 || $stops[$i]['opacity'] < 0) {
                $stops[$i]['opacity'] = 1;
            } else if ($stops[$i]['opacity'] < 1) {
                $trans = true;
            }
            $this->mpdf->gradients[$n]['stops'][$i]['opacity'] = $stops[$i]['opacity'];
            
            if ($i > 0 && $i < (count($stops) - 1)) {
                if (!isset($stops[$i]['offset']) || (isset($stops[$i + 1]['offset']) && $stops[$i]['offset'] > $stops[$i + 1]['offset']) || $stops[$i]['offset'] < $stops[$i - 1]['offset']) {
                    if (isset($stops[$i - 1]['offset']) && isset($stops[$i + 1]['offset'])) {
                        $stops[$i]['offset'] = ($stops[$i - 1]['offset'] + $stops[$i + 1]['offset']) / 2;
                    } else {
                        for ($j = ($i + 1); $j < count($stops); $j++) {
                            if (isset($stops[$j]['offset'])) {
                                break;
                            }
                        }
                        $int = ($stops[$j]['offset'] - $stops[($i - 1)]['offset']) / ($j - $i + 1);
                        for ($f = 0; $f < ($j - $i - 1); $f++) {
                            $stops[($i + $f)]['offset'] = $stops[($i + $f - 1)]['offset'] + ($int);
                        }
                    }
                }
            }
            $this->mpdf->gradients[$n]['stops'][$i]['offset'] = $stops[$i]['offset'];
            $this->mpdf->gradients[$n]['stops'][$i]['offset'] = $stops[$i]['offset'];
        }

        if ($repeat) {
            $ns = count($this->mpdf->gradients[$n]['stops']);
            $offs = array();
            for ($i = 0; $i < $ns; $i++) {
                $offs[$i] = $this->mpdf->gradients[$n]['stops'][$i]['offset'];
            }
            $gp = 0;
            $inside = true;
            while ($inside) {
                $gp++;
                for ($i = 0; $i < $ns; $i++) {
                    $this->mpdf->gradients[$n]['stops'][(($ns * $gp) + $i)] = $this->mpdf->gradients[$n]['stops'][(($ns * ($gp - 1)) + $i)];
                    $tmp = $this->mpdf->gradients[$n]['stops'][(($ns * ($gp - 1)) + ($ns - 1))]['offset'] + $offs[$i];
                    if ($tmp < 1) {
                        $this->mpdf->gradients[$n]['stops'][(($ns * $gp) + $i)]['offset'] = $tmp;
                    } else {
                        $this->mpdf->gradients[$n]['stops'][(($ns * $gp) + $i)]['offset'] = 1;
                        $inside = false;
                        break(2);
                    }
                }
            }
        }

        if ($trans) {
            $this->mpdf->gradients[$n]['trans'] = true;
            $s .= ' /TGS' . $n . ' gs ';
        }
        if (!is_array($extend) || count($extend) < 1) {
            $extend = array('true', 'true'); 
        }
        $this->mpdf->gradients[$n]['coords'] = $coords;
        $this->mpdf->gradients[$n]['extend'] = $extend;
        
        $s .= '/Sh' . $n . ' sh ' . "\n";
        
        $s .= ' Q ' . "\n";
        if ($return) {
            return $s;
        } else {
            $this->mpdf->_out($s);
        }
    }

    function parseMozGradient($bg)
    {
        
        if (preg_match('/repeating-/', $bg)) {
            $repeat = true;
        } else {
            $repeat = false;
        }
        if (preg_match('/linear-gradient\((.*)\)/', $bg, $m)) {
            $g = array();
            $g['type'] = 2;
            $g['colorspace'] = 'RGB';
            $g['extend'] = array('true', 'true');
            $v = trim($m[1]);
            
            while (preg_match('/(\([^\)]*?),/', $v)) {
                $v = preg_replace('/(\([^\)]*?),/', '\\1@', $v);
            }

            while (preg_match('/(\([^\)]*?)[ ]/', $v)) {
                $v = preg_replace('/(\([^\)]*?)[ ]/', '\\1', $v);
            }
            $bgr = preg_split('/\s*,\s*/', $v);
            for ($i = 0; $i < count($bgr); $i++) {
                $bgr[$i] = preg_replace('/@/', ',', $bgr[$i]);
            }
            
            $first = preg_split('/\s+/', trim($bgr[0]));
            if (preg_match('/(left|center|right|bottom|top|deg|grad|rad)/i', $bgr[0]) && !preg_match('/(<#|rgb|rgba|hsl|hsla)/i', $bgr[0])) {
                $startStops = 1;
            } else if (trim($first[(count($first) - 1)]) === "0") {
                $startStops = 1;
            } else {
                $check = $this->mpdf->ConvertColor($first[0]);
                if ($check) {
                    $startStops = 0;
                } else {
                    $startStops = 1;
                }
            }
            
            if ($startStops == 1) {
                
                if (preg_match('/([\-]*[0-9\.]+)(deg|grad|rad)/i', $bgr[0], $m)) {
                    $angle = $m[1] + 0;
                    if (strtolower($m[2]) == 'deg') {
                        $angle = $angle;
                    } else if (strtolower($m[2]) == 'grad') {
                        $angle *= (360 / 400);
                    } else if (strtolower($m[2]) == 'rad') {
                        $angle = rad2deg($angle);
                    }
                    while ($angle < 0) {
                        $angle += 360;
                    }
                    $angle = ($angle % 360);
                } else if (trim($first[(count($first) - 1)]) === "0") {
                    $angle = 0;
                }
                if (preg_match('/left/i', $bgr[0])) {
                    $startx = 0;
                } else if (preg_match('/right/i', $bgr[0])) {
                    $startx = 1;
                }
                if (preg_match('/top/i', $bgr[0])) {
                    $starty = 1;
                } else if (preg_match('/bottom/i', $bgr[0])) {
                    $starty = 0;
                }
                
                if (preg_match('/(\d+)[%]/i', $first[0], $m)) {
                    $startx = $m[1] / 100;
                } else if (!isset($startx) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $first[0], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    if ($tmp) {
                        $startx = $m[1];
                    }
                }
                if (isset($first[1]) && preg_match('/(\d+)[%]/i', $first[1], $m)) {
                    $starty = 1 - ($m[1] / 100);
                } else if (!isset($starty) && isset($first[1]) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $first[1], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    if ($tmp) {
                        $starty = $m[1];
                    }
                }
                if (isset($startx) && !isset($starty)) {
                    $starty = 0.5;
                }
                if (!isset($startx) && isset($starty)) {
                    $startx = 0.5;
                }
            }
            
            else { 
                $starty = 1;
                $startx = 0.5;
                $endy = 0;
                $endx = 0.5;
            }
            $coords = array();
            if (!isset($startx)) {
                $startx = false;
            }
            if (!isset($starty)) {
                $starty = false;
            }
            if (!isset($endx)) {
                $endx = false;
            }
            if (!isset($endy)) {
                $endy = false;
            }
            if (!isset($angle)) {
                $angle = false;
            }
            $g['coords'] = array($startx, $starty, $endx, $endy, $angle, $repeat);
            $g['stops'] = array();
            for ($i = $startStops; $i < count($bgr); $i++) {
                $stop = array();
                
                $el = preg_split('/\s+/', trim($bgr[$i]));
                
                $col = $this->mpdf->ConvertColor($el[0]);
                if ($col) {
                    $stop['col'] = $col;
                } else {
                    $stop['col'] = $col = $this->mpdf->ConvertColor(255);
                }
                if ($col{0} == 1) {
                    $g['colorspace'] = 'Gray';
                } else if ($col{0} == 4 || $col{0} == 6) {
                    $g['colorspace'] = 'CMYK';
                }
                if ($col{0} == 5) {
                    $stop['opacity'] = ord($col{4}) / 100;
                } 
                else if ($col{0} == 6) {
                    $stop['opacity'] = ord($col{5}) / 100;
                } 
                else if ($col{0} == 1 && $col{2} == 1) {
                    $stop['opacity'] = ord($col{3}) / 100;
                }

                if (isset($el[1]) && preg_match('/(\d+)[%]/', $el[1], $m)) {
                    $stop['offset'] = $m[1] / 100;
                    if ($stop['offset'] > 1) {
                        unset($stop['offset']);
                    }
                } else if (isset($el[1]) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $el[1], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    if ($tmp) {
                        $stop['offset'] = $m[1];
                    }
                }
                $g['stops'][] = $stop;
            }
            if (count($g['stops'])) {
                return $g;
            }
        } else if (preg_match('/radial-gradient\((.*)\)/', $bg, $m)) {
            $g = array();
            $g['type'] = 3;
            $g['colorspace'] = 'RGB';
            $g['extend'] = array('true', 'true');
            $v = trim($m[1]);
            
            while (preg_match('/(\([^\)]*?),/', $v)) {
                $v = preg_replace('/(\([^\)]*?),/', '\\1@', $v);
            }
            
            while (preg_match('/(\([^\)]*?)[ ]/', $v)) {
                $v = preg_replace('/(\([^\)]*?)[ ]/', '\\1', $v);
            }
            $bgr = preg_split('/\s*,\s*/', $v);
            for ($i = 0; $i < count($bgr); $i++) {
                $bgr[$i] = preg_replace('/@/', ',', $bgr[$i]);
            }

            
            $startStops = 0;
            $pos_angle = false;
            $shape_size = false;
            $first = preg_split('/\s+/', trim($bgr[0]));
            $checkCol = $this->mpdf->ConvertColor($first[0]);
            if (preg_match('/(left|center|right|bottom|top|deg|grad|rad)/i', $bgr[0]) && !preg_match('/(<#|rgb|rgba|hsl|hsla)/i', $bgr[0])) {
                $startStops = 1;
                $pos_angle = $bgr[0];
            } else if (trim($first[(count($first) - 1)]) === "0") {
                $startStops = 1;
                $pos_angle = $bgr[0];
            } else if (preg_match('/(circle|ellipse|closest-side|closest-corner|farthest-side|farthest-corner|contain|cover)/i', $bgr[0])) {
                $startStops = 1;
                $shape_size = $bgr[0];
            } else if (!$checkCol) {
                $startStops = 1;
                $pos_angle = $bgr[0];
            }
            if (preg_match('/(circle|ellipse|closest-side|closest-corner|farthest-side|farthest-corner|contain|cover)/i', $bgr[1])) {
                $startStops = 2;
                $shape_size = $bgr[1];
            }

            
            if ($pos_angle) { 
                
                if (preg_match('/left/i', $pos_angle)) {
                    $startx = 0;
                } else if (preg_match('/right/i', $pos_angle)) {
                    $startx = 1;
                }
                if (preg_match('/top/i', $pos_angle)) {
                    $starty = 1;
                } else if (preg_match('/bottom/i', $pos_angle)) {
                    $starty = 0;
                }
                
                if (preg_match('/(\d+)[%]/i', $first[0], $m)) {
                    $startx = $m[1] / 100;
                } else if (!isset($startx) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $first[0], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    if ($tmp) {
                        $startx = $m[1];
                    }
                }
                if (isset($first[1]) && preg_match('/(\d+)[%]/i', $first[1], $m)) {
                    $starty = 1 - ($m[1] / 100);
                } else if (!isset($starty) && isset($first[1]) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $first[1], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    if ($tmp) {
                        $starty = $m[1];
                    }
                }

                if (!isset($starty)) {
                    $starty = 0.5;
                }
                if (!isset($startx)) {
                    $startx = 0.5;
                }
            }
            
            else {
                $starty = 0.5;
                $startx = 0.5;
                $endy = 0.5;
                $endx = 0.5;
            }

            
            $shape = 'ellipse'; 
            $size = 'farthest-corner';
            if ($shape_size) { 
                if (preg_match('/(circle|ellipse)/i', $shape_size, $m)) {
                    $shape = $m[1];
                }
                if (preg_match('/(closest-side|closest-corner|farthest-side|farthest-corner|contain|cover)/i', $shape_size, $m)) {
                    $size = $m[1];
                    if ($size == 'contain') {
                        $size = 'closest-side';
                    } else if ($size == 'cover') {
                        $size = 'farthest-corner';
                    }
                }
            }

            $coords = array();
            if (!isset($startx)) {
                $startx = false;
            }
            if (!isset($starty)) {
                $starty = false;
            }
            if (!isset($endx)) {
                $endx = false;
            }
            if (!isset($endy)) {
                $endy = false;
            }
            if (!isset($radius)) {
                $radius = false;
            }
            if (!isset($angle)) {
                $angle = 0;
            }
            $g['coords'] = array($startx, $starty, $endx, $endy, $radius, $angle, $shape, $size, $repeat);

            $g['stops'] = array();
            for ($i = $startStops; $i < count($bgr); $i++) {
                $stop = array();
                
                $el = preg_split('/\s+/', trim($bgr[$i]));

                $col = $this->mpdf->ConvertColor($el[0]);
                if ($col) {
                    $stop['col'] = $col;
                } else {
                    $stop['col'] = $col = $this->mpdf->ConvertColor(255);
                }
                if ($col{0} == 1) {
                    $g['colorspace'] = 'Gray';
                } else if ($col{0} == 4 || $col{0} == 6) {
                    $g['colorspace'] = 'CMYK';
                }
                if ($col{0} == 5) {
                    $stop['opacity'] = ord($col{4}) / 100;
                } 
                else if ($col{0} == 6) {
                    $stop['opacity'] = ord($col{5}) / 100;
                } 
                else if ($col{0} == 1 && $col{2} == 1) {
                    $stop['opacity'] = ord($col{3}) / 100;
                } 

                if (isset($el[1]) && preg_match('/(\d+)[%]/', $el[1], $m)) {
                    $stop['offset'] = $m[1] / 100;
                    if ($stop['offset'] > 1) {
                        unset($stop['offset']);
                    }
                } else if (isset($el[1]) && preg_match('/([0-9.]+(px|em|ex|pc|pt|cm|mm|in))/i', $el[1], $m)) {
                    $tmp = $this->mpdf->ConvertSize($m[1], $this->mpdf->w, $this->mpdf->FontSize, false);
                    $stop['offset'] = $el[1];
                }
                $g['stops'][] = $stop;
            }
            if (count($g['stops'])) {
                return $g;
            }
        }
        return array();
    }

    function parseBackgroundGradient($bg)
    {
    

        $v = trim($bg);
        $bgr = preg_split('/\s+/', $v);
        $g = array();
        if (count($bgr) > 6) {
            if (strtoupper(substr($bgr[0], 0, 1)) == 'L' && count($bgr) == 7) {  
                $g['type'] = 2;
                
                $g['coords'] = array($bgr[3], $bgr[4], $bgr[5], $bgr[6]);
            } else if (count($bgr) == 8) { 
                $g['type'] = 3;
                
                $g['coords'] = array($bgr[3], $bgr[4], $bgr[5], $bgr[6], $bgr[7]);
            }
            $g['colorspace'] = 'RGB';
            
            $cor = $this->mpdf->ConvertColor($bgr[1]);
            if ($cor{0} == 1) {
                $g['colorspace'] = 'Gray';
            } else if ($cor{0} == 4 || $cor{0} == 6) {
                $g['colorspace'] = 'CMYK';
            }
            if ($cor) {
                $g['col'] = $cor;
            } else {
                $g['col'] = $this->mpdf->ConvertColor(255);
            }
            $cor = $this->mpdf->ConvertColor($bgr[2]);
            if ($cor) {
                $g['col2'] = $cor;
            } else {
                $g['col2'] = $this->mpdf->ConvertColor(255);
            }
            $g['extend'] = array('true', 'true');
            $g['stops'] = array(array('col' => $g['col'], 'opacity' => 1, 'offset' => 0), array('col' => $g['col2'], 'opacity' => 1, 'offset' => 1));
            return $g;
        }
        return false;
    }

}
