<?php

/*
  mPDF recognises IETF language tags as:
  - a single primary language subtag composed of a two letter language code from ISO 639-1 (2002), or a three letter code from ISO 639-2 (1998), ISO 639-3 (2007) or ISO 639-5 (2008) (usually written in lower case);
  - an optional script subtag, composed of a four letter script code from ISO 15924 (usually written in title case);
  - an optional region subtag composed of a two letter country code from ISO 3166-1 alpha-2 (usually written in upper case), or a three digit code from UN M.49 for geographical regions;
  Subtags are not case sensitive, but the specification recommends using the same case as in the Language Subtag Registry, where region subtags are uppercase, script subtags are titlecase and all other subtags are lowercase.

  Region subtags are often deprecated by the registration of specific primary language subtags from ISO 639-3 which are now "preferred values". For example, "ar-DZ" is deprecated with the preferred value "arq" for Algerian Spoken Arabic;

  Example: Serbian written in the Cyrillic (sr-Cyrl) or Latin (sr-Latn) script

  und (for undetermined or undefined) is used in situations in which a script must be indicated but the language cannot be identified.
  e.g. und-Cyrl is an undefined language written in Cyrillic script.
 */

function GetLangOpts($llcc, $adobeCJK, &$fontdata)
{
    $tags = preg_split('/-/', $llcc);
    $lang = strtolower($tags[0]);
    $country = '';
    $script = '';
    if (isset($tags[1]) && $tags[1]) {
        if (strlen($tags[1]) == 4) {
            $script = strtolower($tags[1]);
        } else {
            $country = strtolower($tags[1]);
        }
    }
    if (isset($tags[2]) && $tags[2]) {
        $country = strtolower($tags[2]);
    }

    $unifont = "";
    $coreSuitable = false;

    switch ($lang) {

     /* European */
    CASE "en": CASE "eng": 
        CASE "eu": CASE "eus": 
            CASE "br": CASE "bre": 
                CASE "ca": CASE "cat": 
                    CASE "co": CASE "cos": 
                        CASE "kw": CASE "cor": 
                            CASE "cy": CASE "cym": 
                                CASE "cs": CASE "ces": 
                                    CASE "da": CASE "dan": 
                                        CASE "nl": CASE "nld": 
                                            CASE "et": CASE "est": 
                                                CASE "fo": CASE "fao": 
                                                    CASE "fi": CASE "fin": 
                                                        CASE "fr": CASE "fra": 
                                                            CASE "gl": CASE "glg": 
                                                                CASE "de": CASE "deu": 
                                                                    CASE "ht": CASE "hat": 
                                                                        CASE "hu": CASE "hun": 
                                                                            CASE "ga": CASE "gle": 
                                                                                CASE "is": CASE "isl": 
                                                                                    CASE "it": CASE "ita": 
                                                                                        CASE "la": CASE "lat": 
                                                                                            CASE "lb": CASE "ltz": 
                                                                                                CASE "li": CASE "lim": 
                                                                                                    CASE "lt": CASE "lit": 
                                                                                                        CASE "lv": CASE "lav": 
                                                                                                            CASE "gv": CASE "glv": 
                                                                                                                CASE "no": CASE "nor": 
                                                                                                                    CASE "nn": CASE "nno": 
                                                                                                                        CASE "nb": CASE "nob": 
                                                                                                                            CASE "pl": CASE "pol": 
                                                                                                                                CASE "pt": CASE "por": 
                                                                                                                                    CASE "ro": CASE "ron": 
                                                                                                                                        CASE "gd": CASE "gla": 
                                                                                                                                            CASE "es": CASE "spa": 
                                                                                                                                                CASE "sv": CASE "swe": 
                                                                                                                                                    CASE "sl": CASE "slv": 
                                                                                                                                                        CASE "sk": CASE "slk": 
                                                                                                                                                                                                                                                                                                         $unifont = "dejavusanscondensed";
            
                                                                                                                                                                                                                                                                                                         $coreSuitable = true;
        break;

    CASE "ru": CASE "rus": 
        CASE "ab": CASE "abk": 
            CASE "av": CASE "ava": 
                CASE "ba": CASE "bak": 
                    CASE "be": CASE "bel": 
                        CASE "bg": CASE "bul":
                            CASE "ce": CASE "che": 
                                CASE "cv": CASE "chv": 
                                    CASE "kk": CASE "kaz": 
                                        CASE "kv": CASE "kom": 
                                            CASE "ky": CASE "kir": 
                                                CASE "mk": CASE "mkd": 
                                                    CASE "cu": CASE "chu": 
                                                        CASE "os": CASE "oss": 
                                                            CASE "sr": CASE "srp": 
                                                                CASE "tg": CASE "tgk": 
                                                                    CASE "tt": CASE "tat": 
                                                                        CASE "tk": CASE "tuk":
                                                                            CASE "uk": CASE "ukr": 
                                                                                                                                              $unifont = "dejavusanscondensed"; /* freeserif best coverage for supplements etc. */
        break;

    CASE "hy": CASE "hye": 
            $unifont = "dejavusans";
        break;
    CASE "ka": CASE "kat": 
            $unifont = "dejavusans";
        break;

    CASE "el": CASE "ell": 
            $unifont = "dejavusanscondensed";
        break;
    CASE "cop":  
        $unifont = "quivira";
        break;

    CASE "got":  
        $unifont = "freeserif";
        break;

        /* African */
    CASE "nqo":  
        $unifont = "dejavusans";
        break;
        
    CASE "vai":  
        $unifont = "freesans";
        break;
    CASE "am": CASE "amh": 
        CASE "ti": CASE "tir": 
                $unifont = "abyssinicasil";
        break;



        /* Middle Eastern */
    CASE "ar": CASE "ara": 
            $unifont = "xbriyaz";
        break;
    CASE "fa": CASE "fas": 
            $unifont = "xbriyaz";
        break;
    CASE "ps": CASE "pus": 
            $unifont = "xbriyaz";
        break;
    CASE "ku": CASE "kur":
            $unifont = "xbriyaz";
        break;
    CASE "ur": CASE "urd": 
            $unifont = "xbriyaz";
        break;
    CASE "he": CASE "heb": 
        CASE "yi": CASE "yid": 
                $unifont = "taameydavidclm"; 
        break;


    CASE "syr":  
        $unifont = "estrangeloedessa";
        break;


        
    CASE "xcr":  
        $unifont = "aegean";
        break;
    CASE "xlc": 
        $unifont = "aegean";
        break;
    CASE "xld":  
        $unifont = "aegean";
        break;
        
    CASE "phn":  
        $unifont = "aegean";
        break;
        
    CASE "uga": 
        $unifont = "aegean";
        break;


        /* Central Asian */
    CASE "bo": CASE "bod": 
        CASE "dz": CASE "dzo": 
                $unifont = "jomolhari";
        break;
        

        /* South Asian */
    CASE "as": CASE "asm": 
            $unifont = "freeserif";
        break;
    CASE "bn": CASE "ben": 
            $unifont = "freeserif";
        break;
    CASE "ks": CASE "kas": 
            $unifont = "freeserif";
        break;
    CASE "hi": CASE "hin": 
        CASE "bh": CASE "bih": 
            CASE "sa": CASE "san": 
                    $unifont = "freeserif";
        break;
    CASE "gu": CASE "guj": 
            $unifont = "freeserif";
        break;
    CASE "pa": CASE "pan": 
            $unifont = "freeserif";
        break;
    CASE "kn": CASE "kan": 
            $unifont = "lohitkannada";
        break;
    CASE "mr": CASE "mar": 
            $unifont = "freeserif";
        break;
    CASE "ml": CASE "mal": 
            $unifont = "freeserif";
        break;
    CASE "ne": CASE "nep": 
            $unifont = "freeserif";
        break;
    CASE "or": CASE "ori": 
            $unifont = "freeserif";
        break;
    CASE "si": CASE "sin": 
            $unifont = "kaputaunicode";
        break;
    CASE "ta": CASE "tam": 
            $unifont = "freeserif";
        break;
    CASE "te": CASE "tel": 
            $unifont = "pothana2000";
        break;


        
    CASE "sd": CASE "snd": 
            if ($country == "IN") {
                $unifont = "freeserif";
            } else if ($country == "PK") {
                $unifont = "lateef";
            } else {
                $unifont = "lateef";
            }
        break;


        
    CASE "lif":  
        $unifont = "sun-exta";
        break;
        
    CASE "syl": 
        $unifont = "mph2bdamase";
        break;
        
    CASE "dv": CASE "div": 
            $unifont = "freeserif";
        break;



        
    CASE "km": CASE "khm": 
            $unifont = "khmeros";
        break;
    CASE "lo": CASE "lao": 
            $unifont = "dhyana";
        break;
    CASE "my": CASE "mya": 
            $unifont = "tharlon"; 

        break;
    CASE "th": CASE "tha": 
            $unifont = "garuda";
        break;



        
    CASE "vi": CASE "vie": 
            $unifont = "dejavusanscondensed";
        break;


        
    CASE "bug":  
        $unifont = "freeserif";
        break;
        
    CASE "su":  
        $unifont = "sundaneseunicode";
        break;
    CASE "tdd":  
        $unifont = "tharlon";
        break;
    CASE "blt": 
        $unifont = "taiheritagepro";
        break;


        /* Phillipine */
    CASE "bku":  
        $unifont = "quivira";
        break;
    CASE "hnn":  
        $unifont = "quivira";
        break;
    CASE "tl":  
        $unifont = "quivira";
        break;
    CASE "tbw":  
        $unifont = "quivira";
        break;


        /* East Asian */
    CASE "zh": CASE "zho": 
            if ($country == "HK" || $country == "TW") {
                if ($adobeCJK) {
                    $unifont = "big5";
                } else {
                    $unifont = "sun-exta";
                }
            } else if ($country == "CN") {
                if ($adobeCJK) {
                    $unifont = "gb";
                } else {
                    $unifont = "sun-exta";
                }
            } else {
                if ($adobeCJK) {
                    $unifont = "gb";
                } else {
                    $unifont = "sun-exta";
                }
            }
        break;
    CASE "ko": CASE "kor": 
            if ($adobeCJK) {
                $unifont = "uhc";
            } else {
                $unifont = "unbatang";
            }
        break;
    CASE "ja": CASE "jpn": 
            if ($adobeCJK) {
                $unifont = "sjis";
            } else {
                $unifont = "sun-exta";
            }
        break;
    CASE "ii": CASE "iii": 
            if ($adobeCJK) {
                $unifont = "gb";
            } else {
                $unifont = "sun-exta";
            }
        CASE "lis":  
            $unifont = "quivira";
        break;



        /* American */
    CASE "chr":  
    CASE "oj": CASE "oji": 
        CASE "cr": CASE "cre": 
            CASE "iu": CASE "iku": 
                    $unifont = "aboriginalsans";
        break;


        /* Undetermined language - script used */
    CASE "und":

        switch ($script) {

         /* European */
        CASE "latn": 
            $unifont = "dejavusanscondensed";
            break;
        CASE "cyrl": 
            $unifont = "dejavusanscondensed"; /* freeserif best coverage for supplements etc. */
            break;
        CASE "cprt": 
            $unifont = "aegean";
            break;
        CASE "glag": 
            $unifont = "mph2bdamase";
            break;
        CASE "linb": 
            $unifont = "aegean";
            break;
        CASE "ogam": 
            $unifont = "dejavusans";
            break;
        CASE "ital": 
            $unifont = "aegean";
            break;
        CASE "runr": 
            $unifont = "sun-exta";
            break;
        CASE "shaw": 
            $unifont = "mph2bdamase";
            break;

                /* African */
        CASE "egyp": 
            $unifont = "aegyptus";
            break;
        CASE "ethi": 
            $unifont = "abyssinicasil";
            break;
                
        CASE "osma": 
            $unifont = "mph2bdamase";
            break;
        CASE "tfng": 
            $unifont = "dejavusans";
            break;

                /* Middle Eastern */
        CASE "arab":  
            $unifont = "xbriyaz";
            break;
        CASE "xsux": 
            $unifont = "akkadian";
            break;
                

                /* Central Asian */
                
                /* South Asian */
                
        CASE "khar": 
            $unifont = "mph2bdamase";
            break;
        CASE "mtei": 
            $unifont = "eeyekunicode";
            break;
                

                /* South East Asian */
        CASE "kali": 
            $unifont = "freemono";
            break;
                
        CASE "lana": 
            $unifont = "lannaalif";
            break;
        CASE "talu": 
            $unifont = "daibannasilbook";
            break;

                /* East Asian */
        CASE "hans": 
            if ($adobeCJK) {
                $unifont = "gb";
            } else {
                $unifont = "sun-exta";
            }
            break;
        CASE "bopo": 
            $unifont = "sun-exta";
            break;
                
        CASE "yiii": 
            $unifont = "sun-exta";
            break;

                /* American */
        CASE "dsrt": 
            $unifont = "mph2bdamase";
            break;

                /* Other */
        CASE "brai": 
            $unifont = "dejavusans";
            break;
        } 
        break;
    } 


    return array($coreSuitable, $unifont);
}
